#include "boundcurves.h"
#include "partop.h"


boundcurves::boundcurves(long tnbn,long *multip,long **cnbn,long *coarseidentif,long *coarsenadjac,long **coarseadjac,FILE *outfil)
{
  long i,j;
  ncn = tnbn;
  multiplicity = multip;
  cnbn = cnbn;
  
  identif = new long [ncn];
  for(i = 0; i < ncn; i++){
    identif[i] = coarseidentif[i];
  }
  
  nadjacnod = new long[ncn];
  adjacnod = new long*[ncn];
  for(i = 0; i < ncn; i++){
    nadjacnod[i] = coarsenadjac[i];
    adjacnod[i] = new long[nadjacnod[i]];
    for(j = 0; j < nadjacnod[i]; j++){
      adjacnod[i][j] = coarseadjac[i][j];
    }
  }
  out = outfil;
  
  // number of curves
  ncurves = -1;
  // list of start nodes of curves
  start = NULL;
  // list of end nodes of curves
  end =NULL;
  // list of members of curves
  members = NULL;
  // 
  nmembers = NULL;
  //
  nedges = NULL;
  
  // number of subdomain where the curves lies
  ndomid = NULL;
  // 
  domid = NULL;

  // corespondence between coarse numbers of members of curves
  memcor = NULL;

  // reduced graph?
  reduce_flag = -1;

  reducecor = NULL;
  correduce = NULL; 
  
}

boundcurves::~boundcurves(){

  //delete []multiplicity;
  delete []identif;
  //delete []cnbn;
  long j;
  delete []start;
  delete []end;
  
  for(j = 0; j < ncn; j++){
    delete []adjacnod[j];
  }
  delete []nadjacnod;
  delete []adjacnod;
  for(j = 0; j < ncurves; j++){
    delete []members[j];
  }
  delete []nedges;
  delete []nmembers;

  delete []reducecor;
  delete []correduce;
  
}
/*
  function reduce graph
 */

void boundcurves::reduce_graph()
{
  long i,j,k,m,l;
  long *auxnadjac,**auxadjac,*auxidentif;
  long auxncn;
  
  reduce_flag = 1;
  
  auxncn = 0;
  for(i = 0; i < ncn; i++){
    if(multiplicity[i] > 2){
	auxncn++;
    }
  }
  auxnadjac = new long[auxncn];
  auxadjac = new long* [auxncn];
  reducecor = new long[auxncn];
  correduce = new long[ncn]; 
  auxidentif = new long[auxncn];
  k = 0;
  for(i = 0; i < ncn; i++){
    if(multiplicity[i] > 2){
      auxnadjac[k] = 0;
      reducecor[k] = i;
      correduce[i] = k;
      auxidentif[k] = identif[i];
      for(j = 0; j < nadjacnod[i]; j++){
	if(multiplicity[adjacnod[i][j]] > 2 ){
	  auxnadjac[k]++;
	}
      }
      auxadjac[k] = new long[auxnadjac[k]];
      m = 0;
      for(j = 0; j < nadjacnod[i]; j++){
	if(multiplicity[adjacnod[i][j]] > 2 ){
	  auxadjac[k][m] = adjacnod[i][j];
	  m++;
	}
      }
      k++;
    }
    else{
      correduce[i] = -1;
    }
  }
  //fprintf(out,"auxncn je %ld\n",auxncn);    
  
  // rewrite to reduced numbering
  for(i = 0; i < auxncn; i++){
    for(j = 0; j < auxnadjac[i]; j++){
      m = auxadjac[i][j];
      if(correduce[m] != -1){
	auxadjac[i][j] = correduce[m];
      }
    }
  }
  
  
  for(i = 0; i < ncn; i++){
    delete []adjacnod[i];
  }
  delete []adjacnod;
  delete []nadjacnod;
  delete []identif;
  ncn = auxncn;
  adjacnod = new long*[ncn];
  nadjacnod = new long[ncn];
  identif = new long[ncn];
  for(i = 0; i < ncn; i++){
    identif[i] = auxidentif[i];
    nadjacnod[i] = auxnadjac[i];
    adjacnod[i] = new long[nadjacnod[i]];
    for(j = 0; j < nadjacnod[i]; j++){
      adjacnod[i][j] = auxadjac[i][j];
    }
  }
  
  //
  // detection of cross points 
  for( i = 0; i < ncn; i++){
    if(nadjacnod[i] > 2 && identif[i] != 3){
      identif[i] = 3;
    }
    if(nadjacnod[i] == 1 && identif[i] != 3){
      identif[i] = 3;
    }
  }
  
  
//   for(i = 0; i < ncn; i++){
//     fprintf(out,"uzel %ld identif %ld ma %ld sousedu:",i,identif[i],auxnadjac[i]);
//     for(j = 0; j < nadjacnod[i]; j++){
//       fprintf(out,"   %ld",adjacnod[i][j]);
//     }
//     fprintf(out,"\n");
//   }

  
  
  
  
  for(i = 0; i < ncn; i++){
    delete []auxadjac[i];
  }
  delete []auxadjac;
  delete []auxnadjac;
  delete []auxidentif;
  
}


void boundcurves::set_curves( )
{
  long i,j,k;
  long cor,comb,max;
  long prev,cur,stop,n,m,a,b,z,v;
  long *startnod,*endnod,**corneib,*corpoint;
  long **auxmemb;
  long *nauxmemb;
  
  // compution of number of start nodes and check of array identif
  cor = 0;
  if(identif != NULL){
    for( i = 0; i < ncn; i++){
      if(identif[i] == 3 ){
	cor++;
      }
    }
    //fprintf(out,"cor je %ld\n",cor);    
    if(cor == 0){
      for(i = 0; i < ncn; i++){
	if(nadjacnod[i] == 1 && identif[i] != 3 ){
	  identif[i] = 3;
	  cor++;
	}
	if(nadjacnod[i] > 2 && identif[i] != 3){
	  identif[i] = 3;
	  cor++;
	}
      }
    } 
  }
  else{
    identif = new long[ncn];
    for(i = 0; i < ncn; i++){
      if(nadjacnod[i] == 1){
	identif[i] = 3;
      }
      if(nadjacnod[i] > 2){
	identif[i] = 3;
	}
    }
  }
  //fprintf(out,"cor je %ld\n",cor);    
  // allocation of auxiliarry array and estimation of number of curves
  corneib = new long*[cor];
  corpoint = new long[cor];
  k = 0;
  comb = 0;
  for(i = 0; i < ncn; i++){
    if(identif[i] == 3 && nadjacnod[i] != 0){
      comb += nadjacnod[i];
      corpoint[k] = i;
      corneib[k] = new long[nadjacnod[i]];
      for(j = 0; j < nadjacnod[i]; j++){
	corneib[k][j] = -1;
      }
      k++;
    }
  }
  fprintf(out,"Estimate number of curves %ld\n",comb);

  //for(i = 0; i < cor; i++){
  //fprintf(out,"%ld corpoint %ld\n",i,corpoint[i]);
  //for(j = 0; j < nadjacnod[corpoint[i]]; j++){
  //fprintf(out,"   %ld ",corneib[i][j]);
  //}
  //fprintf(out,"\n");
  //}
  
  // old estimation of number of curves
  //comb = compute_number_of_combination(cor,2);
  
  
  startnod = new long[comb];
  endnod = new long[comb];
  auxmemb = new long*[comb];
  nauxmemb = new long[comb];
  
  
  m = 0;
  for(i = 0; i < cor; i++){
    cur = corpoint[i];
    for(j = 0; j < nadjacnod[corpoint[i]]; j++){
      prev = corpoint[i];
      if(corneib[i][j] == -1){
	startnod[m] = corpoint[i];
	//fprintf(out,"startnode %ld  = %ld\n",i,startnod[m]);
	stop = 0;
	n = 0;
	corneib[i][j] = 0;
	cur = adjacnod[corpoint[i]][j];
	nauxmemb[m] = 1;
	auxmemb[m] = new long[ncn];
	auxmemb[m][n] = cur;
	n++;
	while(stop != 1){
	  a = nadjacnod [cur];
	  for(k = 0; k < a; k++){
	    b = adjacnod[cur][k];
	    if(b != prev){
	      if(identif[b] == 2){
		auxmemb[m][n]=b;
		n++;
		prev = cur;
		cur = b;
		stop = 0;
	      }
	      if(identif[b] == 3){
		stop = 1;
		endnod[m] = b;
		nauxmemb[m]= n;
		//fprintf(out,"konecny bod je %ld\n",b);
		// nastaveni stopky
		for(z = 0; z < cor; z++){
		  if(b == corpoint[z]){
		    //fprintf(out,"corpoint je %ld\n",z);
		    for(v = 0; v < nadjacnod[corpoint[z]]; v++){
		      if(cur == adjacnod[corpoint[z]][v]){
			corneib[z][v] = 0;
			//fprintf(out,"nastavena nula na %ld %ld\n",cur,v);
			break;
		      }
		    }
		    break;
		  }
		}
	      }
	    }
	  }
	}
	m++;
      }
    }
  }
  fprintf(out,"Number of curve %ld\n",m);
  
  ncurves = m;
  start = new long[ncurves];
  // list of end nodes of curves
  end = new long[ncurves];
  // list of members of curves
  members = new long*[ncurves];
  // 
  nmembers = new long[ncurves];
  //
  nedges = new long[ncurves];
  for(i = 0; i < ncurves; i++){
    start[i] = startnod[i];
    end[i] = endnod[i];
    nmembers[i] = nauxmemb[i];
    members[i] = new long[nmembers[i]];
    for(j = 0; j < nmembers[i]; j++){
      members[i][j] = auxmemb[i][j];
    }
    nedges[i] = nmembers[i] + 1; 
  }
  
  delete []startnod;
  delete []endnod;
  delete []nauxmemb;
  for(i = 0; i < ncurves; i++){
    delete []auxmemb[i];
  }
  delete []auxmemb;
  delete []corpoint;
  for(i = 0; i < cor; i++){
    delete []corneib[i];
  }
  delete []corneib;
  
  
}

void boundcurves::set_curves_3D( )
{
  long i,j,k;
  long cor,comb,max;
  long prev,cur,stop,n,m,a,b,z,v;
  long *startnod,*endnod,**corneib,*corpoint;
  long **auxmemb;
  long *nauxmemb;
  
  cor = 0;
  if(identif != NULL){
    for( i = 0; i < ncn; i++){
      if(identif[i] == 3 ){
	cor++;
      }
    }
    //fprintf(out,"cor je %ld\n",cor);    
    if(cor == 0){
      for(i = 0; i < ncn; i++){
	if(nadjacnod[i] == 1){
	  identif[i] = 3;
	}
       }
    } 
  }
  else{
    identif = new long[ncn];
    for(i = 0; i < ncn; i++){
      if(nadjacnod[i] == 1){
	identif[i] = 3;
      }
    }
  }

  if(cor == 0){
    for(i = 0; i < ncn; i++){
      if(nadjacnod[i] == 1){
	identif[i] = 3;
      }
    }
  }
  
  cor = 0;
  for( i = 0; i < ncn; i++){
    if(identif[i] == 3 ){
      cor++;
    }
  }
  
  //fprintf(out,"cor je %ld\n",cor);    

  
  corneib = new long*[cor];
  corpoint = new long[cor];
  k = 0;
  comb = 0;
  for(i = 0; i < ncn; i++){
    if(identif[i] == 3){
      comb += nadjacnod[i];
      corpoint[k] = i;
      corneib[k] = new long[nadjacnod[i]];
      for(j = 0; j < nadjacnod[i]; j++){
	corneib[k][j] = -1;
      }
      k++;
    }
  }
  
  //for(i = 0; i < cor; i++){
  //fprintf(out,"%ld corpoint %ld\n",i,corpoint[i]);
  //for(j = 0; j < nadjacnod[corpoint[i]]; j++){
  //fprintf(out,"   %ld ",corneib[i][j]);
  //}
  //fprintf(out,"\n");
  //}
  //old estimation of number of curves
  //comb = compute_number_of_combination(cor,2);
  fprintf(out,"Estimate number of curve %ld\n",comb);
  
  startnod = new long[comb];
  endnod = new long[comb];
  auxmemb = new long*[comb];
  nauxmemb = new long[comb];
  
  
  m = 0;
  for(i = 0; i < cor; i++){
    cur = corpoint[i];
    for(j = 0; j < nadjacnod[corpoint[i]]; j++){
      prev = corpoint[i];
      if(corneib[i][j] == -1){
	startnod[m] = corpoint[i];
	//fprintf(out,"startovaci bod je %ld\n",startnod[m]);
	stop = 0;
	n = 0;
	corneib[i][j] = 0;
	cur = adjacnod[corpoint[i]][j];
	//fprintf(out,"cur = %ld\n",cur);
	if(identif[cur] != 3){
	  nauxmemb[m] = 1;
	  auxmemb[m] = new long[ncn];
	  auxmemb[m][n] = cur;
	  n++;
	  while(stop != 1 ){
	    a = nadjacnod [cur];
	    for(k = 0; k < a; k++){
	      b = adjacnod[cur][k];
	      if(b != prev){
		//fprintf(out,"b=%ld\n",b);
		if(identif[b] == 2){
		  auxmemb[m][n]=b;
		  n++;
		  prev = cur;
		  cur = b;
		  // fprintf(out,"cur = %ld\n",cur);
		  stop = 0;
		  //break;
		}
		if(identif[b] == 3){
		  stop = 1;
		  endnod[m] = b;
		  nauxmemb[m]= n;
		  //fprintf(out,"konecny bod je %ld\n",b);
		  // nastaveni stopky
		  for(z = 0; z < cor; z++){
		    if(b == corpoint[z]){
		      //fprintf(out,"corpoint je %ld\n",z);
		      for(v = 0; v < nadjacnod[corpoint[z]]; v++){
			if(cur == adjacnod[corpoint[z]][v]){
			  corneib[z][v] = 0;
			  //fprintf(out,"nastavena nula na %ld %ld\n",cur,v);
			  break;
			}
		      }
		      break;
		    }
		  }
		  break;
		}
	      }
	    }
	  }
	  m++;
	}
	else{
	  //fprintf(out,"konecny bod je %ld\n",cur);
	  nauxmemb[m] = 0;
	  endnod[m] = cur;
	  // identfikace cile
	  for(z = 0; z < cor; z++){
	    if(cur == corpoint[z]){
	      //fprintf(out,"corpoint je %ld\n",z);
	      for(v = 0; v < nadjacnod[corpoint[z]]; v++){
		if(prev == adjacnod[corpoint[z]][v]){
		  corneib[z][v] = 0;
		  //fprintf(out,"nastavena nula na %ld %ld\n",cur,v);
		  break;
		}
	      }
	      break;
	    }
	  }
	}
      }
    }
  }
  fprintf(out,"Number of curve %ld\n",m);
  
  ncurves = m;
  start = new long[ncurves];
  // list of end nodes of curves
  end = new long[ncurves];
  // list of members of curves
  members = new long*[ncurves];
  // 
  nmembers = new long[ncurves];
  //
  nedges = new long[ncurves];
  for(i = 0; i < ncurves; i++){
    start[i] = startnod[i];
    end[i] = endnod[i];
    nmembers[i] = nauxmemb[i];
    members[i] = new long[nmembers[i]];
    for(j = 0; j < nmembers[i]; j++){
      members[i][j] = auxmemb[i][j];
    }
    nedges[i] = nmembers[i] + 1; 
  }
  
  delete []startnod;
  delete []endnod;
  delete []nauxmemb;
  for(i = 0; i < ncurves; i++){
    delete []auxmemb[i];
  }
  delete []auxmemb;
  delete []corpoint;
  for(i = 0; i < cor; i++){
    delete []corneib[i];
  }
  delete []corneib;
    
  
}


long boundcurves::compute_number_of_combination(long n, long k)
{
  long i;
  long denom,nom,diff;
  long combination;
  
  diff = n - k;
  nom = 1;
  for(i = diff+1; i < n+1; i++){
    nom *= i;
  }
  denom = 1;
  for(i = 2; i < k+1; i++){
    denom *= i;
  }
  combination = nom/denom;
  
  return(combination);
}

void boundcurves::print_info()
{
  long i,j;
  
  fprintf(out,"Number of curves %ld\n",ncurves);
  for(i = 0; i < ncurves; i++){
    fprintf(out,"Curve %ld:\n",i+1);
    fprintf(out,"Start %ld\n",start[i]);
    fprintf(out,"End %ld\n",end[i]);
    fprintf(out,"Number of members %ld\n",nmembers[i]);
    fprintf(out,"Members: ");
    for(j = 0; j < nmembers[i]; j++){
      fprintf(out,"   %ld",members[i][j]);
    }
    fprintf(out,"\n");
    fprintf(out,"Number of edges %ld\n",nedges[i]);
  }
}

void boundcurves::add_corners(long condcorner,long typecondcorner,long nmember,long intpointo)
{
  
  switch(typecondcorner){
  case each_nth_memb:{
    add_each_nth_member(nmember);
    break;
  }
  case nth_memb:{
    add_nth_member(nmember);
    break;
  }
  case all_memb:{
    add_all_members( );
    break;
  }
  case rand_memb:{
    add_rand(nmember);
    break;
  }
  case centroid:{
    add_centroid();
    break;
  }
  case int_memb:{
    add_intpoint(intpointo);
    break;
  }
  case n_part_curve:{
    add_n_part_curve(nmember);
    break;
  }
  }
  

}

void boundcurves::add_rand(long nmember)
{
  long i,j,k;
  long member;
  
  // inicialization of pseudorandom numbers
  srand((unsigned int) time(NULL));
  
  for(i = 0; i < ncurves; i++){
    if(nmembers[i] != 0){
      for(j = 0; j < nmember; j++){
	member = rand()%nmembers[i];
	//fprintf(out,"Select will be %ld-th node %ld\n",member,members[i][member]);
	k = members[i][member];
	identif[k] = 3;
      }
    }
  }
  fflush(out);
  
}



void boundcurves::add_each_nth_member(long nmember)
{
  long i,j,stop,s,e;
  double zbytek;
  long cycles,member,lnmember;
  
  for(i = 0; i < ncurves; i++){
    //fprintf(out,"krivka cislo %ld\n",i);
    lnmember = nmember;
    if(nmembers[i] != 0){
      if(lnmember > nedges[i]){
	//fprintf(out,"nmember is too big\n");
	lnmember = nedges[i]-1;
      }
      zbytek = nedges[i] % lnmember;
      if(zbytek == 0){
	cycles = nedges[i] / lnmember;
      }
      else{
	// nova volba nmember
	stop = 0;
	e = 0;
	while (stop == 0){
	  zbytek = nedges[i] % lnmember;
	  if( zbytek != 0){
	    //fprintf(out,"zbytek = %lf\n",zbytek);
	    if(e == 0){
	      lnmember++;
	    }
	    if(e == 1){
	      lnmember--;
	    }
	    //fprintf(out,"lnmember = %ld\n",lnmember);
	    stop = 0;
	  }
	  else{
	    cycles = nedges[i] / lnmember;
	    //fprintf(out,"nova volba nmember = %ld\n",lnmember);
	    stop = 1;
	  }
	  if( nedges[i] == lnmember){
	    e = 1;
	    lnmember = nmember;
	    stop = 0;
	  }
	  if(lnmember == 1){
	    stop = 1;
	    e = 2;
	  }
	}
      }
      if(e == 2){
	//fprintf(out,"%ld je prvocislo, nelze delit - nelze vybrat kazdy n-ty\nbude vybran n-ty clen\n",nedges[i]);
	member = 0;
	for(j = 0; j < nmembers[i]; j++){
	  member++;
	  if(member == nmember){
	    //  fprintf(out,"vybran bude uzel cislo %ld\n",members[i][j]);
	    member = 0;
	    identif[members[i][j]] = 3;
	  }
	}
      }
      //fprintf(out,"cycles = %ld\n",cycles);
      for(j = 1; j < cycles; j++){
	member = j*lnmember-1;
	//fprintf(out,"vybran bude uzel cislo %ld\n",members[i][member]);
	identif[members[i][member]] = 3;
      }
    }
  }
  fflush(out);
}

void boundcurves::add_nth_member(long nmember)
{
  long i,j;
  long member;
  for(i = 0; i < ncurves; i++){
    //fprintf(out,"krivka cislo %ld\n",i);
    if(nmembers[i] != 0){
      if(nmember >= nmembers[i]){
	//fprintf(out,"nmember is too big\n");
      }
      else{
	member = 0;
	for(j = 0; j < nmembers[i]; j++){
	  member++;
	  if(member == nmember){
	    //  fprintf(out,"vybran bude uzel cislo %ld\n",members[i][j]);
	    member = 0;
	    identif[members[i][j]] = 3;
	  }
	}
      }
    }
  }
  fflush(out);
}

void boundcurves::add_all_members( )
{
  long i,j;
  for(i = 0; i < ncurves; i++){
    //fprintf(out,"krivka cislo %ld\n",i);
    for(j = 0; j < nmembers[i]; j++){
      //fprintf(out,"vybran bude uzel cislo %ld\n",members[i][j]);
      identif[members[i][j]] = 3;
    }
  }
}

void boundcurves:: add_centroid()
{
  long i,j,k;
  long select;
  for(i = 0; i < ncurves; i++){
    if((nedges[i]%2) == 0){
      select = nedges[i]/2;
      select--;
    }
    else{
      k = nedges[i]+1;
      select = k/2;
      select--;
    }
    //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][select]);
    identif[members[i][select]] = 3;
  }

}

void boundcurves:: add_intpoint(long intpointo)
{
  // numbers of nodes which will be added
  long order,i;
  double dpos;
  long lpos;
  // intpointo - number of integration points
  
  switch(intpointo){
  case 1:
    for(i = 0; i < ncurves; i++){
      // 1st
      dpos = 0.5*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
    }
    break;
  case 2:
    for(i = 0; i < ncurves; i++){
      // 1st
      dpos = 0.2*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 2nd
      dpos = 0.8*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
    }
    break;
  case 3:
    for(i = 0; i < ncurves; i++){
      // 1st
      dpos = 0.1*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 2nd
      dpos = 0.5*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 3rd
      dpos = 0.9*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
    }
    break;
  case 4:
    for(i = 0; i < ncurves; i++){
      // 1st
      dpos = 0.07*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 2nd
      dpos = 0.3*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 3rd
      dpos = 0.7*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 4th
      dpos = 0.93*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
    }
    break;
  case 5:
    for(i = 0; i < ncurves; i++){
      // 1st
      dpos = 0.07*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 2nd
      dpos = 0.23*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 3rd
      dpos = 0.5*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 4th
      dpos = 0.77*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 5th
      dpos = 0.93*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
    }
    break;
  case 6:
    for(i = 0; i < ncurves; i++){
      //1st
      dpos = 0.03*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 2nd
      dpos = 0.17*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 3rd
      dpos = 0.38*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 4rd
      dpos = 0.62*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 5th
      dpos = 0.83*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 6th
      dpos = 0.97*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
    }
    break;
  case 7:
    for(i = 0; i < ncurves; i++){
      //1st
      dpos = 0.03*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 2nd
      dpos = 0.13*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 3rd
      dpos = 0.30*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 4th
      dpos = 0.50*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 5th
      dpos = 0.70*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 6th
      dpos = 0.87*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 7th
      dpos = 0.97*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
    }
    break;
  case 8:
    for(i = 0; i < ncurves; i++){
      //1st
      dpos = 0.02*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 2nd
      dpos = 0.10*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 3rd
      dpos = 0.24*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 4th
      dpos = 0.41*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 5th
      dpos = 0.59*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 6th
      dpos = 0.76*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 7th
      dpos = 0.90*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
      // 8th
      dpos = 0.98*nedges[i];
      dpos = round(dpos);
      lpos = (long) dpos;
      lpos--;
      identif[members[i][lpos]] = 3;
      //fprintf(out,"Curve %ld:vybran bude uzel cislo %ld\n",i,members[i][lpos]);
    }
    break;
  }
  
  fflush(out);
}

void boundcurves::add_n_part_curve(long nmember)
{
  long i,j,k;
  long smember,stop,e,lnmember,lnedges;
  double zbytek;
  for(i = 0; i < ncurves; i++){
    //fprintf(out,"krivka cislo %ld nmember %ld nedges %ld\n",i,nmember,nedges[i]);
    if(nedges[i]%nmember == 0) {
      smember = nedges[i]/nmember;
    } 
    else{
      lnmember = nmember;
      lnedges = nedges[i];
      // nova volba nmember
      stop = 0;
      e = 0;
      while (stop == 0){
	zbytek = lnedges % lnmember;
	if( zbytek != 0){
	  //fprintf(out,"zbytek = %lf\n",zbytek);
	  if(e == 0){
	    lnmember--;
	  }
	  if(e == 1){
	    lnmember++;
	  }
	  //fprintf(out,"lnmember = %ld\n",lnmember);
	  stop = 0;
	}
	else{
	  smember = lnedges / lnmember;
	  //fprintf(out,"nova volba nmember = %ld\n",lnmember);
	  stop = 1;
	}
	if( lnedges == lnmember){
	  e = 1;
	  lnmember = nmember;
	  stop = 0;
	}
	if(lnmember == 1){
	  stop = 0;
	  e = 2;
	  //fprintf(out,"%ld je prvocislo, nelze delit - nelze vybrat kazdy n-ty\nbude upraven pocet hran na nedges+1 %ld\n",lnedges);
	  lnedges++;
	  lnmember = nmember;
	}
      }
    }
    //fprintf(out,"cycle %ld nmembers %ld \n",smember,nmembers[i]);
    k = 0;
    for(j = 0; j < nmembers[i];j++){
      k++;
      if(k == smember){
	k = 0;
	//fprintf(out,"vybran bude uzel %ld -ty  cislo %ld\n",j+1,members[i][j]);
	identif[members[i][j]] = 3;
      }
    }
  }
  
}

void boundcurves::get_coarseidentif(long *coarseidentif)
{
  long i;
  if(reduce_flag == 0){
    for(i = 0; i < ncn; i++){
      coarseidentif[i] = identif[i];
    }
  }
  else{
    for(i = 0; i < ncn; i++){
    coarseidentif[reducecor[i]] = identif[i];
    }
  }
}
