#include "parcongrad.h"
#include <string.h>
#include <math.h>
#include <time.h>

parcongrad::parcongrad(int np,int mr,long nd)
{
  nproc=np;  myrank=mr;  ndom=nd;
  ndof=0;  nbdof=0;
  nbn=0;  maxnbdof=0;
}

parcongrad::~parcongrad()
{
}


/**
   function orders unknowns on subdomains
   
   @param top - pointer to domain topology
   
   JK, 4.1.2006
*/
void parcongrad::subdomain_ordering (partop *ptop,gtopology *top,FILE *out)
{
  ndof = top->codenum_generation (out);
  maxndof = ptop->maxndof;
}





/**
   function orders unknowns in coarse problem
   
   @param top - pointer to domain topology
   @param ptop - pointer to parallel topology
   
   JK, 4.1.2006
*/
void parcongrad::coarse_problem_ordering (long *domproc,gtopology *top,partop *ptop,FILE *out)
{
  long i,j,k,m,tnbn,nid,ndofn,cn;
  long *buff,*nbnd,*ndofncn;
  long **cnbn,**nbdofnd,**cncn;
  long ***lbcn;
  MPI_Status stat;
  
  if (myrank==0){


    //  array of numbers of boundary nodes on subdomains
    nbnd = ptop->nbnd;
    //  array containing coarse node numbers of boundary nodes
    cnbn = ptop->cnbn;
    //  array containing numbers of DOFs on boundary nodes
    nbdofnd = ptop->nbdofnd;
    //  total number of boundary nodes
    tnbn = ptop->tnbn;
    //  array containing local boundary code numbers
    lbcn = ptop->lbcn;
    
    if (nbnd==NULL){
      fprintf (stderr,"\n\n array containing local boundary code numbers (lbcn) is not allocated");
      fprintf (stderr,"\n in function coarse_problem_ordering (file %s, line %d)",__FILE__,__LINE__);
      fprintf (stderr,"\n function code_numbers_on_master (class partop) should be called first,");
    }
    
    //  number of DOFs at coarse nodes
    ndofncn = new long [tnbn];
    
    for (i=0;i<nproc;i++){
      for (j=0;j<nbnd[i];j++){
	k=cnbn[i][j];
	ndofncn[k]=nbdofnd[i][j];
      }
    }
    

    //  assignment of constraint indicators
    cncn = new long* [tnbn];
    for (i=0;i<tnbn;i++){
      cncn[i] = new long [ndofncn[i]];
    }


    //  number of boundary DOFs on subdomains
    //  it must not be deleted, it is used in further subroutines
    nbdofdom = new long [nproc];
    
    maxnbdof=0;
    for (i=0;i<nproc;i++){
      nbdofdom[i]=0;
      for (j=0;j<nbnd[i];j++){
	m=cnbn[i][j];
	for (k=0;k<nbdofnd[i][j];k++){
	  cncn[m][k]=lbcn[i][j][k];
	  if (lbcn[i][j][k]>0)
	    nbdofdom[i]++;
	}
      }
      if (maxnbdof<nbdofdom[i])
	maxnbdof=nbdofdom[i];
    }
    

    //  generation of code numbers (unknown ordering)
    ndofcp=1;
    for (i=0;i<tnbn;i++){
      for (j=0;j<ndofncn[i];j++){
	if (cncn[i][j]>0){
	  cncn[i][j]=ndofcp;
	  ndofcp++;
	}
      }
    }
    ndofcp--;
    
    
    bcn = new long* [nproc];
    for (i=0;i<nproc;i++){
      bcn[i] = new long [nbdofdom[i]];
      nbdofdom[i]=0;
    }
    
    //  assembling of lists of boundary code numbers on subdomains
    for (i=0;i<nproc;i++){
      for (j=0;j<nbnd[i];j++){
	m=cnbn[i][j];
	for (k=0;k<ndofncn[m];k++){
	  if (cncn[m][k]>0){
	    bcn[i][nbdofdom[i]]=cncn[m][k];
	    nbdofdom[i]++;
	  }
	}
      }
    }
    
    fprintf (out,"\n\n kontrolni tisk\n");
    for (i=0;i<nproc;i++){
      fprintf (out,"\n domena cislo %ld\n",i+1);
      for (j=0;j<nbdofdom[i];j++){
	fprintf (out,"  %ld",bcn[i][j]);
      }
    }
    
    fprintf (out,"\n\n kontrola lbcn");
    for (i=0;i<nproc;i++){
      fprintf (out,"\n");
      for (j=0;j<nbnd[i];j++){
	fprintf (out,"   ");
	for (k=0;k<nbdofnd[i][j];k++){
	  fprintf (out," %ld",lbcn[i][j][k]);
	}
      }
    }
    
    for (i=0;i<tnbn;i++){
      delete [] cncn[i];
    }
    delete [] cncn;
    
    delete [] ndofncn;

  }

  if (myrank==0){
    for (i=1;i<nproc;i++){
      MPI_Send (&maxnbdof,1,MPI_LONG,i,myrank,MPI_COMM_WORLD);
    }
  }
  else{
    MPI_Recv (&maxnbdof,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  

  nbn = ptop->nbn;

  buff = new long [ndof];
  for (i=0;i<ndof;i++){
    buff[i]=0;
  }
  
  nbdof=0;
  for (i=0;i<nbn;i++){
    nid=ptop->lnbn[i];
    ndofn=top->give_ndofn (nid);
    for (j=0;j<ndofn;j++){
      cn = top->give_dof (nid,j)-1;
      if (cn>-1){
	buff[cn]=1;
	nbdof++;
      }
    }
  }
  
  nidof=ndof-nbdof;

  lidof = new long [nidof];
  lbdof = new long [nbdof];

  j=0;  k=0;
  for (i=0;i<ndof;i++){
    if (buff[i]==0){  lidof[j]=i;  j++;  }
    if (buff[i]==1){  lbdof[k]=i;  k++; }
  }
  
  fprintf (out,"\n\n kontrola vnitrnich DOF\n");
  for (i=0;i<nidof;i++){
    fprintf (out," %ld",lidof[i]);
  }
  fprintf (out,"\n\n kontrola hranicnich DOF\n");
  for (i=0;i<nbdof;i++){
    fprintf (out," %ld",lbdof[i]);
  }
  
  
}

/**
   function extracts boundary DOFs from local %vector to buffer
   
   @param lv - local %vector
   @param buff - buffer
   
   JK, 4.1.2006
*/
void parcongrad::local_buff (double *lv,double *buff)
{
  long i,j;
  
  for (i=0;i<nbdof;i++){
    j=lbdof[i]-1;
    buff[j]=lv[j];
  }
}

/**
   function scaters boundary DOFs from buffer to coarse %vector
   
   @param lv - local %vector
   @param buff - buffer
   @param nd - number of subdomain
   
   JK, 4.1.2006
*/
void parcongrad::buff_coarse (double *cv,double *buff,long nd)
{
  long i,j;
  
  //for (i=0;i<nbdofdom[nd];i++){
  for (i=0;i<ndofdom[nd];i++){
    j=bcn[nd][i]-1;
    cv[j]+=buff[i];
  }
}

/**
   function scaters boundary DOFs from buffer to coarse %vector
   
   @param lv - local %vector
   @param buff - buffer
   @param nd - number of subdomain
   
   JK, 4.1.2006
*/
void parcongrad::coarse_buff (double *cv,double *buff,long nd)
{
  long i,j;
  
  //for (i=0;i<nbdofdom[nd];i++){
  for (i=0;i<ndofdom[nd];i++){
    j=bcn[nd][i]-1;
    buff[i]=cv[j];
  }
}

/**
   function scaters boundary DOFs from buffer to local %vector
   
   @param lv - local %vector
   @param buff - buffer
   
   JK, 4.1.2006
*/
void parcongrad::buff_local (double *lv,double *buff)
{
//  long i,j;
  
  //for (i=0;i<nbdofdom[nd];i++){
  //  for (i=0;i<ndofdom[nd];i++){
  //  j=bcn[nd][i]-1;
  //lv[j]=buff[i];
  //}
}


double parcongrad::mod_scal_prod (double *u,double *v)
{
  long i,j;
  double ss;
  
  ss=0.0;
  for (i=0;i<nidof;i++){
    j=lidof[i];
    ss+=v[j]*u[j];
  }
  return ss;
}


/**
   function solves system of algebraic equations by parallel
   conjugate gradient method
   
   @param top - topology
   @param gm - %matrix of the system
   @param domproc - domain-processor correspondence
   @param lhs - array containing solution of the system
   @param rhs - array containing right hand side
   @param out - output stream
   @param iv - initial values indicator
   
   iv=0 - initial vector is zero %vector
   iv=1 - initial vector is taken from x array
   
   JK, 4.1.2006
*/
void parcongrad::solve_system (partop *ptop,gtopology *top,gmatrix *gm,
			       long *domproc,double *lhs,double *rhs,FILE *out,long iv)
{
  long i,j,k,buffsize;
  long *dofmultip;
  double nom,denom,alpha,beta;
  double *d,*p,*r,*dd,*dp,*dr,*dx,*buff,*master_rhs;
  MPI_Status stat;

  buffsize=maxndof+1;
  
  //  auxiliary vector
  p = new double [ndof];
  //  residuum vector
  r = new double [ndof];
  //  direction vector
  d = new double [ndof];
  //  buffer
  buff = new double [buffsize];
  
  dofmultip = ptop->dofmultip;
  
  if (myrank==0){
    for (i=1;i<nproc;i++){
      MPI_Send (&ni,1,MPI_LONG,i,myrank,MPI_COMM_WORLD); 
    }
  }
  else{
    MPI_Recv (&ni,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
  }
  MPI_Barrier (MPI_COMM_WORLD);	 

  
  
  /* pomocne shromazdeni vektoru zatizeni */
  
  // rhs 
  // fprintf (out,"\n\n\n\n Kontrola prave strany na domene\n\n");
  for(i = 0; i < ndof; i++){
    
    rhs[i]/=double(dofmultip[i]);
    buff[i] = rhs[i];
    fprintf (out,"%ld   %ld   %le\n",i,dofmultip[i],rhs[i]); 
  }
  
  
  
  
  if(myrank == 0){
    tndof=ptop->tndof;
    bcn = ptop->gcnd;
    ndofdom = ptop->nud;
    master_rhs = new double[tndof];
    nullv (master_rhs,tndof);
    //  master contribution
    k=domproc[0];
    buff_coarse (master_rhs,buff,k);
    //  slave contributions
    for (j=1;j<nproc;j++){
      MPI_Recv (buff,buffsize,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      k=domproc[stat.MPI_TAG];
      buff_coarse (master_rhs,buff,k);
    }
    
    fprintf (out,"\n\n\n\n Kontrola prave strany na masteru\n\n");
    double norm=0.0;
    for(i = 0; i < tndof; i++){
      norm+=master_rhs[i];
    }
    for(i = 0; i < tndof; i++){
      fprintf (out,"%ld    %le\n",i,master_rhs[i]);
    }
    fprintf (stdout,"\n\n velikost rhs je %le",norm);

    delete []master_rhs;
  }
  else{
    MPI_Send (buff,buffsize,MPI_DOUBLE,0,myrank,MPI_COMM_WORLD); 
  }
  MPI_Barrier (MPI_COMM_WORLD);	 
  
  /* konec vektoru zatizeni  */




  //  initial vector
  if (iv==0){
    for (i=0;i<ndof;i++){
      lhs[i]=0.0;
    }
  }
  
  //  matrix multiplied by initial vector
  gm->gmxv (lhs,p);
  
  //  initial residuum
  subv (rhs,p,r,ndof);
  
  nullv (buff,buffsize);
  copyv (r,buff,ndof);
  
  if (myrank==0){
    tndof=ptop->tndof;
    bcn = ptop->gcnd;
    ndofdom = ptop->nud;
    

    //  vector of coarse residuum
    dr = new double [tndof];
    nullv (dr,tndof);
    //  coarse direction vector
    dd = new double [tndof];
    nullv (dd,tndof);
    //  vector of unknowns
    dx = new double [tndof];
    nullv (dx,tndof);
    //  auxiliary vector
    dp = new double [tndof];
    nullv (dp,tndof);
    
    //  master contribution
    k=domproc[0];
    buff_coarse (dr,buff,k);
    
    //  slave contributions
    for (i=1;i<nproc;i++){
      MPI_Recv (buff,buffsize,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      k=domproc[stat.MPI_TAG];
      buff_coarse (dr,buff,k);
    }
    
    nom = ss (dr,dr,tndof);
    
    copyv (dr,dd,tndof);
    
    
    for (i=1;i<nproc;i++){
      k=domproc[i];
      nullv (buff,buffsize);
      coarse_buff (dd,buff,k);
      MPI_Send (buff,buffsize,MPI_DOUBLE,i,myrank,MPI_COMM_WORLD); 
    }
    
    k=domproc[0];
    nullv (buff,buffsize);
    coarse_buff (dd,buff,k);
  }
  else{
    MPI_Send (buff,buffsize,MPI_DOUBLE,0,myrank,MPI_COMM_WORLD); 
    MPI_Recv (buff,buffsize,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
  }
  MPI_Barrier (MPI_COMM_WORLD);	 
  

  copyv (buff,d,ndof);
  
  //  iteration loop
  for (i=0;i<ni;i++){
    
    //  A s = p
    gm->gmxv (d,p);
    
    copyv (p,buff,ndof);
    

    if (myrank==0){
      nullv (dp,tndof);
      
      //  master contribution
      k=domproc[0];
      buff_coarse (dp,buff,k);
      
      //  slave contributions
      for (j=1;j<nproc;j++){
	MPI_Recv (buff,buffsize,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
	k=domproc[stat.MPI_TAG];
	buff_coarse (dp,buff,k);
      }
      
      denom = ss (dd,dp,tndof);
      
      alpha = nom/denom;
      

      //  new approximation of x and r
      for (j=0;j<tndof;j++){
	dx[j]+=alpha*dd[j];
	dr[j]-=alpha*dp[j];
      }
      
      denom=nom;
      
      nom = ss (dr,dr,tndof);
      
      fprintf (stdout,"\n iteration number   %ld    nom %le",i,nom);

      if (nom<err)
	buff[buffsize-1]=100.0;
      
      beta = nom/denom;

      //  new vector of direction
      for (j=0;j<tndof;j++){
	dd[j]=beta*dd[j]+dr[j];
      }
      

      for (j=1;j<nproc;j++){
	k=domproc[j];
	nullv (buff,buffsize-1);
	coarse_buff (dd,buff,k);
	MPI_Send (buff,buffsize,MPI_DOUBLE,j,myrank,MPI_COMM_WORLD); 
      }
      
      k=domproc[0];
      nullv (buff,buffsize-1);
      coarse_buff (dd,buff,k);
      
    }
    else{
      MPI_Send (buff,buffsize,MPI_DOUBLE,0,myrank,MPI_COMM_WORLD); 
      MPI_Recv (buff,buffsize,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    }
    MPI_Barrier (MPI_COMM_WORLD);	 
    
    
    copyv (buff,d,ndof);
    
    if (buff[buffsize-1]>1.0){
      
      if (myrank==0){
	for (j=1;j<nproc;j++){
	  k=domproc[j];
	  nullv (buff,buffsize-1);
	  coarse_buff (dx,buff,k);
	  MPI_Send (buff,buffsize,MPI_DOUBLE,j,myrank,MPI_COMM_WORLD); 
	}
	k=domproc[0];
	nullv (buff,buffsize-1);
	coarse_buff (dx,buff,k);
      }
      else{
	MPI_Recv (buff,buffsize,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      }
      
      copyv (buff,lhs,ndof);
      
      break;
    }
  }
  MPI_Barrier (MPI_COMM_WORLD);	 
  
  
  
//  time_t t1,t2,t3,t4;
  
  
  
  
    



  if (myrank==0){
    
    fprintf (out,"\n\n\n\n\n");
    j=domproc[myrank];
    fprintf (out,"\n\n\n Domain %ld",j);
    fprintf (out,"\n time of condensation             %g",buff[0]);
    fprintf (out,"\n time of solution of red. sys.    %g",buff[1]);
    fprintf (out,"\n time of back substitution        %g",buff[2]);
    
    j=domproc[myrank];
    fprintf (stdout,"\n\n\n Domain %ld",j);
    fprintf (stdout,"\n time of condensation             %g",buff[0]);
    fprintf (stdout,"\n time of solution of red. sys.    %g",buff[1]);
    fprintf (stdout,"\n time of back substitution        %g",buff[2]);

    for (i=1;i<nproc;i++){
      MPI_Recv (buff,3,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      j=domproc[stat.MPI_TAG];
      fprintf (out,"\n\n\n Domain %ld",j);
      fprintf (out,"\n time of condensation             %g",buff[0]);
      fprintf (out,"\n time of solution of red. sys.    %g",buff[1]);
      fprintf (out,"\n time of back substitution        %g",buff[2]);

      fprintf (stdout,"\n\n\n Domain %ld",j);
      fprintf (stdout,"\n time of condensation             %g",buff[0]);
      fprintf (stdout,"\n time of solution of red. sys.    %g",buff[1]);
      fprintf (stdout,"\n time of back substitution        %g",buff[2]);
    }
  }
  else{
    MPI_Send(buff,3,MPI_LONG,0,myrank,MPI_COMM_WORLD);
  }
  MPI_Barrier (MPI_COMM_WORLD);

  


}
