#include "partop.h"
#include <string.h>
#include <math.h>
#include <time.h>
#include <float.h>
#include <limits.h>
#include <stdlib.h>
#include "../GEFEL/galias.h"
#include "../GEFEL/intp.h"

/**
   constructor
   
   @param np - number of processors
   @param mr - my rank
   @param nd - number of subdomain
   @param meshd - type of mesh description
   @param *nameproc - name of processor
   @param nl - length of processor name  
   JK, 3.5.2004
*/
partop::partop(int np,int mr,long nd,meshdescription meshd,char *nameproc, int nl)
{
  //  number of processors
  nproc=np;
  //  my rank
  myrank=mr;
  //  mesh description
  md = meshd;

  //  number of nodes on subdomain
  nn=0;
  //  the number of DOFs on subdomain
  ndof=0;
  //  the number of coupled DOFs
  ncdof=0;
  //  the number of boundary/interface coupled DOFs
  nbcdof=0;

  //  the maximum number of nodes on one subdomain
  maxnn=0;
  
  //  the number of internal nodes
  nin=0;
  //  the number of boundary/interface nodes on subdomain
  nbn=0;
  //  the maximum number of boundary nodes on subdomain
  maxnbn = 0;
  //  the total number of all boundary/interface nodes in the problem
  tnbn=0;
  //  the total number of nodes in the whole problem
  tnnp=0;
  
  //  maximum number of all DOFs on one subdomain
  maxndof=0;


  
  //  array containing nodal multiplicity
  nodmultip = NULL;
  //  array containing coarse numbers of interface/boundary nodes on one subdomain
  icnbn = NULL;
  //  global numbers of boundary/interface nodes on subdomains
  gnbndom = NULL;
  
  
  if (myrank==0){
    //  array containing the numbers of nodes on subdomains
    nnsd = NULL;
    //  array containing number of subdomains which each boundary/interface node belongs to
    bmultip = NULL;
    //  array containing number of subdomains which each node belongs to
    amultip = NULL;
    //  array containing the numbers of boundary/interface nodes on subdomains
    nbnd = NULL;
    //  array containing the numbers of internal nodes on subdomains
    nind = NULL;
    //  array containing global node numbers
    allnodes = NULL;
    //  array containing coarse node numbers
    bnodes = NULL;
    //  array containing indicators of coupled DOFs
    coupdof = NULL;
    //  array containing suspicious indicators of coupled DOFs
    coupdofmas = NULL;
    //  array containing numbers of all degrees of freedom on subdomains
    nalldof = NULL;
    //  array containing the numbers of DOF on nodes
    ndofnmas = NULL;
    //  array containing DOF indicators
    dofindmas = NULL;
    //  array containing DOF indicators
    dofind = NULL;
    //  array containing coarse numbers of interface/boundary nodes
    icnbnmas = NULL;
    //  array containing node multiplicity of boundary/interface nodes
    icmultip = NULL;
    //  global numbers of boundary/interface nodes on the master
    gnbn = NULL;
    //  global numbers of internal nodes on the master
    gnin = NULL;
    //  global numbers of boundary/interface nodes appropriate to coarse node
    gnbncn = NULL;
    //  subdomain id of interface/boundary nodes appropriate to coarse node
    sid = NULL;
    
  }

  cornod = -1;



  // ****************************************************
  //  to co je nize neni zkontrolovano
  // ****************************************************


  //  number of subdomain
  ndom=nd;
  // name of processor
  strcpy (procName,nameproc);
  nameLength=nl;
  
  
  // number of elements on subdomain
  ne=0;

  //  local numbers of boundary nodes
  lnbn = NULL;
  //  local numbers of internal nodes
  lnin = NULL;
  //
  lgnbn = NULL;
  // aray containing DOF multiplicity
  dofmultip = NULL;
  // array containing identification of nodes - vertex, edge, internal
  nodeidentif=NULL;
  // array containing local number of vertex node
  // vertices[i]=j i-th vertex node has local number j
  vertices = NULL;
  // number of vertex (corner) nodes on one subdomain
  nv = 0;
  // spatial dimension of mesh
  dim = -1;
  
  /*
  nboundnodes = 0;
  boundnodes = NULL;
  nadjacboundnod = NULL;
  adjacboundnod = NULL;
  nbcurves = 0;

  bcurves = NULL;
  */

  fictdom = NULL;
  sizefictdom = -1.0;
  nmember = 0;
  condcorner = -1;
  typecondcorner = 0;
    
  if (myrank==0){
    multip = NULL;
    cnbn = NULL;
    nbdofnd = NULL;
    nbdofd = NULL;
    lbcn = NULL;
    bnmultip = NULL;
    llnbn = NULL;
    ldn = NULL;
    
    
    gnn = NULL;
    pgcn = NULL;
    gcnbn = NULL;
  
    tncen = 0;
    ncend = NULL;
    cncen = NULL;
  }
  
}

/**
   destructor
   
   JK, 3.5.2004
*/
partop::~partop()
{
  long i;

  delete [] nodmultip;
  delete [] lnbn;
  
  if (myrank==0){
    delete [] nnsd;
    delete [] nbnd;
    delete [] multip;
    delete [] nbdofd;
    delete [] bnmultip;
    
    
    for (i=0;i<nproc;i++){
      delete [] allnodes[i];
    }
    delete [] allnodes;
    
    for (i=0;i<nproc;i++){
      delete [] cnbn[i];
    }
    delete [] cnbn;
    
    for (i=0;i<nproc;i++){
      delete [] nbdofnd[i];
    }
    delete [] nbdofnd;
    
    for (i=0;i<tnbn;i++){
      delete [] llnbn[i];
    }
    delete [] llnbn;
    
    for (i=0;i<tnbn;i++){
      delete [] ldn[i];
    }
    delete [] ldn;
    
    //  missing deallocation of array lbcn
  }
}


/**
   function assignes node numbers to auxiliary indicators at nodes
   
   @param top - pointer to general topology
   @param ltg - array of local to global correcpondence
   
   JK, 1.7.2005
*/
void partop::initiation (gtopology *top,long *ltg)
{
  long i;
  
  //  number of nodes on subdomain
  nn = top->nn;
  //  number of elements on subdomain
  ne = top->ne;
  

  switch (md){
  case all_nodes:{
    for (i=0;i<nn;i++){
      top->gnodes[i].ai = ltg[i];
    }
    break;
  }
  case bound_nodes:{
    for (i=0;i<nn;i++){
      top->gnodes[i].ai = ltg[i];
    }
    break;
  }
  case neg_bound_nodes:{
    for (i=0;i<nn;i++){
      top->gnodes[i].ai = ltg[i];
    }
    break;
  }
  default:{
    par_print_err(myrank,"unknown type of mesh description", __FILE__, __LINE__, __func__);
  }
  }


}





/**
   function computes node multiplicity
   function is used only for mesh description = all_nodes or neg_bound_nodes
   
   the following functions have to be called before this function:
   void initiation (gtopology *top,long *ltg);
   void numbers_of_all_nodes_on_subdomains (gtopology *top,long *domproc,FILE *out);

   
   function establishes:
   maxnbn - maximum number of boundary nodes
   
   nbn - number of boundary nodes on each subdomain
   
   tnbn - total number of boundary nodes
   
   nbnd - array containing numbers of boundary nodes on subdomains
   
   tnnp - total number of nodes in problem
   
   multip - array of boundary node multiplicity (on the master)
   
   nodmultip - array of node multiplicity (on all processors)
   
   
   @param top - pointer to subdomain topology
   @param domproc - array containing domain-processor correspondence
   
   JK, 30.6.2005
*/
void partop::compute_multiplicity (long *ltg,long *domproc,FILE *out)
{
  long i,j,k,buffsize;
  long *buff;
  MPI_Status stat;
  
  if (maxnn==0){
    fprintf (stderr,"\n\n maximum number of nodes on one subdomain is equal to zero,\n");
    fprintf (stderr,"\n function numbers_of_all_nodes_on_subdomains has to be called first,\n");
    abort ();
  }
  
  switch (md){
  case bound_nodes:{
    
    buffsize=2;
    buff = new long [buffsize];
    
    //  determination of total number of boundary nodes
    nbn=0;
    maxnbn = 0;
    tnbn=0;
    for (i=0;i<nn;i++){
      if (tnbn<ltg[i])
        tnbn=ltg[i];
      if (ltg[i]>-1)
        nbn++;
    }
    
    buff[0]=tnbn;
    buff[1]=nbn;
    
    if (myrank==0){
      //  list of number of boundary nodes on subdomains
      if (nbnd != NULL)
        delete [] nbnd;
      nbnd = new long [nproc];
      
      //  master contribution
      j=domproc[0];
      if (tnbn<buff[0])  tnbn=buff[0];
      if (maxnbn<buff[1])  maxnbn=buff[1];
      nbnd[j]=buff[1];
      
      for (i=1;i<nproc;i++){
        MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
        j=domproc[stat.MPI_TAG];
        //  slave contributions
        if (tnbn<buff[0])
          tnbn=buff[0];
        if (maxnbn<buff[1])
          maxnbn=buff[1];
        nbnd[j]=buff[1];
      }
      tnbn++;
      
      buff[0]=tnbn;
      buff[1]=maxnbn;
      
      for (i=1;i<nproc;i++){
        MPI_Send (buff,buffsize,MPI_LONG,i,myrank,MPI_COMM_WORLD);
      }
    }
    else{
      MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    }
    MPI_Barrier (MPI_COMM_WORLD);
    
    tnbn=buff[0];
    maxnbn=buff[1];
    
    delete [] buff;
    
    buffsize = maxnbn;
    buff = new long [buffsize];
    for (i=0;i<maxnbn;i++){
      buff[i]=-1;
    }
    
    j=0;
    for(i=0;i<nn;i++){
      if (ltg[i]>-1){
        buff[j] = ltg[i];
        j++;
      }
    }
    
    
    if(myrank == 0){
      if (multip!=NULL){
        delete [] multip;
      }
      multip = new long [tnbn];
      for(i=0;i<tnbn;i++){
        multip[i]=0;
      }
      
      // master contribution
      for(j=0;j<maxnbn;j++){
        if (buff[j]>-1){
          multip[buff[j]]++;
        }
      }
      
      //  slave contributions
      for (i=1;i<nproc;i++){
        MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
        for(j=0;j<maxnbn;j++){
          if (buff[j]>-1){
            multip[buff[j]]++;
          }
        }
      }
      
    }
    else{
      MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
    }
    MPI_Barrier (MPI_COMM_WORLD);
    delete [] buff;
    
    buffsize=tnbn;
    buff = new long [buffsize];
    
    if (myrank==0){
      for (i=0;i<tnbn;i++){
        buff[i]=multip[i];
      }
      
      for (i=1;i<nproc;i++){
        MPI_Send (buff,buffsize,MPI_LONG,i,myrank,MPI_COMM_WORLD);
      }
    }
    else{
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    }
    MPI_Barrier (MPI_COMM_WORLD);
    
    if (nodmultip!=NULL)
      delete [] nodmultip;
    nodmultip = new long [nn];
    
    for (i=0;i<nn;i++){
      if (ltg[i]>-1){
        nodmultip[i]=buff[ltg[i]];
      }
      else{
        nodmultip[i]=1;
      }
    }
    
    delete [] buff;
       
    break;
  }




  case all_nodes:
  case neg_bound_nodes:{
    
    buffsize=maxnn+1;
    buff = new long [buffsize];
    tnnp=0;
    for (i=0;i<nn;i++){
      buff[i]=ltg[i];
      if (tnnp<ltg[i])
        tnnp=ltg[i];
    }
    buff[maxnn]=tnnp;
    
    // *********
    //  master
    // *********
    if (myrank==0){
      if (allnodes != NULL){
        for (i=0;i<nproc;i++){
          delete [] allnodes[i];
        }
        delete [] allnodes;
      }
      allnodes = new long* [nproc];
      for (i=0;i<nproc;i++){
        allnodes[i] = new long [nnsd[i]];
      }
      
      
      k=domproc[0];
      for (j=0;j<nnsd[k];j++){
        allnodes[k][j]=buff[j];
      }
      if (tnnp<buff[maxnn])
        tnnp=buff[maxnn];
      

      for (i=1;i<nproc;i++){
        MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
        
        k=domproc[stat.MPI_TAG];
        for (j=0;j<nnsd[k];j++){
          allnodes[k][j]=buff[j];
        }
        if (tnnp<buff[maxnn])
          tnnp=buff[maxnn];
      }
    }
    
    // ********
    // slaves
    // ********
    else{
      MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
    }
    
    if (myrank==0){
      //  array containing number of subdomains which share node
      //  this is called node multiplicity
      
      //  must be increased due to indices from 0 instead of 1
      tnnp++;
      
      fprintf (out,"\n\n\n total number of nodes on whole problem %ld\n",tnnp);
      if (multip != NULL)
        delete [] multip;
      multip = new long [tnnp];
      for (i=0;i<tnnp;i++){
        multip[i]=0;
      }
      
      //  computation of node incidences
      for (i=0;i<nproc;i++){
        for (j=0;j<nnsd[i];j++){
          multip[allnodes[i][j]]++;
        }
      }
    }
    
    
    if (myrank==0){
      for (i=1;i<nproc;i++){
        k=domproc[i];
        for (j=0;j<nnsd[k];j++){
          buff[j]=multip[allnodes[k][j]];
        }
        MPI_Send (buff,buffsize,MPI_LONG,i,myrank,MPI_COMM_WORLD);
      }
      
      k=domproc[0];
      for (j=0;j<nnsd[k];j++){
        buff[j]=multip[allnodes[k][j]];
      }
    }
    else{
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    }
    
    //  nodal multiplicity
    if (nodmultip != NULL)
      delete [] nodmultip;
    nodmultip = new long [nn];
    for (i=0;i<nn;i++){
      nodmultip[i] = buff[i];
    }
    
    delete [] buff;
    
    if (myrank==0){
      //  number of all boundary nodes
      tnbn=0;
      for (i=0;i<tnnp;i++){
        if (multip[i]!=1)  tnbn++;
      }
      
      //  list of number of boundary nodes on subdomains
      if (nbnd != NULL)
        delete [] nbnd;
      nbnd = new long [nproc];
      
      maxnbn=0;
      for (i=0;i<nproc;i++){
        k=0;
        for (j=0;j<nnsd[i];j++){
          if (multip[allnodes[i][j]]>1)  k++;
        }
        if (maxnbn<k)  maxnbn=k;
        nbnd[i]=k;
      }
      
      for (i=1;i<nproc;i++){
        MPI_Send (&maxnbn,1,MPI_LONG,i,myrank,MPI_COMM_WORLD);
      }
    }
    else{
      MPI_Recv (&maxnbn,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    }
    
    
    //  number of boundary nodes on subdomain
    nbn=0;
    for (i=0;i<nn;i++){
      if (nodmultip[i]>1)
        nbn++;
    }

    break;
  }
  default:{
    fprintf (stderr,"\n\n unknown type of mesh description is required in");
    fprintf (stderr,"\n function compute_multiplicity (file %s, line %d)\n",__FILE__,__LINE__);
  }
  }
  MPI_Barrier (MPI_COMM_WORLD);
  
  
  fprintf(out,"\n\n\n Multiplicity of nodes on subdomain\n\n");  
  for (i=0;i<nn;i++)
    fprintf(out,"%ld   %ld\n",i,nodmultip[i]);  
  

  fprintf(out,"\n\n\n total number of boundary nodes in whole problem is %ld\n\n",tnbn);
  fprintf (out,"\n\n\n maximum number of boundary nodes on subdomain is %ld\n\n",maxnbn);
  if (myrank==0){
    fprintf (out,"\n\n\n numbers of boundary nodes on each subdomain \n\n");
    for (i=0;i<nproc;i++){
      fprintf (out,"\n %ld",nbnd[i]);
    }
  }  
  fprintf(out,"\n\n\n total number of boundary nodes in whole problem is %ld\n\n",tnbn);
  fprintf (out,"\n\n\n maximum number of boundary nodes on subdomain is %ld\n\n",maxnbn);
  if (myrank==0){
    fprintf (out,"\n\n\n numbers of boundary nodes on each subdomain \n\n");
    for (i=0;i<nproc;i++){
      fprintf (out,"\n %ld",nbnd[i]);
    }
  }  
  
}

/**
   function assembles multiplicity of degrees of freedom
   it is different from the function compute_multiplicity, which
   computes multiplicity of nodes
   
   JK, 14.6.2007
*/
void partop::dof_multiplicity (gtopology *top,FILE *out)
{
  long i,j,k,nm;
  long ndofn;
  
  ndof = 0;
  for (i=0;i<nn;i++){
    ndofn=top->give_ndofn (i);
    for (j=0;j<ndofn;j++){
      k=top->give_dof (i,j);
      if (k>0){
        ndof++;
      }
    }
  }
  
  //printf (out,"\n\n ndof   %ld\n",ndof);
  
  
  dofmultip = new long [ndof];
  
  for (i=0;i<nn;i++){
    ndofn=top->give_ndofn (i);
    nm=nodmultip[i];
    for (j=0;j<ndofn;j++){
      k=top->give_dof (i,j);
      if (k>0){
        dofmultip[k-1]=nm;
      }
    }
  }
  /*
  fprintf (out,"\n\n\n kontrola dofmultip \n");
  for (i=0;i<ndof;i++){
    fprintf (out,"%ld   %ld\n",i+1,dofmultip[i]);
  }
  */

}


/**
   function selects boundary nodes
   
   the following functions have to be called before this function:
   void initiation (gtopology *top,long *ltg);
   void numbers_of_all_nodes_on_subdomains (gtopology *top,long *domproc,FILE *out);
   void compute_multiplicity (gtopology *top,long *domproc,FILE *out);

   
   function establishes:
   
   lgnbn - global numbers of boundary nodes
   lnbn - local numbers of boundary nodes
   nin - number of interanl nodes
   lnin - local numbers of internal nodes
   
   @param ltg - local to global correspondence
   @param domproc - array containing domain-processor correspondence
   
   JK, 30.6.2005
   
*/
void partop::find_boundary_nodes (long *ltg,long *domproc,FILE *out)
{
  long i,j,k,l,buffsize;
  long *buff;
  MPI_Status stat;
  
  
  // *******************************************************
  //  searching for number of boundary nodes on subdomains
  // *******************************************************
  /*  
  switch (md){
  case bound_nodes:{
    
    //  number of boundary nodes on subdomain
    nbn=0;  maxnbn=0;
    for (i=0;i<nn;i++){
      if (ltg[i]>-1)  nbn++;
      if (maxnbn<ltg[i])  maxnbn=ltg[i];
    }
    
    buffsize=2;
    buff = new long [buffsize];
    buff[0]=nbn;
    buff[1]=maxnbn;
    
    
    // *********
    //  master
    // *********
    if (myrank==0){
      //  list of number of boundary nodes on subdomains
      if (nbnd != NULL)
        delete [] nbnd;
      nbnd = new long [nproc];
      
      //  master contribution
      j=domproc[0];
      if (maxnbn<buff[0])  maxnbn=buff[0];
      if (tnbn<buff[1])  tnbn=buff[1];
      nbnd[j]=buff[0];
      
      for (i=1;i<nproc;i++){
        MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
        
        //  slave contributions
        j=domproc[stat.MPI_TAG];
        if (maxnbn<buff[0])  maxnbn=buff[0];
        if (tnbn<buff[1])  tnbn=buff[1];
        nbnd[j]=buff[0];
        
      }
      
      //  node numbers start from 0 with respect to C language notation
      tnbn++;
      
      for (i=1;i<nproc;i++){
        MPI_Send (&maxnbn,1,MPI_LONG,i,myrank,MPI_COMM_WORLD);
      }
    }
    // ********
    // slaves
    // ********
    else{
      MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
      MPI_Recv (&maxnbn,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    }
    
    delete [] buff;
    fprintf (out,"\n\n maxnbn is %ld",maxnbn);
    break;
  }

  case all_nodes:
  case neg_bound_nodes:{
    
    if (myrank==0){
      if (multip==NULL){
        fprintf (stderr,"\n\n array multip is not allocated in function find_boundary_nodes,\n");
        fprintf (stderr,"\n function compute_multiplicity has to be called first (file %s, line %d),\n",__FILE__,__LINE__);
      }
    }
    
    if (myrank==0){
      //  number of all boundary nodes
      tnbn=0;
      for (i=0;i<tnnp;i++){
        if (multip[i]!=1)  tnbn++;
      }
      
      //  list of number of boundary nodes on subdomains
      if (nbnd != NULL)
        delete [] nbnd;
      nbnd = new long [nproc];

      maxnbn=0;
      for (i=0;i<nproc;i++){
        k=0;
        for (j=0;j<nnsd[i];j++){
          if (multip[allnodes[i][j]]>1)  k++;
        }
        if (maxnbn<k)  maxnbn=k;
        nbnd[i]=k;
      }
      
      for (i=1;i<nproc;i++){
        MPI_Send (&maxnbn,1,MPI_LONG,i,myrank,MPI_COMM_WORLD);
      }
    }
    else{
      MPI_Recv (&maxnbn,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    }
    
    
    //  number of boundary nodes on subdomain
    nbn=0;
    for (i=0;i<nn;i++){
      if (nodmultip[i]>1)
        nbn++;
    }

    break;
  }
  default:{
    fprintf (stderr,"\n\n unknown type of mesh description is required in");
    fprintf (stderr,"\n function find_boundary_nodes (file %s, line %d)\n",__FILE__,__LINE__);
  }
  }
  */
  
  
  buffsize=maxnbn;
  buff = new long [buffsize];
  
  switch (md){
  case bound_nodes:{

    //  list of coarse numbers of boundary nodes
    if (lgnbn!=NULL)
      delete [] lgnbn;
    lgnbn = new long [nbn];

    j=0;
    for (i=0;i<nn;i++){
      if (ltg[i]>-1){
        lgnbn[j]=ltg[i];
        j++;
      }
    }
    break;
  }
  case all_nodes:
  case neg_bound_nodes:{
    
    if (myrank==0){
      for (i=1;i<nproc;i++){
        k=domproc[i];
        l=0;
        for (j=0;j<nnsd[k];j++){
          if (multip[allnodes[k][j]]>1){
            buff[l]=allnodes[k][j];
            l++;
          }
        }
        
        MPI_Send (buff,buffsize,MPI_LONG,i,myrank,MPI_COMM_WORLD);
      }
      
      k=domproc[0];
      l=0;
      for (j=0;j<nnsd[k];j++){
        if (multip[allnodes[k][j]]>1){
          buff[l]=allnodes[k][j];
          l++;
        }
      }
      
    }
    else{
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    }

    //  list of coarse numbers of boundary nodes
    if (lgnbn!=NULL)
      delete [] lgnbn;
    lgnbn = new long [nbn];
    
    for (i=0;i<nbn;i++){
      lgnbn[i]=buff[i];
    }
    
    break;
  }
  default:{
    fprintf (stderr,"\n\n unknown type of mesh description is required in");
    fprintf (stderr,"\n function find_boundary_nodes (file %s, line %d)\n",__FILE__,__LINE__);
  }
  }
  
  
  delete [] buff;
  
  
  // *******************************
  //  indication of boundary nodes 
  // *******************************
  //  local numbers of boundary nodes
  if (lnbn != NULL)
    delete [] lnbn;
  lnbn = new long [nbn];
  
  switch (md){
  case all_nodes:
  case neg_bound_nodes:{
    j=0;
    for (i=0;i<nn;i++){
      if (nodmultip[i]>1){
        lnbn[j]=i;
        j++;
      }
    }
    
    break;
  }
    
  case bound_nodes:{
    j=0;
    for (i=0;i<nn;i++){
      if (ltg[i]>-1){
        lnbn[j]=i;
        j++;
      }
    }
    break;
  }
  default:{
    fprintf (stderr,"\n\n unknown type of mesh description is required in");
    fprintf (stderr,"\n function find_boundary_nodes (file %s, line %d)\n",__FILE__,__LINE__);
  }
  }
  

  //  number of internal nodes
  nin = nn-nbn;
  
  //  array containing local numbers of internal nodes
  if (lnin != NULL)
    delete [] lnin;
  lnin = new long [nin];
  
  buff = new long [nn];
  for (i=0;i<nn;i++){
    buff[i]=0;
  }
  
  for (i=0;i<nbn;i++){
    j=lnbn[i];
    buff[j]=1;
  }
  
  j=0;
  for (i=0;i<nn;i++){
    if (buff[i]==0){
      lnin[j]=i;
      j++;
    }
  }

  delete [] buff;

  MPI_Barrier (MPI_COMM_WORLD);
  

  if (myrank==0){
    fprintf (out,"\n\n\n numbers of boundary nodes on each subdomain \n\n");
    for (i=0;i<nproc;i++){
      fprintf (out,"\n %ld",nbnd[i]);
    }
  }
  
  fprintf (out,"\n\n\n numbers of boundary nodes on subdomain \n\n");
  fprintf (out,"\n %ld",nbn);
  fprintf (out,"\n\n\n local numbers of boundary nodes on subdomain");
  for (i=0;i<nbn;i++){
    fprintf (out,"\n %ld   %ld",i,lnbn[i]);
  }

  fprintf (out,"\n\n\n numbers of internal nodes on subdomain \n\n");
  fprintf (out,"\n %ld",nin);
  fprintf (out,"\n\n\n local numbers of internal nodes on subdomain");
  for (i=0;i<nin;i++){
    fprintf (out,"\n %ld   %ld",i,lnin[i]);
  }
  

}

/**
   function rewrites array ltg
   
   JK, 8.8.2007
*/
void partop::rewrite_ltg (long *ltg)
{
  long i;
  
  for (i=0;i<nn;i++){
    ltg[i]=-1;
  }
  for (i=0;i<nbn;i++){
    ltg[lnbn[i]]=lgnbn[i];
  }
}


/**
   function assembles boundary nodes on the master processor
   
   the following functions have to be called before this function:
   void initiation (gtopology *top,long *ltg);
   void numbers_of_all_nodes_on_subdomains (gtopology *top,long *domproc,FILE *out);
   void compute_multiplicity (gtopology *top,long *domproc,FILE *out);
   void find_boundary_nodes (gtopology *top,long *domproc,FILE *out);
   
   
   function establishes:
   
   cnbn - array of coarse numbers of boundary nodes
   
   @param top - pointer to subdomain topology
   @param domproc - array containing domain-processor correspondence
   
   JK, 30.6.2005
*/
void partop::boundary_nodes_on_master (gtopology *top,long *domproc,FILE *out)
{
  long i,j,k,m,buffsize;
  long *buff;
  MPI_Status stat;
  
  if (myrank==0){
    if (nbnd==NULL){
      fprintf (stderr,"\n\n array containing numbers of boundary nodes on subdomains is not allocated,\n");
      fprintf (stderr,"\n function find_boundary_nodes has to be called first (file %s, line %d),\n",__FILE__,__LINE__);
    }
  }
  
  
  if (myrank==0){
    if (cnbn != NULL){
      for (i=0;i<nproc;i++){
        delete [] cnbn[i];
      }
      delete [] cnbn;
    }
    cnbn = new long* [nproc];
    for (i=0;i<nproc;i++){
      cnbn[i] = new long [nbnd[i]];
    }
  }
  
  switch (md){
  case all_nodes:
  case neg_bound_nodes:{
    
    if (myrank==0){
      if (gcnbn != NULL)
        delete [] gcnbn;
      gcnbn = new long [tnnp];

      j=0;
      for (i=0;i<tnnp;i++){
        if (multip[i]>1){
          gcnbn[i]=j;
          j++;
        }
        else{
          gcnbn[i]=-1;
        }
      }
      
      /*
      // kontrolni tisk pole gcnbn - pro allnodes je to co gnodes.ai pro bounadry nodes
      fprintf (out,"\n\n gcnbn\n");
      for (i=0;i<tnnp;i++){
        fprintf (out,"%ld  %ld\n",i+1,gcnbn[i]); 
      }
      */
      
      if (j!=tnbn){
        fprintf (stderr,"\n\n total number of boundary nodes computed in function boundary_nodes_on_master");
        fprintf (stderr,"\n differs from the number computed in function find_boundary_nodes (file %s, line %d),\n",__FILE__,__LINE__);
      }
      
      for (i=0;i<nproc;i++){
        m=0;
        for (j=0;j<nnsd[i];j++){
          k=allnodes[i][j];
          if (gcnbn[k]>-1){
            cnbn[i][m]=gcnbn[k];
            m++;
          }
        }
      }
      delete [] gcnbn;
      gcnbn = NULL;
    }
    
    break;
  }
  
  case bound_nodes:{
    
    if (maxnbn==0){
      fprintf (stderr,"\n\n maximum number of boundary nodes on one subdomain is equal to zero,");
      fprintf (stderr,"\n function find_boundary_nodes should be called first (file %s, line %d),\n",__FILE__,__LINE__);
    }
    
    buffsize=maxnbn;
    buff = new long [buffsize];
    
    
    j=0;
    for (i=0;i<nn;i++){
      if (top->gnodes[i].ai>-1){
        buff[j]=top->gnodes[i].ai;
        j++;
      }
    }
    
    if (myrank==0){
      
      //  master contribution
      k=domproc[0];
      for (j=0;j<nbnd[k];j++){
        cnbn[k][j]=buff[j];
      }
      
      for (i=1;i<nproc;i++){
        MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
        
        //  slave contributions
        k=domproc[stat.MPI_TAG];
        for (j=0;j<nbnd[k];j++){
          cnbn[k][j]=buff[j];
        }
        
      }
 
    }
    else{
      MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
    }
    
    delete [] buff;
    break;
  }
  default:{
    fprintf (stderr,"\n\n unknown type of mesh description is required in");
    fprintf (stderr,"\n function boundary_nodes_on_master (file %s, line %d)\n",__FILE__,__LINE__);
  }
  }

  
  MPI_Barrier (MPI_COMM_WORLD);
  
  /*
  if (myrank==0){
    fprintf (out,"\n\n\n\n coarse numbering of boundary nodes  - cnbn\n\n");
    for (i=0;i<nproc;i++){
      fprintf (out,"\n\n domain number %ld\n",i+1);
      for (j=0;j<nbnd[i];j++){
        fprintf (out,"%ld %ld\n",j,cnbn[i][j]);
      }
    }
  }
  fprintf(out,"\ntnbn je %ld\n\n",tnbn);
  */
}


/**
   function assembles correspondence among coarse nodes and local boundary nodes
   
   function assembles:
   
   llnbn - list of local numbers of boundary nodes belonging to coarse node
   ldn - list of subdomains which contain boundary nodes belonging to coarse node
   
   not checked

   JK, 6.7.2005
*/
void partop::coarse_local_nodes ()
{
  long i,j,k,m;
  
  if (myrank==0){
    //  multiplicity of boundary nodes
    bnmultip = new long [tnbn];
    for (i=0;i<tnbn;i++){
      bnmultip[i]=0;
    }
    
    if (md == all_nodes){
      j=0;
      for (i=0;i<tnnp;i++){
        if (multip[i]>1){
          bnmultip[j]=multip[i];
          j++;
        }
      }
    }
    
    if (md == bound_nodes){
      for (i=0;i<nproc;i++){
        for (j=0;j<nbnd[i];j++){
          bnmultip[cnbn[i][j]]++;
        }
      }
    }
    
    llnbn = new long* [tnbn];
    ldn = new long* [tnbn];
    for (i=0;i<tnbn;i++){
      llnbn[i] = new long [bnmultip[i]];
      ldn[i] = new long [bnmultip[i]];
      bnmultip[i]=0;
    }
    
    for (i=0;i<nproc;i++){
      for (j=0;j<nbnd[i];j++){
        k=cnbn[i][j];
        m=bnmultip[k];
        llnbn[k][m]=j;
        ldn[k][m]=i;
        bnmultip[k]++;
      }
    }
  }
}

/**
   function sorts node numbers of nodes shared at coarse node increasingly
   
   not checked

   JK, 11.7.2005
*/
void partop::sort_nodes (FILE *out)
{
  long i,j,k,m,n,min;
  
  if (myrank==0){
    for (i=0;i<tnbn;i++){
      for (j=0;j<bnmultip[i];j++){
        min=nproc;
        for (k=j;k<bnmultip[i];k++){
          if (ldn[i][k]<min){
            min=ldn[i][k];
            m=k;
          }
        }
        n=ldn[i][j];
        ldn[i][j]=ldn[i][m];
        ldn[i][m]=n;
        
        n=llnbn[i][j];
        llnbn[i][j]=llnbn[i][m];
        llnbn[i][m]=n;
      }
    }
  }
  
  /*
  if (myrank==0){
    if(md == all_nodes){
      fprintf (out,"\n\n Multiplicity of all nodes on whole problem\n");
      for (i=0;i<tnnp;i++){
        fprintf (out,"%ld  %ld\n",i+1,multip[i]); 
      }
    }
    fprintf (out,"\n\n Multiplicity of all boundary nodes on whole problem\n");
    for (i=0;i<tnbn;i++){
      fprintf (out,"%ld  %ld\n",i+1,bnmultip[i]); 
    }
    
    
    
    //fprintf (out,"\n\n  ldn and llnbn array\n");
    // llnbn ukazuje na pozici v poli lnbn
    //for (i=0;i<tnbn;i++){
    //fprintf (out,"\n boundary node %4ld  ",i+1);
    //for (j=0;j<bnmultip[i];j++){
    //fprintf (out," ldn =  %ld  llnbn =  %ld  lnbn = %ld,",ldn[i][j],llnbn[i][j],lnbn[llnbn[i][j]]);
    //}
    //}
  }
  */
}
  


/**
   function detects numbers of degrees of freedom on boundary nodes
   

   the following functions have to be called before this function:
   void initiation (gtopology *top,long *ltg);
   void numbers_of_all_nodes_on_subdomains (gtopology *top,long *domproc,FILE *out);
   void compute_multiplicity (gtopology *top,long *domproc,FILE *out);
   void find_boundary_nodes (gtopology *top,long *domproc,FILE *out);
   void boundary_nodes_on_master (gtopology *top,long *domproc,FILE *out);

   function assembles:
   
   nbdofnd - array containing numbers of DOFs on boundary nodes
   
   @param top - pointer to general topology
   @param domproc - array containing domain-processor correspondence

   JK, 6.7.2005
*/
void partop::number_of_bdofs_on_nodes (gtopology *top,long *domproc,FILE *out)
{
  long i,j,k,buffsize;
  long *buff;
  MPI_Status stat;
  
  if (lnbn==NULL){
    fprintf (stderr,"\n\n array containing local numbers of boundary nodes is not allocated,\n");
    fprintf (stderr,"\n function find_boundary_nodes has to be called first (file %s, line %d),\n",__FILE__,__LINE__);
  }

  buffsize = maxnbn;
  buff = new long [buffsize];
  
  //  detection of numbers of DOFs on boundary nodes
  for (i=0;i<nbn;i++){
    j=lnbn[i];
    buff[i]=top->give_ndofn (j);
  }
  
  if (myrank==0){
    if (nbdofnd != NULL){
      for (i=0;i<nproc;i++){
        delete [] nbdofnd[i];
      }
      delete [] nbdofnd;
    }
    nbdofnd = new long* [nproc];
    for (i=0;i<nproc;i++){
      nbdofnd[i] = new long [nbnd[i]];
    }
    
    //  contribution from master
    k=domproc[0];
    for (j=0;j<nbnd[k];j++){
      nbdofnd[k][j]=buff[j];
    }
    
    for (i=1;i<nproc;i++){
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      
      //  slave contributions
      k=domproc[stat.MPI_TAG];
      for (j=0;j<nbnd[k];j++){
        nbdofnd[k][j]=buff[j];
      }
      
    }
  }
  else{
    MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  
  delete [] buff;
  
  /*
  if (myrank==0){
    fprintf (out,"\n\n\n numbers of DOFs at boundary nodes");
    for (i=0;i<nproc;i++){
      fprintf (out,"\n\n domain %ld",i);
      for (j=0;j<nbnd[i];j++){
        fprintf (out,"\n %ld %ld",j+1,nbdofnd[i][j]);
      }
    }
  }
  */
}

/**
   function assembles code numbers on master

   in fact, it assembles indicators 0 or 1 which will be
   used later for code number generation
   
   the following functions have to be called before this function:
   void initiation (gtopology *top,long *ltg);
   void numbers_of_all_nodes_on_subdomains (gtopology *top,long *domproc,FILE *out);
   void compute_multiplicity (gtopology *top,long *domproc,FILE *out);
   void find_boundary_nodes (gtopology *top,long *domproc,FILE *out);
   void boundary_nodes_on_master (gtopology *top,long *domproc,FILE *out);
   void number_of_bdofs_on_nodes (gtopology *top,long *domproc,FILE *out);


   function assembles:
   
   nbdofd - array containing number of boundary DOFs on subdomains
   lbcn - array containing local boundary code numbers
   
   @param top - pointer to general topology
   @param domproc - array containing domain-processor correspondence
   
   JK, 6.7.2005
*/
void partop::code_numbers_on_master (gtopology *top,long *domproc,FILE *out)
{
  long i,j,k,m,l,ndofn,buffsize;
  long *buff;
  MPI_Status stat;
  
  if (myrank==0){
    if (nbdofnd==NULL){
      fprintf (stderr,"\n\n array ndofnd is not allocated in function code_numbers_on_master,\n");
      fprintf (stderr,"\n function number_of_dofs_on_nodes has to be called first (file %s, line %d),\n",__FILE__,__LINE__);
    }
  }
  
  //  detection of numbers of DOFs on particular subdomains
  if (myrank==0){
    if (nbdofd != NULL)
      delete [] nbdofd;
    nbdofd = new long [nproc];
    
    maxnbdof=0;
    for (i=0;i<nproc;i++){
      nbdofd[i]=0;
      for (j=0;j<nbnd[i];j++){
        nbdofd[i]+=nbdofnd[i][j];
      }
      if (maxnbdof<nbdofd[i])
        maxnbdof=nbdofd[i];
    }
    
    for (i=1;i<nproc;i++){
      MPI_Send (&maxnbdof,1,MPI_LONG,i,myrank,MPI_COMM_WORLD);
    }
  }
  else{
    MPI_Recv (&maxnbdof,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
  }
  
  buffsize = maxnbdof;
  buff = new long [buffsize];
  
  m=0;
  for (i=0;i<nbn;i++){
    k=lnbn[i];
    ndofn = top->give_ndofn (k);
    for (j=0;j<ndofn;j++){
      buff[m]=top->give_dof (k,j);
      m++;
    }
  }
  
  if (myrank==0){
    if (lbcn != NULL){
      for (i=0;i<nproc;i++){
        for (j=0;j<nbnd[i];j++){
          delete [] lbcn[i][j];
        }
        delete [] lbcn[i];
      }
      delete [] lbcn;
    }
    lbcn = new long** [nproc];
    for (i=0;i<nproc;i++){
      lbcn[i] = new long* [nbnd[i]];
      for (j=0;j<nbnd[i];j++){
        lbcn[i][j] = new long [nbdofnd[i][j]];
      }
    }
    
    //  master contribution
    l=0;
    m=domproc[0];
    for (j=0;j<nbnd[m];j++){
      for (k=0;k<nbdofnd[m][j];k++){
        lbcn[m][j][k]=buff[l];
        l++;
      }
    }
    
    for (i=1;i<nproc;i++){
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      
      //  slave contributions
      l=0;
      m=domproc[stat.MPI_TAG];
      for (j=0;j<nbnd[m];j++){
        for (k=0;k<nbdofnd[m][j];k++){
          lbcn[m][j][k]=buff[l];
          l++;
        }
      }

    }
  }
  else{
    MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
  }
  
  /*
  if (myrank==0){
    fprintf (out,"\n\n\n kontrola pole lbcn \n\n");
    for (i=0;i<nproc;i++){
      fprintf (out,"\n\n domena  %ld",i);
      for (j=0;j<nbnd[i];j++){
        fprintf (out,"\n boundary uzel  %ld",j);
        for (k=0;k<nbdofnd[i][j];k++){
          fprintf (out,"  %ld",lbcn[i][j][k]);
        }
        fprintf (out,"\n");
      }
    }
  }
  */
  
  
  MPI_Barrier (MPI_COMM_WORLD);
  
}














































































/**
   Function checks array ltg if there are corner/fixing nodes.
   If corner nodes are not detected in array ltg,
   function corner_detection  is called.
   
   @param domproc - array containing domain-processor correspondence
   @param ltg - local to global correspondence
   @param top - pointer to the sequential general topology
   @param out - output file (used for auxiliary output)
   @param proc_name - processor name
   
   JB 22.11.2009
 **/

void partop::check_corner_ltg (gtopology *top, long *domproc,long *ltg, FILE *out,char *proc_name)
{
  
  long i,neg,pos,one;
   switch (md){
   case all_nodes:{
     
     cornod = 1;
     break;
   }
   case bound_nodes:{
     //fprintf(out,"\n\nCheking ltg\n nn=%ld\n",nn);
     neg = 0;
     pos = 0;
     one = 0;
     for (i = 0; i < nn; i++ ){
       if (ltg[i] <= -2){
         neg++;
       }
       if (ltg[i] > -1){
         pos++;
       }
       if (ltg[i] == -1){
         one++;
       }
     }
     //fprintf(out,"Cheking neg=%ld pos=%ld one=%ld\n",neg,pos,one);
     //if ((neg+pos+one) == nn){
     //fprintf(out,"Cheking OK\n");
     //}
     // corner nodes are found
     if ( neg > 0){
       fprintf(out,"\n\n\nCorner nodes are found in ltg array\n");
       cornod = 0;
     }
     // corner nodes must be found
     else{
       fprintf(out,"\n\n\nCorner nodes must be found by function corner_detection\n");
       corner_detection(top,domproc,ltg,out,proc_name);
       cornod = 1; 
     }
     
     break;
   }
   case neg_bound_nodes:{
     
     cornod = 1;
     break;
   }
   default:{
     
   }
   }
   fflush(out);
}

/**
   Function changes array ltg. Function calls functions for 
   corner/fixing nodes selection. Output is new ltg array
   with corner/fixing nodes.
   
   @param domproc - array containing domain-processor correspondence
   @param ltg - local to global correspondence
   @param top - pointer to the sequential general topology
   @param out - output file (used for auxiliary output)
   @param proc_name - processor name
   
   JB 22.11.2009
 **/

void partop::corner_detection(gtopology *top, long *domproc, long *ltg, FILE *out,char *proc_name)
{
  time_t t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12;
  long i,j;
    
  //t1=time(NULL);
  // set of fictitious subdomain
  set_fictitious_subdomain(top,out);
  //t2=time(NULL);
  // size of fictitious subdomain
  compute_size_of_fictitious_subdomain(top,out);
  //t3=time(NULL);
    
  // set of spatial dimension of problem
  give_whole_dim(top);
  //t4=time(NULL);
  fprintf(out,"Dimension of problem is %ld\n",dim);
  switch(dim){
  case -1:
    par_print_err(myrank,"Spatial dimension is not detected", __FILE__, __LINE__, __func__);
    fflush(stderr);
    break;
  case 1:
    
    break;
  case 2:
    // node identification in 2D
    identification_node_2d (top,out);
    if(condcorner == automatic || condcorner == userdef){    
      prepare_master_curves(domproc,out);
      if(myrank == 0){
	// sestaveni krivek
	mascurves = new boundcurves(tnbn,icmultip,icnbnmas,coarseidentif,coarsenadjac,coarseadjac,out);
	//reduction of graph - no
	mascurves->reduce_flag = 0;
	mascurves->set_curves();
	mascurves->print_info();
	mascurves->add_corners(condcorner,typecondcorner,nmember,intpointo);
	mascurves->get_coarseidentif(coarseidentif);
      }
    }
    else{
      //check_geometry_of_corner_nodes(top,out);
      check_triangle(top,out);
      send_and_check_corner_nodes(domproc,ltg,out);
    }
    assemble_ltg_with_corners(domproc,ltg,out);
    if(condcorner == automatic || condcorner == userdef){    
      if(myrank == 0){
	mascurves->~boundcurves();
      }
    }
    fflush(out);
    break;
  case 3:
    //vyvojove umisteni
    //boundary_surface_detection(top,out);
    
    
    //node identification in 3d
    switch(condcorner){
    case nocondconer:{
      //first option - statistics
      compute_statistics_of_multiplicity(domproc,out);
      select_corner_nodes_on_master(domproc,out);
      check_triangle(top,out);
      //assemble_ltg_with_corners(domproc,ltg,out);
      
      //second option - identification
      //identification_node_3d (top,out);
      //check_triangle(top,out);
      send_and_check_corner_nodes(domproc,ltg,out);
      assemble_ltg_with_corners(domproc,ltg,out);
      break;
    }
    case automatic:{
      
      break;
    }
    case userdef:{
      identification_node_3d (top,out);
      check_geometry_of_corner_nodes(top,out);
      send_and_check_corner_nodes(domproc,ltg,out);
      prepare_master_curves(domproc,out);
      if(myrank == 0){
	// sestaveni krivek
	mascurves = new boundcurves(tnbn,icmultip,icnbnmas,coarseidentif,coarsenadjac,coarseadjac,out);
	mascurves->reduce_graph();
	mascurves->set_curves_3D();
	mascurves->print_info();
	mascurves->add_corners(condcorner,typecondcorner,nmember,intpointo);
	mascurves->get_coarseidentif(coarseidentif);
      }
      assemble_ltg_with_corners(domproc,ltg,out);
      if(myrank == 0){
	mascurves->~boundcurves();
      }
      break;
    }
    }
    fflush(out);
    break;
    
  default:
    par_print_err(myrank,"Spatial dimension is not detected", __FILE__, __LINE__, __func__);
    fflush(stderr);
   break;
  }
  
  
  delete []fictdom;
  delete []nodeidentif;
}


/*
  function establishes fictitious subdomain around real subdomain
  fictitious subdomain is rectangle in two dimension
  fictitious subdomain is cube in three dimension

  fictdom[0] = k - min_x coordinates
  fictdom[1] = k - min_y coordinates
  fictdom[2] = k - min_z coordinates
  fictdom[3] = k - max_x coordinates
  fictdom[4] = k - max_y coordinates
  fictdom[5] = k - max_z coordinates

  @param top - pointer to the sequential general topology
  @param out - output file (used for auxiliary output)
  
  10.02.2009 JB
 */

void partop::set_fictitious_subdomain(gtopology *top,FILE *out)
{
  long i;
  double  miny,minx,minz,maxy,maxx,maxz,x,y,z;
  
  if(fictdom != NULL){
    delete []fictdom;
  }
  fictdom = new double[6]; 
  //miny=DBL_MAX;
  //minx=DBL_MAX;
  //minz=DBL_MAX;
  //maxy=DBL_MIN;
  //maxx=DBL_MIN;
  //maxz=DBL_MIN;
  minx = top->gnodes[0].x;
  miny = top->gnodes[0].y;
  minz = top->gnodes[0].z;
  maxx = top->gnodes[0].x;
  maxy = top->gnodes[0].y;
  maxz = top->gnodes[0].z;
  for(i = 0; i < nn; i++){
    x = top->gnodes[i].x;
    y = top->gnodes[i].y;
    z = top->gnodes[i].z;	
    //fprintf(out,"%ld %lf %lf %lf\n",i,x,y,z);
    // min_x
    if(x <= minx){
      minx = x;
      fictdom[0] = x;
    }
    // min_y
    if(y <= miny){
       miny = y;
      fictdom[1] = y;
    }
    // min_z
    if(z <= minz){
      minz = z;
      fictdom[2] = z;
    }
    // max_x
    if(x >= maxx){
      maxx = x;
      fictdom[3] = x;
    }
    // max_y
    if(y >= maxy){
      maxy = y;
      fictdom[4] = y;
    }
    // max_z
    if(z >= maxz){
      maxz = z;
      fictdom[5] = z;
    }
  }
  fprintf(out,"Coordinates of fictitious domain: %lf %lf %lf %lf %lf %lf\n",fictdom[0],fictdom[1],fictdom[2],fictdom[3],fictdom[4],fictdom[5]);
  fflush(out);
}

/**
   Function computes distance between nodes with minimal and maximal coordinates
   of fictitious subdomain. Output is variable double sizefictdom.
   
   @param top - pointer to the sequential general topology
   @param out - output file (used for auxiliary output)

   10.02.2009 JB
   
**/

void partop::compute_size_of_fictitious_subdomain(gtopology *top,FILE *out)
{
  double vetcx,vetcy,vetcz,sizevect;
  //size_fictdom
  vetcx = fictdom[3] - fictdom[0];
  vetcy = fictdom[4] - fictdom[1];
  vetcz = fictdom[5] - fictdom[2];
  //fprintf(out,"velikost fictdom %lf %lf %lf\n",vetc_x,vetc_y,vetc_z);
  sizevect = vetcx*vetcx + vetcy*vetcy + vetcz*vetcz;
  sizefictdom = sqrt(sizevect);
  fprintf(out,"Size of fictitious domain %lf\n",sizefictdom);
  fflush(out);
}


/**
   Function recognizes spatial dimension of the used mesh.
   Output is variable long dim.
   
   @param top - general topology
   
   29.5.2007, JB
**/
void partop::give_whole_dim(gtopology *top)
{
  
  gelemtype get;
  long *elem;
  long i,t;
 
  
  elem = new long[ne];
  
  for(i = 0; i < ne; i++){
    get =  top->give_elem_type (i);
    switch(get){
    case linbar:{     t = 1;    break;  }
    case quadbar:{    t = 1;    break;  }
    case lintriag:{   t = 2;    break;  }
    case quadtriag:{  t = 2;    break;  }
    case linquad:{    t = 2;    break;  }
    case quadquad:{   t = 2;    break;  }
    case lintetra:{   t = 3;    break;  }
    case quadtetra:{  t = 3;    break;  }
    case linhexa:{    t = 3;    break;  }
    case quadhexa:{   t = 3;    break;  }
    default:{
      par_print_err(myrank,"Unknown element type is required", __FILE__, __LINE__, __func__);
    }
    }
    elem[i] = t;
  }
  
  
  for(i = 1; i < ne; i++){
    if(elem[i] != elem[i-1]){
      par_print_err(myrank,"different spatial dimension is found in the mesh, function identification_node can not be used", __FILE__, __LINE__, __func__);
    }
    else{
      dim = elem[i];
    }
  }
  
  delete []elem;
}

/**
   Function computes length of vector which is defined by nodes a and b
   Output is variable double u.
   
   @param top - pointer to the sequential general topology
   @param a - the first node of vector
   @param b - the second node of vector
   @param out - output file (used for auxiliary output)
      
   JB 24.09.2009
 **/
double partop::compute_length_of_vector(gtopology *top,long a,long b,FILE *out)
{
  double u1,u2,u3,x1,x2,y1,y2,z1,z2,u;
    
  //node 1 
  x1 = top->gnodes[a].x;
  y1 = top->gnodes[a].y;
  z1 = top->gnodes[a].z;
  // node 2
  x2 = top->gnodes[b].x;
  y2 = top->gnodes[b].y;
  z2 = top->gnodes[b].z;
  // vecotor u - nodes 1 and 2
  u1 = x2 - x1;
  //fprintf(out,"u1 = %le\n",u1);
  u2 = y2 - y1;
  //fprintf(out,"u2 = %le\n",u2);
  u3 = z2 - z1;
  //  magnitude of vectoru u
  u = u1*u1+u2*u2+u3*u3;
  u=sqrt(u);
  return(u);
}

/**
   Function computes angle between two vectors  which are defined by nodes a, b and c
   vector u = b - a
   vector v = c - a
   Output is variable double angle.
   
   @param top - pointer to the sequential general topology
   @param a - the first node of vector u and vector v
   @param b - the second node of vector u
   @param c - the second node of vector v
   @param out - output file (used for auxiliary output)
   
   JB 24.09.2009
**/

double partop::compute_angle_of_vector_a_and_b(gtopology *top,long a,long b,long c,FILE *out)
{
  double u1,u2,u3,v1,v2,v3,x1,x2,x3,y1,y2,y3,z1,z2,z3,scal,u,v,cosalpha,angle;
  
  //node 1 
  x1 = top->gnodes[a].x;
  y1 = top->gnodes[a].y;
  z1 = top->gnodes[a].z;
  // node 2
  x2 = top->gnodes[b].x;
  y2 = top->gnodes[b].y;
  z2 = top->gnodes[b].z;
  // vecotor u - nodes 1 and 2
  u1 = x2 - x1;
  //fprintf(out,"u1 = %le\n",u1);
  u2 = y2 - y1;
  //fprintf(out,"u2 = %le\n",u2);
  u3 = z2 - z1;
  //fprintf(out,"u3 = %le\n",u3);
  // node 3
  x3 = top->gnodes[c].x;
  y3 = top->gnodes[c].y;
  z3 = top->gnodes[c].z; 
  // vector v - nodes 1 and 3
  v1 = x3 - x1;
  //fprintf(out,"v1 = %le\n",v1);
  v2 = y3 - y1;
  //fprintf(out,"v2 = %le\n",v2);
  v3 = z3 - z1;
  //fprintf(out,"v3 = %le\n",v3);
  // scalar product
  scal = u1*v1+u2*v2+u3*v3;
  //fprintf(out,"scal = %le\n",scal);
  //  magnitude of vector  u
  u = u1*u1+u2*u2+u3*u3;
  //fprintf(out,"u = %le\n",u);
  u = sqrt(u);
  //fprintf(out,"u = %le\n",u);
  // magnitude of vector  v
  v = v1*v1+v2*v2+v3*v3;
  //fprintf(out,"v = %le\n",v);
  v = sqrt(v);
  //fprintf(out,"v = %le\n",v);
  // angle
  cosalpha = scal/(u*v);
  
  if(cosalpha > 1.0){
    cosalpha = 1.0;
  }
  if(cosalpha < -1.0){
    cosalpha = -1.0;
  }
  angle = acos(cosalpha);
  return(angle);

}

/**
   Function compute area of triangle which is defined by nodes a, b and c.
   Computation is based on determinant of coordinates of nodes.
   Output is variable double area.
   
   @param top - pointer to the sequential general topology
   @param a - the first node of triangle
   @param b - the second node of triangle
   @param c - the third node of triangle
   @param out - output file (used for auxiliary output)

   
   JB 24.09.2009
**/

double partop::compute_area_of_triangle(gtopology *top,long a,long b,long c,FILE *out)
{
  double x1,x2,x3,y1,y2,y3,z1,z2,z3,area;
  
  //node 1 
  x1 = top->gnodes[a].x;
  y1 = top->gnodes[a].y;
  z1 = top->gnodes[a].z;
  // node 2
  x2 = top->gnodes[b].x;
  y2 = top->gnodes[b].y;
  z2 = top->gnodes[b].z;
  // node 3
  x3 = top->gnodes[c].x;
  y3 = top->gnodes[c].y;
  z3 = top->gnodes[c].z;
  
  area = x1*y2*z3+y1*z2*x3+z1*x2*y3-z1*y2*x3-y1*x2*z3-x1*z2*y3;
  
  return(area);
}

// identification of the minimal number of corner/fixing nodes
void partop::identification_node_2d (gtopology *top, FILE *out)
{
  long i,j,k,l,m;
  long ned,nned,auxnedges,auxnv;
  long *auxnned,*dupledges,*aux,*locbound;
  long **auxedgenodes;
  
  // identifikace uzlu na domene
  if (nodeidentif != NULL)
    delete []nodeidentif;
  nodeidentif = new long [nbn];
  for(i = 0; i < nbn; i++){
    nodeidentif[i] = 2;
  }
  k = 0;
  auxnv = 0;
  for(i = 0; i < nbn; i++){
    // interface node
    if(nodmultip[lnbndom[i]] > 2){
      nodeidentif[i] = 3;
      auxnv++;
    }
  }
  if(condcorner == automatic || condcorner == userdef){    
    auxnv  = 0;
  }
  if( auxnv < 3){
    //fprintf(out,"maly pocet vrchulu %ld\n",k);
    
    // number of all edges in mesh
    auxnedges = 0;
    // pomocne pole
    aux = new long[3];
    for(i = 0; i < ne; i++){
      // number of edge on i-th element - gtopology
      ned = top->give_ned(i);
      nned = top->give_nned(i);
      for(j = 0; j < ned; j++ ){
	top->give_edge_nodes(i,j,aux);
	m = 0;
	for(l = 0; l < nned; l++){
	  if(nodmultip[aux[l]] > 1){
	    m++;
	  }
	}
	if(m > 1){
	  auxnedges++;
	}
      }
    }
    //fprintf(out,"pocet hran %ld\n",auxnedges);
    
    // plneni pole s uzky na hranach
    // plneni s pole s poctem uzlu na hrane
    auxnned = new long[auxnedges];
    auxedgenodes = new long*[auxnedges];
    dupledges = new long[auxnedges];
    // pomocne pole
    //aux = new long[3];
    k = 0;
    for(i = 0; i < ne; i++){
      ned = top->give_ned(i);
      nned = top->give_nned(i);
      for(j = 0; j < ned; j++ ){
	top->give_edge_nodes(i,j,aux);
	m = 0;
	for(l = 0; l < nned; l++){
	  if(nodmultip[aux[l]] > 1){
	    m++;
	  }
	}
	if(m > 1){
	  auxnned[k] = nned;
	  auxedgenodes[k] = new long[nned];
	  for(l = 0; l < nned; l++){
	    auxedgenodes[k][l] = aux[l];
	  }
	  // trideni pole s uzly na hrane(od nejmensiho po nejvetsi)
	  if(nned == 2){
	    if(auxedgenodes[k][0] > auxedgenodes[k][1]){
	      l = auxedgenodes[k][1];
	      auxedgenodes[k][1] = auxedgenodes[k][0];
	      auxedgenodes[k][0] = l;
	    }
	  }
	  if(nned == 3){
	    if(auxedgenodes[k][0] > auxedgenodes[k][2]){
	      l = auxedgenodes[k][2];
	      auxedgenodes[k][2] = auxedgenodes[k][0];
	      auxedgenodes[k][0] = l;
	    }
	    if(auxedgenodes[k][0] > auxedgenodes[k][1]){
	      l = auxedgenodes[k][1];
	      auxedgenodes[k][1] = auxedgenodes[k][0];
	      auxedgenodes[k][0] = l;
	    }
	    if(auxedgenodes[k][1] > auxedgenodes[k][2]){
	      l = auxedgenodes[k][2];
	      auxedgenodes[k][2] = auxedgenodes[k][1];
	      auxedgenodes[k][1] = l;
	    }
	  }
	  dupledges[k] = 0;
	  k++;
	}
      }
    }
    delete []aux;
    //auxnedges = k;
    //fprintf(out,"pocet hran %ld\n",auxnedges);
    
    
    // kontrolni tisk
    //for(i = 0; i < auxnedges; i++){
    //fprintf(out,"hrana %ld ma uzly:",i);
    //for(j = 0; j < auxnned[i]; j++ ){
    //  fprintf(out,"   %ld",auxedgenodes[i][j]);
    //}
    //fprintf(out,"\n");
    //}
    
    // hledani duplikovanych hran na hranici
    for(i = 0; i < auxnedges; i++){
      for(j = i+1; j < auxnedges; j++){
	if(auxedgenodes[i][0] == auxedgenodes[j][0]){
	  if(auxnned[i] == 2 && auxnned[j] == 2){
	    if(auxedgenodes[i][1] == auxedgenodes[j][1]){
	      dupledges[j]++;
	      dupledges[i]++;
	    }
	  }
	  if(auxnned[i] == 3 && auxnned[i] == 3){
	    if(auxedgenodes[i][1] == auxedgenodes[j][1]){
	      if(auxedgenodes[i][2] == auxedgenodes[j][2]){
		dupledges[j]++;
		dupledges[i]++;
	      }
	    }
	  }
	}
      }
    }
    
    // kontrolni tisk
    //for(i = 0; i < auxnedges; i++){
    //fprintf(out,"%ld dupledges %ld\n",i,dupledges[i]);
    //}
    
    locbound = new long[nn];
    m = 0;
    for(i = 0; i < nn; i++){
      if(nodmultip[i] > 1){
	locbound[i] = m;
	m++;
      }
      else{
	locbound[i] = -1;
      }
    }
    
    // kontrolni tisk
    //for(i = 0; i < nbn; i++){
    // fprintf(out,"%ld lnbndom %ld\n",i,lnbndom[i]);
    //}
    //for(i = 0; i < nn; i++){
    //fprintf(out,"%ld locbound %ld\n",i,locbound[i]);
    //}
    
    // renumbering
    //if(adjacboundnod != NULL){
    //for(j = 0; j < nadjacboundnod[i]; j++){
    //delete []adjacboundnod[i];
    //}
    //delete []adjacboundnod;
    //}
    //if(nadjacboundnod != NULL){
    //delete []nadjacboundnod;
    //}
    
    
    nadjacboundnod = new long[nbn];
    for(i = 0; i < nbn; i++){
      nadjacboundnod[i] = 0;
      for(j = 0; j < auxnedges; j++){
	if(dupledges[j] == 0) {
	  for(k = 0; k < auxnned[j]; k++){
	    if(auxedgenodes[j][k] == lnbndom[i]){
	      auxedgenodes[j][k] = i;
	      nadjacboundnod[i]++;
	    }
	  }
	}
      } 
    }
    
    // kontrolni tisk
    //for(i = 0; i < auxnedges; i++){
    //fprintf(out,"hrana %ld ma uzly:",i);
    ///for(j = 0; j < auxnned[i]; j++ ){
    // fprintf(out,"   %ld",auxedgenodes[i][j]);
    //}
    //fprintf(out,"\n");
    //}
    //for(i = 0; i < nbn; i++){
    //fprintf(out,"%ld\n",nadjacboundnod[i]);
    //}  
    
    
    // identifikace uzlu na domene
    k = 0;
    for(i = 0; i < nn; i++){
      
      // interface node
      if(nodmultip[i] > 2){
	if(nodeidentif[k] == 2){
	  nodeidentif[k] = 3;
	}
	k++;
      }
      if(nodmultip[i] == 2){
	if(nadjacboundnod[locbound[i]] == 1){
	  nodeidentif[k] = 3;
	  k++;
	}
	else{
	  nodeidentif[k] = 2;
	  k++;
	}
      }
    }
  }
  
  //for (i = 0; i < nbn; i++){
  //if(nodeidentif[i] != 2 || nodeidentif[i] != 3) fprintf(out,"\n node number %ld is not identification",lnbndom[i]+1);
  //if(nodeidentif[i] == 2) fprintf(out,"\n node number %ld is edge node",lnbndom[i]+1);
  //if(nodeidentif[i] == 3) fprintf(out,"\n node number %ld is vertex node",lnbndom[i]+1);
  //}
  //fflush(out);
  //fprintf(out,"\n\n");
    
  nv = 0;
  for (i = 0; i < nbn; i++){
    if(nodeidentif[i] == 3){
      nv++;
    }
  }
  fprintf(out,"Number of potential corner nodes %ld\n\n",nv);
  if(nv < 2){
    // funkce pro zahusteni - vzdy vyber 3
  }
  
  vertices = new long[nv];
  k = 0;
  for (i = 0; i < nbn; i++){
    if(nodeidentif[i] == 3){
      vertices[k]=i;
      k++;
    }
  }
  
  
  if(condcorner == automatic || condcorner == userdef){    
    
    // sestaveni sousedu hranicnich uzlu (pouze na hranici) pro pripadne sestaveni krivek
    adjacboundnod = new long*[nbn];
    for(i = 0; i < nbn; i++){
      adjacboundnod[i] = new long[nadjacboundnod[i]];
      m = 0;
      for(j = 0; j < auxnedges; j++){
	if(dupledges[j] == 0) {
	  if(auxnned[j] == 2){
	    for(k = 0; k < auxnned[j]; k++){
	      if(auxedgenodes[j][k] == i){
		if(k == 0){
		  adjacboundnod[i][m] = auxedgenodes[j][1];
		  m++;
		}
		if(k == 1){
		  adjacboundnod[i][m] = auxedgenodes[j][0];
		  m++;
		}
	      }
	    }
	  }
	  if(auxnned[j] == 3){
	    for(k = 0; k < auxnned[j]; k++){
	      if(auxedgenodes[j][k] == i){
		if(k == 0){
		  adjacboundnod[i][m] = auxedgenodes[j][1];
		  m++;
		}
		if(k == 1){
		  adjacboundnod[i][m] = auxedgenodes[j][0];
		  m++;
		  adjacboundnod[i][m] = auxedgenodes[j][2];
		  m++;
		}
		if(k == 2){
		  adjacboundnod[i][m] = auxedgenodes[j][1];
		  m++;
		}
	      }
	    }
	  }
	}
      }
    }
    // kontrolni tisk
    //for(i = 0; i < nbn; i++){
    // fprintf(out,"uzel %ld ma  %ld sousedu: ",i,nadjacboundnod[i]);
    //for(j = 0; j < nadjacboundnod[i]; j++){
    //fprintf(out,"   %ld",adjacboundnod[i][j]);
    //}
    //fprintf(out,"\n");
    //}
    //for(i = 0; i < nbn; i++){
    // fprintf(out,"uzel %ld ma icnbn %ld\n",i,icnbn[i]);
    //}
    
    for(i = 0; i < nbn; i++){
      for(j = 0; j < nadjacboundnod[i]; j++){
	k = adjacboundnod[i][j];
	adjacboundnod[i][j] = icnbn[k];
      }
    }
    
    
    //sorting 
    for(i = 0; i < nbn; i++){
      if(nadjacboundnod[i] == 2){
	if(adjacboundnod[i][0] > adjacboundnod[i][1]){
	  j = adjacboundnod[i][1];
	  adjacboundnod[i][1] = adjacboundnod[i][0];
	  adjacboundnod[i][0] = j;
	}
      }
    }
    
    
    // kontrolni tisk
    // for(i = 0; i < nbn; i++){
    //       fprintf(out,"uzel %ld ma  %ld sousedu: ",i,nadjacboundnod[i]);
    //       for(j = 0; j < nadjacboundnod[i]; j++){
    // 	fprintf(out,"   %ld",adjacboundnod[i][j]);
    //       }
    //       fprintf(out,"\n");
    //     }
    
      
  }
  if(auxnv < 3){
    // mazani poli
    for(i = 0; i < auxnedges; i++){
      delete []auxedgenodes[i];
    }
    delete []auxnned;
    delete []auxedgenodes;
    delete []dupledges;
    delete []locbound;
  }
  
  
  j = 0;
  for(i = 0; i < nbn; i++){
    if(nodeidentif[i] == 3){
      fprintf(out,"%ld is corner node\n\n",lnbndom[i]);
      j++;
    }
  }
  fprintf(out,"Number of potential corner nodes %ld\n\n",j);
}


void partop::identification_node_3d (gtopology *top,FILE *out)
{
  long i,j,k,l,m,a,b,c,d,n;
  long ned,nned,auxnedges,auxnv;
  long nsurf,nnsurf,auxnsurf;
  long *auxnned,*dupledges,*aux,*locbound;
  long *auxnnsurf,*duplsurf;
  long **auxedgenodes;
  long **auxsurfnodes;
  time_t t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12;
  long max;
  long *vcomb;
  long *ver;
  double radius;
  double maxradius;
  double area;
  double maxarea;
  double cosalpha;
  double *contrad;
    
  
  t1=time(NULL);
  
  // number of all edges in mesh
    auxnedges = 0;
    // pomocne pole
    aux = new long[3];
    for(i = 0; i < ne; i++){
      // number of edge on i-th element - gtopology
      ned = top->give_ned(i);
      nned = top->give_nned(i);
      for(j = 0; j < ned; j++ ){
	top->give_edge_nodes(i,j,aux);
	m = 0;
	for(l = 0; l < nned; l++){
	  if(nodmultip[aux[l]] > 1){
	    m++;
	  }
	}
	if(m > 1){
	  auxnedges++;
	}
      }
    }
    //fprintf(out,"pocet hran %ld\n",auxnedges);
    
    // plneni pole s uzky na hranach
    // plneni s pole s poctem uzlu na hrane
    auxnned = new long[auxnedges];
    auxedgenodes = new long*[auxnedges];
    dupledges = new long[auxnedges];
    k = 0;
    for(i = 0; i < ne; i++){
      ned = top->give_ned(i);
      nned = top->give_nned(i);
      for(j = 0; j < ned; j++ ){
	top->give_edge_nodes(i,j,aux);
	m = 0;
	for(l = 0; l < nned; l++){
	  if(nodmultip[aux[l]] > 1){
	    m++;
	  }
	}
	if(m > 1){
	  auxnned[k] = nned;
	  auxedgenodes[k] = new long[nned];
	  for(l = 0; l < nned; l++){
	    auxedgenodes[k][l] = aux[l];
	  }
	  // trideni pole s uzly na hrane(od nejmensiho po nejvetsi)
	  if(nned == 2){
	    if(auxedgenodes[k][0] > auxedgenodes[k][1]){
	      l = auxedgenodes[k][1];
	      auxedgenodes[k][1] = auxedgenodes[k][0];
	      auxedgenodes[k][0] = l;
	    }
	  }
	  if(nned == 3){
	    if(auxedgenodes[k][0] > auxedgenodes[k][2]){
	      l = auxedgenodes[k][2];
	      auxedgenodes[k][2] = auxedgenodes[k][0];
	      auxedgenodes[k][0] = l;
	    }
	    if(auxedgenodes[k][0] > auxedgenodes[k][1]){
	      l = auxedgenodes[k][1];
	      auxedgenodes[k][1] = auxedgenodes[k][0];
	      auxedgenodes[k][0] = l;
	    }
	    if(auxedgenodes[k][1] > auxedgenodes[k][2]){
	      l = auxedgenodes[k][2];
	      auxedgenodes[k][2] = auxedgenodes[k][1];
	      auxedgenodes[k][1] = l;
	    }
	  }
	  dupledges[k] = 0;
	  k++;
	}
      }
    }
    delete []aux;
    
    //fprintf(out,"pocet hran %ld\n",auxnedges);
    
    
    // kontrolni tisk
    //for(i = 0; i < auxnedges; i++){
    //fprintf(out,"hrana %ld ma uzly:",i);
    //for(j = 0; j < auxnned[i]; j++ ){
    //fprintf(out,"   %ld",auxedgenodes[i][j]);
    //}
    //fprintf(out,"\n");
    //}
    
    // hledani duplikovanych hran na hranici
    for(i = 0; i < auxnedges; i++){
      for(j = i+1; j < auxnedges; j++){
	if(auxedgenodes[i][0] == auxedgenodes[j][0]){
	  if(auxnned[i] == 2 && auxnned[j] == 2){
	    if(auxedgenodes[i][1] == auxedgenodes[j][1]){
	      dupledges[j]++;
	      //dupledges[i]++;
	    }
	  }
	  if(auxnned[i] == 3 && auxnned[i] == 3){
	    if(auxedgenodes[i][1] == auxedgenodes[j][1]){
	      if(auxedgenodes[i][2] == auxedgenodes[j][2]){
		dupledges[j]++;
		//dupledges[i]++;
	      }
	    }
	  }
	}
      }
    }
    
    // kontrolni tisk
    //for(i = 0; i < auxnedges; i++){
    //fprintf(out,"%ld dupledges %ld\n",i,dupledges[i]);
    //}
    k = 0;
    for(i = 0; i < auxnedges; i++){
      if(dupledges[i] == 0){
	k++;
      }
    }
    //long nedges;
    //nedges = k;
    //fprintf(out," nedges je %ld\n",nedges);

    if(nadjacboundnod != NULL ){
      delete []nadjacboundnod;
    }
    
    nadjacboundnod = new long[nbn];
    for(i = 0; i < nbn; i++){
      nadjacboundnod[i] = 0;
      for(j = 0; j < auxnedges; j++){
	if(dupledges[j] == 0) {
	  for(k = 0; k < auxnned[j]; k++){
	    if(auxedgenodes[j][k] == lnbndom[i]){
	      auxedgenodes[j][k] = i;
	      nadjacboundnod[i]++;
	    }
	  }
	}
      } 
    }
    
    //for(i = 0; i < nbn; i++){
    //fprintf(out,"%ld\n",nadjacboundnod[i]);
    //}  
    
    if(adjacboundnod != NULL){
      for(i = 0; i < nbn; i++){
	 delete []adjacboundnod[i];
      }
      delete []adjacboundnod;
    }
    adjacboundnod = new long*[nbn];
    for(i = 0; i < nbn; i++){
      adjacboundnod[i] = new long[nadjacboundnod[i]];
      m = 0;
      for(j = 0; j < auxnedges; j++){
	if(dupledges[j] == 0) {
	  if(auxnned[j] == 2){
	    for(k = 0; k < auxnned[j]; k++){
	      if(auxedgenodes[j][k] == i){
		if(k == 0){
		  adjacboundnod[i][m] = auxedgenodes[j][1];
		  m++;
		}
		if(k == 1){
		  adjacboundnod[i][m] = auxedgenodes[j][0];
		  m++;
		}
	      }
	    }
	  }
	  if(auxnned[j] == 3){
	    for(k = 0; k < auxnned[j]; k++){
	      if(auxedgenodes[j][k] == i){
		if(k == 0){
		  adjacboundnod[i][m] = auxedgenodes[j][1];
		  m++;
		}
		if(k == 1){
		  adjacboundnod[i][m] = auxedgenodes[j][0];
		  m++;
		  adjacboundnod[i][m] = auxedgenodes[j][2];
		  m++;
		}
		if(k == 2){
		  adjacboundnod[i][m] = auxedgenodes[j][1];
		  m++;
		}
	      }
	    }
	  }
	}
      }
    }
    // kontrolni tisk
    //for(i = 0; i < nbn; i++){
    // fprintf(out,"uzel %ld ma  %ld sousedu: ",i,nadjacboundnod[i]);
    //for(j = 0; j < nadjacboundnod[i]; j++){
    //fprintf(out,"   %ld",adjacboundnod[i][j]);
    //}
    //fprintf(out,"\n");
    //}
    //for(i = 0; i < nbn; i++){
    // fprintf(out,"uzel %ld ma icnbn %ld\n",i,icnbn[i]);
    //}
    
    
    
    //sorting 
    for(i = 0; i < nbn; i++){
      if(nadjacboundnod[i] == 2){
	if(adjacboundnod[i][0] > adjacboundnod[i][1]){
	  j = adjacboundnod[i][1];
	  adjacboundnod[i][1] = adjacboundnod[i][0];
	  adjacboundnod[i][0] = j;
	}
      }
    }
    
    
    // kontrolni tisk
    // for(i = 0; i < nbn; i++){
    //       fprintf(out,"uzel %ld ma  %ld sousedu: ",i,nadjacboundnod[i]);
    //       for(j = 0; j < nadjacboundnod[i]; j++){
    // 	fprintf(out,"   %ld",adjacboundnod[i][j]);
    //       }
    //       fprintf(out,"\n");
    //     }
    

    // identifikace vnitrnich uzlu na domene
    if (nodeidentif != NULL)
      delete []nodeidentif;
    nodeidentif = new long [nbn];
    auxnv = 0;
    for(i = 0; i < nbn; i++){
      nodeidentif[i] = 2;
      if(nodmultip[lnbndom[i]] > 2){
	//fprintf(out,"uzel %ld multip %ld\n",i,nodmultip[lnbndom[i]]);
	b = 0;
	c = 0;
	d = 0;
	for(j = 0; j < nadjacboundnod[i]; j++){
	  a = lnbndom[adjacboundnod[i][j]];
	  //fprintf(out,"multiplicita %ld je %ld\n",adjacboundnod[i][j],nodmultip[a]);
	  if(nodmultip[a] <  nodmultip[lnbndom[i]]){
	    b++;
	  }
	  if(nodmultip[a] == nodmultip[lnbndom[i]]){
	    c++;
	  }
	  if(nodmultip[a] > nodmultip[lnbndom[i]]){
	    d++;
	  }
	}
	//fprintf(out,"b %ld c %ld d %ld\n",b,c,d);
	// bod dotyku nodmultip[lnbndom[i]] podoblasti - krizeni krivek o multiplicite > 2
	if(c == 0 && d == 0 && b == nadjacboundnod[i]){
	  nodeidentif[i] = 3;
	}
	// pocatek krivky o multiplicite nodmultip[lnbndom[i]] 
	if(c == 1 && d == 0 ){
	  nodeidentif[i] = 3;
	}
      }
      if(nodeidentif[i] == 3){
	auxnv++;
      }
    }
    fprintf(out,"Number of potential corner nodes %ld\n\n",auxnv);
    t2=time(NULL);
    //fprintf(out,"cas %ld\n\n",t2-t1);
    if(auxnv < 3){
      //eke_corner_nodes_on_subdomain();
      if(auxnv == 0){
	ver = new long[2];
	max = 0;
	for(i = 0; i < nbn; i++){
	  if(max < icnbn[i]){
	    max = icnbn[i];
	    ver[0] = i;
	  }
	}
	//fprintf(out,"max je %ld\n\n",max);
	//fprintf(out,"prvni je %ld\n\n",lnbndom[ver[0]]);
	
	vcomb = new long[3];
	vcomb[0] = lnbndom[ver[0]];
	maxradius = 0.0;
	for(i = 0; i < nbn; i++){
	  vcomb[1] = lnbndom[i];
	  radius = compute_length_of_vector(top,vcomb[0],vcomb[1],out);
	  //fprintf(out,"radius je %lf\n\n",radius);
	  if(radius >= maxradius){
	    maxradius = radius;
	    ver[1] = i;
	  }
	}
	//fprintf(out,"dalsi je %ld\n\n",lnbndom[ver[1]]);
	//fprintf(out,"maxradius je %lf\n\n",maxradius);
	//fprintf(out,"potencialni vrcholy jsou %ld %ld\n\n",lnbndom[ver[0]],lnbndom[ver[1]]);
	
	vcomb[0] = lnbndom[ver[1]];
	for(i = 0; i < nbn; i++){
	  vcomb[1] = lnbndom[i];
	  radius = compute_length_of_vector(top,vcomb[0],vcomb[1],out);
	  if(radius >= maxradius){
	    j = i;
	    maxradius = radius;
	  }
	}
	if(lnbndom[j] != lnbndom[ver[0]]){
	  ver[0] = j;
	}
	nodeidentif[ver[0]]=3;
	nodeidentif[ver[1]]=3;
	vcomb[0] = lnbndom[ver[0]];
	vcomb[1] = lnbndom[ver[1]];
	//fprintf(out,"potencialni vrcholy jsou %ld %ld\n\n",lnbndom[ver[0]],lnbndom[ver[1]]);
	
	delete []ver;
	
	maxarea = 0.0;
	for(i = 0; i < nbn; i++){
	  vcomb[2] = lnbndom[i];
	  area = compute_area_of_triangle(top,vcomb[0],vcomb[1],vcomb[2],out);
	  if(maxarea <= area){
	    maxarea = area;
	  }
	}
	
	for(i = 0; i < nbn; i++){
	  vcomb[2] = lnbndom[i];
	  area = compute_area_of_triangle(top,vcomb[0],vcomb[1],vcomb[2],out);
	  if(maxarea <= area){
	    nodeidentif[i] = 3;
	  }
	}
	delete []vcomb;
	
	// for(i = 0; i < nbn; i++){
	// 	  if(nodeidentif[i] == 3){
	// 	    fprintf(out,"%ld je corner\n\n",lnbndom[i]);
	// 	  }
	// 	}
	
      }
      
      if(auxnv == 1){
	ver = new long[2];
	max = 0;
	for(i = 0; i < nbn; i++){
	  if(nodeidentif[i] == 3){
	    ver[0] = i;
	  }
	}
	//fprintf(out,"max je %ld\n\n",max);
	//fprintf(out,"prvni je %ld\n\n",lnbndom[ver[0]]);
	
	vcomb = new long[3];
	vcomb[0] = lnbndom[ver[0]];
	maxradius = 0.0;
	for(i = 0; i < nbn; i++){
	  vcomb[1] = lnbndom[i];
	  radius = compute_length_of_vector(top,vcomb[0],vcomb[1],out);
	  //fprintf(out,"radius je %lf\n\n",radius);
	  if(radius >= maxradius){
	    maxradius = radius;
	    ver[1] = i;
	  }
	}
	//fprintf(out,"dalsi je %ld\n\n",lnbndom[ver[1]]);
	//fprintf(out,"maxradius je %lf\n\n",maxradius);
	//fprintf(out,"potencialni vrcholy jsou %ld %ld\n\n",lnbndom[ver[0]],lnbndom[ver[1]]);
	
	nodeidentif[ver[1]]=3;
	vcomb[0] = lnbndom[ver[0]];
	vcomb[1] = lnbndom[ver[1]];
	
	delete []ver;
	
	maxarea = 0.0;
	for(i = 0; i < nbn; i++){
	  vcomb[2] = lnbndom[i];
	  area = compute_area_of_triangle(top,vcomb[0],vcomb[1],vcomb[2],out);
	  if(maxarea <= area){
	    maxarea = area;
	  }
	}
	for(i = 0; i < nbn; i++){
	  vcomb[2] = lnbndom[i];
	  area = compute_area_of_triangle(top,vcomb[0],vcomb[1],vcomb[2],out);
	  if(maxarea <= area){
	    nodeidentif[i] = 3;
	  }
	}
	delete []vcomb;
	
	// for(i = 0; i < nbn; i++){
	// 	  if(nodeidentif[i] == 3){
	// 	    fprintf(out,"%ld je corner\n\n",lnbndom[i]);
	// 	  }
	// 	}
      
      }
      if(auxnv == 2){
	ver = new long[2];
	vcomb = new long[3];
	j = 0;
	k = 0;
	for(i = 0; i < nbn; i++){
	  if(nodeidentif[i] == 3){
	    ver[j] = lnbndom[i];
	    j++;
	  }
	  if(nodeidentif[i] == 2){
	    k++;
	  }
	}
	radius = compute_length_of_vector(top,ver[0],ver[1],out);
	//fprintf(out,"radius %lf mezi uzly %ld %ld\n\n",radius,ver[0],ver[1]);
	maxradius = 0.0;
	j = 0;
	contrad = new double[k];
	vcomb[0]=ver[0];
	for(i = 0; i < nbn; i++){
    	  if(nodeidentif[i] == 2){
	    vcomb[1] = lnbndom[i];
	    contrad[j] = compute_length_of_vector(top,vcomb[0],vcomb[1],out);
	    //fprintf(out,"uzel %ld contrad %lf\n\n",lnbndom[i],contrad[j]);
	    if(maxradius <= contrad[j]){
	      maxradius = contrad[j];
	    }
	    j++;
	  }
	}
	//fprintf(out,"maxradius %lf\n\n",maxradius);
	j = 0;
	vcomb[0] = ver[1];
	for(i = 0; i < nbn; i++){
	  if(nodeidentif[i] == 2){
	    //fprintf(out,"uzel %ld contrad %lf maxradius %lf\n\n",lnbndom[i],contrad[j],maxradius);
	    if(maxradius <= contrad[j]){
	      vcomb[1] = lnbndom[i];
	      radius = compute_length_of_vector(top,vcomb[0],vcomb[1],out);
	      //fprintf(out,"radius %lf\n\n",radius);
	      if(radius > 0.05*sizefictdom){
		vcomb[2] = ver[0];
		cosalpha = compute_angle_of_vector_a_and_b(top,vcomb[0],vcomb[1],vcomb[2],out);
		if(cosalpha > 0.034906585 || cosalpha < 3.1066861){
		  nodeidentif[i] = 3;
		  //fprintf(out,"uzel  %ld bude oznacen jako corner\n\n",lnbndom[i]);
		}
	      }
	    }
	    j++;
	  }
	}
	delete []contrad;
	delete []ver;
	delete []vcomb;
      }
    }
    t3=time(NULL);
    //fprintf(out,"cas %ld\n\n",t3-t2);    
    //fprintf(out,"cas celkem %ld\n\n",t3-t1);    

    // rewrite for creation of master graph
    for(i = 0; i < nbn; i++){
      for(j = 0; j < nadjacboundnod[i]; j++){
	k = adjacboundnod[i][j];
	adjacboundnod[i][j] = icnbn[k];
      }
    }
    
    j = 0;
    for(i = 0; i < nbn; i++){
      if(nodeidentif[i] == 3){
	fprintf(out,"%ld is corner node\n\n",lnbndom[i]);
	j++;
      }
    }
    fprintf(out,"Number of potential corner nodes %ld\n\n",j);
    
}



void partop::prepare_master_curves(long *domproc,FILE *out)
{
  long i,j,k,m,a,b;
  long buffsize,maxnv,max,adr1,adr2;
  long *buff,*nvdom,**aux,*naux,**auxnadjac,*help,**vazbanod,**vazbadom;
  MPI_Status stat;
  
  buffsize = maxnbn;
  buff = new long[buffsize];
  
  for(i = 0; i < nbn; i++){
    buff[i] = nodeidentif[i];
  }
  
  // fprintf(out,"\n\nKONTROLA V prepare_master_curves\n\n");
  //   for(j = 0; j < nbn; j++){
  //     fprintf(out,"buff %ld %ld\n",j,nodeidentif[j]);
  //   }
  
  if(myrank == 0){
      
    // vazby
    vazbanod = new long*[tnbn];
    vazbadom = new long*[tnbn];
    
    for(i = 0; i < tnbn; i++){
      vazbanod[i] = new long[icmultip[i]];
      vazbadom[i] = new long[icmultip[i]];
    }
    
    for(i = 0; i < tnbn; i++){
      m = 0;
      for(j = 0; j < nproc; j++){
	for(k = 0; k < nbnd[j]; k++){
	  if(icnbnmas[j][k] == i){
	    vazbanod[i][m] = k;
	    vazbadom[i][m] = j;
	    m++;
	    break;
	  }
	}
      }
    }
    
    //     for(i = 0; i < tnbn; i++){
    //       fprintf(out,"%ld vazbadom na ",i);
    //       for(j = 0; j < icmultip[i]; j++){
    // 	fprintf(out,"  %ld",vazbadom[i][j]);
    //       }
    //       fprintf(out,"\n");
    //     }
    //     for(i = 0; i < tnbn; i++){
    //       fprintf(out,"%ld vazbanod na ",i);
    //       for(j = 0; j < icmultip[i]; j++){
    // 	fprintf(out,"  %ld",vazbanod[i][j]);
    //       }
    //       fprintf(out,"\n");
    //     }
    
    aux = new long* [nproc];
    //  master contribution
    k=domproc[0];
    aux[k] = new long[nbnd[k]];
    for(j = 0; j < nbnd[k]; j++){
      aux[k][j] = buff[j];
    }
    //slaves contributions
    for (i=1;i<nproc;i++){
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      k=domproc[stat.MPI_TAG];
      aux[k] = new long[nbnd[k]];
      for(j = 0; j < nbnd[k]; j++){
	aux[k][j] = buff[j];
      } 
    }
    
    //         for(i = 0; i < nproc; i++){
    //           fprintf(out,"domain %ld\n",i);
    //           for(j = 0; j < nbnd[i]; j++){
    //     	fprintf(out,"%ld\n",aux[i][j]);
    //           }
    //         }
    
    coarseidentif = new long[tnbn];
    
    for(i = 0; i < tnbn; i++){
      // initial import
      coarseidentif[i] = aux[vazbadom[i][0]][vazbanod[i][0]];
      // controlling
      m = 0;
      for(j = 1; j < icmultip[i]; j++){
	if(coarseidentif[i] != aux[vazbadom[i][j]][vazbanod[i][j]]){
	  coarseidentif[i] = aux[vazbadom[i][j]][vazbanod[i][j]];
	  m++;
	}
      }
      if( m != 0){
	fprintf(out,"prepis %ld\n",m);
      }
    }
    
    
    //     for(i = 0; i < tnbn; i++){
    //       fprintf(out,"%ld coarseidentif %ld\n",i,coarseidentif[i]);
    //     }
    
    nvdom = new long[nproc];
    maxnv = 0;
    for(i = 0; i < nproc; i++){
      nvdom[i] = 0;
      for(j = 0; j < nbnd[i]; j++){
	if(aux[i][j] == 3){
	  nvdom[i]++;
	}
      }
      if(nvdom[i] >= maxnv){
	maxnv = nvdom[i];
      }
      delete []aux[i];
    }
    delete []aux;
    
    for(i= 0; i < nproc; i++){
      fprintf(out,"domain %ld has %ld corner nodes\n",i,nvdom[i]);
    }
    fprintf(out,"Maximal number of corner nodes on one subdomain %ld\n",maxnv);
    
  }
  else{
    MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  delete []buff;
  
  
  buffsize = nbn;
  for(i = 0; i < nbn; i++){
    buffsize += nadjacboundnod[i];
  }

  //fprintf(out,"buffsize = %ld\n",buffsize);

  
  if (myrank==0){
    //  master contribution
    max = buffsize;
    
    for (i=1;i<nproc;i++){
      MPI_Recv (&k,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      
      //  slave contributions
      if (max<k){
	max=k;
      }
    }
    
    buffsize = max;
    
    for (i=1;i<nproc;i++){
      MPI_Send (&buffsize,1,MPI_LONG,i,myrank,MPI_COMM_WORLD);
    }
  }
  // ********
  // slaves
  // ********
  else{
    MPI_Send (&buffsize,1,MPI_LONG,0,myrank,MPI_COMM_WORLD);
    MPI_Recv (&buffsize,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
  }
  MPI_Barrier (MPI_COMM_WORLD);
    
//   fprintf(out,"buffsize = %ld\n",buffsize);

  buff = new long[buffsize];
  
  buff[0] = 0;
  for(i = 1; i < nbn; i++){
    buff[i] = buff[i-1]+nadjacboundnod[i-1];
  }
  
  m = nbn;
  for(i = 0; i < nbn; i++){
    for(j = 0; j < nadjacboundnod[i]; j++){
      buff[m]  = adjacboundnod[i][j];
      m++;
    }
  }
  for(i = m; i < buffsize; i++){
    buff[m] = -1;
    m++;
  }
  
  //   for(i = 0; i < buffsize; i++){
  //     fprintf(out,"kont buff %ld %ld\n",i,buff[i]);
  //   }
  
  //delete nadjacboundnod
  for(i = 0; i < nbn; i++){
    delete []adjacboundnod[i];
  }
  delete []adjacboundnod;
  delete []nadjacboundnod;
  
  if (myrank == 0){
    auxnadjac = new long*[nproc];
    aux = new long*[nproc];
    naux = new long[nproc];
    // master contribution
    k=domproc[0];
    auxnadjac[k] = new long[nbnd[k]+1];
    for(j = 0; j < nbnd[k]; j++){
      auxnadjac[k][j] = buff[j];
    }
    naux[k] = 0;
    m = 0;
    for(j = nbnd[k]; j < buffsize; j++){
      if(buff[j] != -1){
	naux[k]++;
      }
      else{
	m++;
      }
      if(m == 2){
	break;
      }
    }
    //fprintf(out,"%ld naux = %ld %ld\n",k,naux[k],j);
    aux[k] = new long[naux[k]];
    auxnadjac[k][nbnd[k]] = naux[k];
    for(j = 0; j < naux[k]; j++){
      aux[k][j] = buff[j+nbnd[k]];
    }
    
    
    //slaves contributions
    for (i=1;i<nproc;i++){
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      k=domproc[stat.MPI_TAG];
      auxnadjac[k] = new long[nbnd[k]+1];
      for(j = 0; j < nbnd[k]; j++){
	auxnadjac[k][j] = buff[j];
      }
      naux[k] = 0;
      m = 0;
      for(j = nbnd[k]; j < buffsize; j++){
	if(buff[j] != -1){
	  naux[k]++;
	}
	else{
	  m++;
	}
	if(m == 2){
	  break;
	}
      }
      //fprintf(out,"%ld naux = %ld %ld\n",k,naux[k],j);
      aux[k] = new long[naux[k]];
      auxnadjac[k][nbnd[k]] = naux[k];
      for(j = 0; j < naux[k]; j++){
	aux[k][j] = buff[j+nbnd[k]];
      }
    }
    
    //     for(i = 0; i < nproc; i++){
    //       fprintf(out,"domain %ld %ld:",i,naux[i]);
    //       for(j = 0; j < nbnd[i]; j++){
    // 	fprintf(out,"   %ld",auxnadjac[i][j]);
    //       }
    //       fprintf(out,"\n");
    //     }
    
    //     for(i = 0; i < nproc; i++){
    //       fprintf(out,"domain %ld\n",i);
    //       for(j = 0; j < nbnd[i]; j++){
    // 	adr1=auxnadjac[i][j];
    // 	adr2=auxnadjac[i][j+1];
    // 	fprintf(out,"uzel %ld pocet sousedu: %ld ",j,adr2-adr1);
    // 	for(k = adr1; k < adr2; k++){
    // 	  fprintf(out," %ld",aux[i][k]);
    // 	}
    // 	 fprintf(out,"\n");
    //       }
    //     }
    
    max = 0;
    for(i = 0; i < tnbn; i++){
      b = 0;
      for(j = 0; j < icmultip[i]; j++){
      adr1=auxnadjac[vazbadom[i][j]][vazbanod[i][j]];
      adr2=auxnadjac[vazbadom[i][j]][vazbanod[i][j]+1];
      a = adr2-adr1;
      b+=a;
      }
      if(max < b){
	max = b;
      }
    }
    //fprintf(out,"max je %ld\n",max);    
    
    coarsenadjac = new long[tnbn];
    coarseadjac = new long*[tnbn];
    help = new long[max];
    for(i = 0; i < tnbn; i++){
      //fprintf(out,"uzel  %ld\n",i);    
      b = 0;
      for(j = 0; j < icmultip[i]; j++){
	adr1=auxnadjac[vazbadom[i][j]][vazbanod[i][j]];
	adr2=auxnadjac[vazbadom[i][j]][vazbanod[i][j]+1];
	for(k = adr1; k < adr2; k++){
	  help[b] = aux[vazbadom[i][j]][k];
	  b++;
	}
      }
      //fprintf(out,"b je %ld\n\n",b);    
      //print
      //for(j = 0; j < b; j++){
      //fprintf(out,"    %ld",help[j]);    
      //}
      //fprintf(out,"\n"); 
      if(b == 2){
	if(help[0] != help[1]){
	  fprintf(out,"problem u uzlu %ld\n",i); 
	}
	else{
	  coarsenadjac[i] = 1;
	  coarseadjac[i] = new long[coarsenadjac[i]];
	  coarseadjac[i][0] = help[0];
	  //fprintf(out,"ps %ld  s %ld\n",coarsenadjac[i],coarseadjac[i][0]); 
	}
      }
      else{
	// bubble sorting
	for(j = b - 1; j > 0; j--){
	  for(k = 0; k < j; k++){
	    if(help[k] > help[k+1]){
	      m = help[k];   
	      help[k] = help[k+1];
	      help[k+1] = m;
	    }
	  } 
	}
	
	// print
	//for(j = 0; j < b; j++){
	//fprintf(out,"    %ld",help[j]);    
	//}
	//fprintf(out,"\n"); 
	
	coarsenadjac[i] = 1;
	for(j = 1; j < b; j++){
	  if(help[j] != help[j-1]){
	    coarsenadjac[i]++;
	  }
	}
	//fprintf(out,"s %ld\n",coarsenadjac[i]); 
	coarseadjac[i] = new long[coarsenadjac[i]];
	coarseadjac[i][0] = help[0];
	m = 1;
	for(j = 1; j < b; j++){
	  if(help[j] != help[j-1]){
	    coarseadjac[i][m] = help[j];
	    m++;
	  }
	}
      }
    }
    delete []help;
    
//     for(i = 0; i < tnbn; i++){
//       fprintf(out,"uzel %ld ma  %ld sousedu:",i,coarsenadjac[i]); 
//       for(j = 0; j < coarsenadjac[i]; j++){
//         fprintf(out,"    %ld",coarseadjac[i][j]);    
//       }
//       fprintf(out,"\n");    
//     }
    
    for(i = 0; i < nproc; i++){
      delete []aux[i];
      delete []auxnadjac[i];
    }
    delete []aux;
    delete []naux;
    delete []auxnadjac;
    
    for(i = 0; i < tnbn; i++){
      delete []vazbadom[i];
      delete []vazbanod[i];
    }
    
  }
  else{
    MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  delete [] buff;
    
}


long partop::compute_number_of_combination(long n, long k)
{
  long i;
  long denom,nom,diff;
  long combination;
  
  diff = n - k;
  nom = 1;
  for(i = diff+1; i < n+1; i++){
    nom *= i;
  }
  denom = 1;
  for(i = 2; i < k+1; i++){
    denom *= i;
  }
  combination = nom/denom;
  
  return(combination);
}


void partop::check_geometry_of_corner_nodes(gtopology *top,FILE *out)
{
  long i,j,k,m;
  double length,cosalpha,length1,length2;
  long combl,comba,v;
  long *vcomb,*ver,*lengthok,*cosalphaok,*pointver;

  //fprintf(out,"kontrola v geometrii\n");
  
  vcomb = new long[3];
  
  v = 0 ;
  for(i = 0; i < nbn; i++){
    if(nodeidentif[i] == 3){
      v++;
    }
  }
  ver = new long[v];
  pointver = new long[v];
  j = 0 ;
  for(i = 0; i < nbn; i++){
    if(nodeidentif[i] == 3){
      ver[j] = lnbndom[i];
      pointver[j] = i;
      j++;
    }
  }
  //fprintf(out,"kontrola v geometrii %ld\n",v);
  // check length between two corners
  //combl = compute_number_of_combination(v,2);


  k = 0;
  for(i = 0; i < v; i++){
    if(nodeidentif[pointver[i]] == 3){
      vcomb[0] = ver[i];
      for(j = i+1; j < v; j++){
	if(nodeidentif[pointver[j]] == 3){
	  //fprintf(out,"ver: %ld a %ld:",ver[i],ver[j]);
	  vcomb[1] = ver[j];
	  length=compute_length_of_vector(top,vcomb[0],vcomb[1],out);
	  if(length > 0.05*sizefictdom ){
	    //lengthok[k] = 1;
	    //fprintf(out,"delka je OK\n");
	  }
	  else{
	    //lengthok[k] = 0;
	    nodeidentif[pointver[j]]=2;
	    //fprintf(out,"delka je kratka %ld bude vymazan ze seznamu vrcholu\n",i,lnbndom[pointver[j]]);
	  }
	  k++;
	}
      }
    }
  }
  //for(i = 0; i < combl; i++){
  //if(lengthok[i] == 1){
  //  fprintf(out,"delka %ld je OK\n",i);
  //}
  //else{
  //  fprintf(out,"delka %ld je kratka\n",i);
  //}
  //}
  
  // check angle between two vectors
  //comba = compute_number_of_combination(v,3);
  //fprintf(out,"pocet kombinaci = %ld\n",comba);
  
  m = 0;
  lengthok = new long[3];
  cosalphaok = new long[3];
  for(i = 0; i < v; i++){
    if(nodeidentif[pointver[i]] == 3){
      vcomb[0] = ver[i];
      for(j = i+1; j < v; j++){
	if(nodeidentif[pointver[j]] == 3){
	  vcomb[1] = ver[j];
	  length=compute_length_of_vector(top,vcomb[0],vcomb[1],out);
	  for(k = j+1; k < v; k++){
	    if(nodeidentif[pointver[k]] == 3){
	      vcomb[2]= ver[k];
	      //fprintf(out,"ver: %ld  %ld %ld\n",ver[i],ver[j],ver[k]);
	      cosalpha=compute_angle_of_vector_a_and_b(top,vcomb[0],vcomb[1],vcomb[2],out);
	      if(cosalpha > 0.034906585 || cosalpha < 3.1066861){
		//cosalphaok[m] = 1;
		//fprintf(out,"uzly na primce nelezi\n",i);
	      }
	      else{
		//cosalphaok[m] = 0;
		//fprintf(out,"uzly na primce lezi\n",i);
		// 1 a 2
		length1=compute_length_of_vector(top,vcomb[0],vcomb[1],out);
		// 1 a 3
		length2=compute_length_of_vector(top,vcomb[0],vcomb[2],out);
		if(length1 >= length2){
		  nodeidentif[pointver[k]]=2;
		}
		else{
		  nodeidentif[pointver[j]]=2;
		}
	      }
	      m++;
	    }
	  }
	}
      }
    }
  }
  //for(i = 0; i < comba; i++){
  //if(cosalphaok[i] == 1){
  //  fprintf(out,"uzly na primce nelezi\n",i);
  //}
  //else{
  //  fprintf(out,"uzly na primce lezi\n",i);
  //}
  
//   j = 0;
//   for(i = 0; i < nbn; i++){
//     if(nodeidentif[i] == 3){
//       fprintf(out,"uzel %ld je vrchol\n",lnbndom[i]);
//       j++;
//     }
//   }
//   fprintf(out,"pocet vrcholu je %ld\n",j);

  delete []vcomb;
  //delete []cosalphaok;
  //delete []lengthok;
  delete []ver;
  delete []pointver;
}



void partop::check_triangle(gtopology *top,FILE *out)
{
  long i,j,k,m;
  double *length,*angle;
  long combl,comba,v;
  long *vcomb,*ver,*lengthok,*cosalphaok,*pointver;
  long stop;

  //fprintf(out,"check triangle\n");
  
  vcomb = new long[3];
  
  v = 0 ;
  for(i = 0; i < nbn; i++){
    if(nodeidentif[i] == 3){
      v++;
    }
  }
  ver = new long[v];
  pointver = new long[v];
  j = 0 ;
  for(i = 0; i < nbn; i++){
    if(nodeidentif[i] == 3){
      ver[j] = lnbndom[i];
      pointver[j] = i;
      j++;
    }
  }
  
  length = new double[3];
  angle = new double[3];
  for(i = 0; i < v; i++){
    stop = 0;
    if(nodeidentif[pointver[i]] == 3){
    vcomb[0] = ver[i];
      for(j = i+1; j < v; j++){
	stop = 0;
	if(nodeidentif[pointver[j]] == 3){
	  vcomb[1] = ver[j];
	  length[2]= compute_length_of_vector(top,vcomb[0],vcomb[1],out);
	  for(k = j+1; k < v; k++){
	    if(nodeidentif[pointver[k]] == 3){
	      vcomb[2] = ver[k];
	      //fprintf(out,"uzly : %ld %ld %ld\n",vcomb[0],vcomb[1],vcomb[2]);
	      length[1] = compute_length_of_vector(top,vcomb[0],vcomb[2],out);
	      length[0] = compute_length_of_vector(top,vcomb[1],vcomb[2],out);
	      angle[0] = compute_angle_of_vector_a_and_b(top,vcomb[0],vcomb[1],vcomb[2],out);
	      length[0] = compute_length_of_vector(top,vcomb[1],vcomb[2],out);
	      angle[1] = compute_angle_of_vector_a_and_b(top,vcomb[1],vcomb[0],vcomb[2],out);
	      angle[2] = compute_angle_of_vector_a_and_b(top,vcomb[2],vcomb[0],vcomb[1],out);
	      //test uhlu
	      //if(angle[0]+angle[1]+angle[2] > 3.1416 || angle[0]+angle[1]+angle[2] < 3.1414){
		//fprintf(out,"divne uhly\n");
	      //}
	      //test print
	      //fprintf(out,"delky : %le %le %le\n",length[0],length[1],length[2]);
	      //fprintf(out,"uhly : %lf %lf %lf\n",angle[0]*180/3.141592653589793238462643,angle[1]*180/3.141592653589793238462643,angle[2]*180/3.141592653589793238462643);
	      // vyhodnoceni
	      // kontrola primky
	      //if(fabs(cos(angle[0])) > 0.99){
	      if((angle[0] < 0.034906585 || angle[0] > 3.1066861) || (angle[1] < 0.034906585 || angle[1] > 3.1066861) || (angle[2] < 0.034906585 || angle[2] > 3.1066861)){
	      // lezi na primce
		if(length[1] >= length[2]){
		  nodeidentif[pointver[j]]=2;
		  stop = 1;
		  //fprintf(out,"uzly lezi na primce, bude vymazan uzel %ld\n",lnbndom[pointver[j]]);
		}
		else{
		  nodeidentif[pointver[k]]=2;
		  //fprintf(out,"uzly lezi na primce, bude vymazan uzel %ld\n",lnbndom[pointver[k]]);
		}
	      }
	      
	      // toto neni dotazeno do konce
	      // kontrola uhlu 
	      if(angle[0] < (3.141592653589793238462643*5)/180 && angle[1] < (3.141592653589793238462643*5)/180){
		nodeidentif[pointver[k]]=2;
		//fprintf(out,"male uhly 0 a 1 bude vymazan uzel %ld\n",lnbndom[pointver[k]]);
	      }
	      if(angle[0] < (3.141592653589793238462643*5)/180 && angle[2] < (3.141592653589793238462643*5)/180 ){
		nodeidentif[pointver[j]]=2;
		//fprintf(out,"male uhly 0 a 2 bude vymazan uzel %ld\n",lnbndom[pointver[j]]);
		
	      }
	      if(angle[1] < (3.141592653589793238462643*5)/180 && angle[2] < (3.141592653589793238462643*5)/180){
		nodeidentif[pointver[i]]=2;
		//fprintf(out,"male uhly 1 a 2 bude vymazan uzel %ld\n",lnbndom[pointver[i]]);
	      }
	      
	      // kontrola delek
	      if(length[0]< 0.05*sizefictdom ){
		if(length[1] > 0.05*sizefictdom && length[1] > length[2]){
		  nodeidentif[pointver[j]]=2;
		  //fprintf(out,"mala vzdalenost mezi uzly bude vymazan uzel %ld\n",lnbndom[pointver[j]]);
		  stop = 1;
		}
		if(length[2] > 0.05*sizefictdom && length[2] > length[1]){
		  //fprintf(out,"mala vzdalenost mezi uzly bude vymazan uzel %ld\n",lnbndom[pointver[k]]);
		  nodeidentif[pointver[k]]=2;
		}
	      }
	      if(length[1]< 0.05*sizefictdom ){
		if(length[0] > 0.05*sizefictdom && length[0] > length[2]){
		  //fprintf(out,"mala vzdalenost mezi uzly, bude vymazan uzel %ld\n",lnbndom[pointver[i]]);
		  nodeidentif[pointver[i]]=2;
		  stop = 2;
		}
		if(length[2] > 0.05*sizefictdom && length[2] > length[0]){
		  nodeidentif[pointver[k]]=2;
		  //fprintf(out,"mala vzdalenost mezi uzly, bude vymazan uzel %ld\n",lnbndom[pointver[k]]);
		}
	      }
	      if(length[2]< 0.05*sizefictdom ){
		if(length[0] > 0.05*sizefictdom && length[0] > length[1]){
		  nodeidentif[pointver[i]]=2;
		  //fprintf(out,"mala vzdalenost mezi uzly, bude vymazan uzel %ld\n",lnbndom[pointver[i]]);
		  stop = 2;
		}
		if(length[1] > 0.05*sizefictdom && length[1] > length[0]){
		  nodeidentif[pointver[j]]=2;
		  //fprintf(out,"mala vzdalenost mezi uzly, bude vymazan uzel %ld\n",lnbndom[pointver[j]]);
		  stop = 1;
		}
	      }
	      if(stop == 1){
		break;
	      }
	    }
	  }
	  if(stop == 2){
	    break;
	  }
	}
      }
    }
  }

  j = 0;
  for(i = 0; i < nbn; i++){
    //fprintf(out,"uzel %ld nodeidentif %ld\n",lnbndom[i],nodeidentif[i]);
    if(nodeidentif[i] == 3){
      fprintf(out,"Node %ld is corner nodes\n",lnbndom[i]);
      j++;
    }
  }
  fprintf(out,"Number of corner node is %ld, after check_triangle\n",j);

  delete []vcomb;
  delete []ver;
  delete []pointver;
  
  delete []length;
  delete []angle;
}



void partop::compute_statistics_of_multiplicity(long *domproc,FILE *out)
{
  long i,j,k;
  long buffsize;
  long *buff;
  MPI_Status stat;
  
  fprintf(out,"\n\n\nStatistics:\n");
  
  maxmultip = 0;
  for(i = 0; i < nn; i++){
    if(nodmultip[i] > maxmultip){
      maxmultip = nodmultip[i];
      }
  }
  
  if(myrank == 0){
    //  array containing number of nodes on subdomains
    nstatdom = new long [nproc];
    statdom = new long*[nproc];
    //  master contribution
    j=domproc[0];
    nstatdom[j]=maxmultip;
    statdom[j] = new long[nstatdom[j]];
    for (i = 1; i < nproc; i++){
      MPI_Recv (&k,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      
      //  slave contributions
      j=domproc[stat.MPI_TAG];
      if (maxmultip < k){
	maxmultip = k;
      }
      nstatdom[j] = k;
      statdom[j] = new long[maxmultip];
    }
    
    fprintf(out,"\n\n\nMaximal multiplicity  %ld\n",maxmultip);
    for (i = 0; i < nproc; i++){
      fprintf(out,"\n\nDomain %ld has maximal multiplicity %ld\n",i,nstatdom[i]);
    }
    
    
    
    for (i=1;i<nproc;i++){
      MPI_Send (&maxmultip,1,MPI_LONG,i,myrank,MPI_COMM_WORLD);
    }
  }
  else{
    MPI_Send (&maxmultip,1,MPI_LONG,0,myrank,MPI_COMM_WORLD);
    MPI_Recv (&maxmultip,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  
  buffsize = maxmultip;
  buff = new long[buffsize];
  
  for(j = 0; j < maxmultip; j++){
    buff[j] = 0;
    for(i = 0; i < nn; i++){
      if(nodmultip[i] == j+1){
	buff[j]++;
      }
    }
    //fprintf(out,"Pocet uzlu s multiplicitou %ld je %ld\n",j+1,buff[j]);
  }
  
  if(myrank == 0){
    k=domproc[0];
    for(j = 0; j < maxmultip; j++){
      statdom[k][j] = buff[j];
    }
    
    for (i = 1; i < nproc; i++){
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
        
      //  slave contributions
      k=domproc[stat.MPI_TAG];
      for(j = 0; j < maxmultip; j++){
      statdom[k][j] = buff[j];
      }
    }
    for (i = 0; i < nproc; i++){
      fprintf(out,"\n\nDomain %ld:\n",i);
      for(j= 0; j < maxmultip; j++){
	fprintf(out,"Number of nodes with multiplicity %ld is %ld\n",j+1,statdom[i][j]);
      }
    }
  }
  else{
    MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  delete []buff;
}


void partop::select_corner_nodes_on_master(long *domproc,FILE *out)
{
  long i,j,k,m;
  long **vazbanod,**vazbadom;
  long *auxnv,*buff;
  long acticmultip,stop,buffsize;
  MPI_Status stat;
  
  
  buffsize = maxnbn;
  buff = new long[buffsize];
  
  if(myrank == 0){
    
     // vazby
    vazbanod = new long*[tnbn];
    vazbadom = new long*[tnbn];
    
    for(i = 0; i < tnbn; i++){
      vazbanod[i] = new long[icmultip[i]];
      vazbadom[i] = new long[icmultip[i]];
    }
    
    for(i = 0; i < tnbn; i++){
      m = 0;
      for(j = 0; j < nproc; j++){
       for(k = 0; k < nbnd[j]; k++){
	 if(icnbnmas[j][k] == i){
	   vazbanod[i][m] = k;
	   vazbadom[i][m] = j;
	   m++;
	   break;
	 }
       }
      }
    }
    auxnv = new long[nproc];
    for(i = 0; i < nproc; i++){
      auxnv[i]=0;
    }
    // indentifikace na masteru 
    coarseidentif = new long[tnbn];
    // vyber uzlu s nejvetsi icmultiplicitou
    for(i = 0; i < tnbn; i++){
      if(icmultip[i] == maxmultip){
	coarseidentif[i] = 3;
	for(j = 0; j < icmultip[i]; j++){
	  auxnv[vazbadom[i][j]]++;
	}
      }
      else{
	coarseidentif[i] = 2;
      }
    }
    
    //for(i = 0; i < nproc; i++){
    //fprintf(out,"auxnv[%ld] = %ld\n",i,auxnv[i]);
    //}

    
    for(i = 0; i < nproc; i++){
      if(auxnv[i] < 3){
	stop = 0;
	acticmultip = maxmultip-1;
	while (stop != 1){
	  if(statdom[i][acticmultip-1] != 0){
	    for(j = 0; j < nbnd[i]; j++){
	      if(icmultip[icnbnmas[i][j]] == acticmultip){
		coarseidentif[icnbnmas[i][j]] = 3;
		for(k = 0; k < icmultip[icnbnmas[i][j]]; k++){
		  auxnv[vazbadom[icnbnmas[i][j]][k]]++;
		}
	      }
	    }
	    if(auxnv[i] >= 3){
		  stop = 1;
	    }
	    else{
	      stop = 0;
	      acticmultip--;
	    }
	  }
	  else{
	    acticmultip--;
	  }
	}
      }
    }
    
    // for(i = 0; i < nproc; i++){
    //       fprintf(out,"auxnv[%ld] = %ld\n",i,auxnv[i]);
    //     }
    
    j = 0;
    k = 0;
    for(i = 0; i < tnbn; i++){
      if(coarseidentif[i] == 3){
	j++;
      }
      else{
	k++;
      }
    }
    // fprintf(out,"Number of corner nodes in whole problem is %ld",j);
    //     fprintf(out,"Number of boundary nodes in whole problem is %ld",k);
    
    for(j = 1; j < nproc; j++){
      for(k = 0; k < nbnd[j]; k++){
	buff[k] = coarseidentif[icnbnmas[j][k]];
      }
      MPI_Send (buff,buffsize,MPI_LONG,j,myrank,MPI_COMM_WORLD);
    }
    
    nodeidentif = new long[nbn];
    for(j = 0; j < nbnd[0]; j++){
      nodeidentif[j] = coarseidentif[icnbnmas[0][j]];;
    }
    
    for(i = 0; i < tnbn; i++){
      delete []vazbadom[i];
      delete []vazbanod[i];
    }
    delete []vazbadom;
    delete []vazbanod;
    delete []auxnv;
  }
  else{
    MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    
    nodeidentif = new long[nbn];
    for(i = 0; i < nbn; i++){
      nodeidentif[i] = buff[i];
    }
  }
  
  //   for (i = 0; i < nbn; i++){
  //     if(nodeidentif[i] == 2) fprintf(out,"\n node number %ld is edge node",lnbndom[i]+1);
  //     if(nodeidentif[i] == 3) fprintf(out,"\n node number %ld is vertex node",lnbndom[i]+1);
  //   }
  //   fflush(out);
  //   fprintf(out,"\n\n");
  
  delete []buff;
}

void partop::boundary_surface_detection(gtopology *top,FILE *out)
{
  
  long i,j,k,m,n,l,a;
  long auxnsurf,nsurf,nnsurf;
  long *aux;
  long *auxnnsurf,**auxsurfnodes,*duplsurf;
  
  
  // number of all surf in mesh
  auxnsurf = 0;
  // pomocne pole
  aux = new long[8];
  for(i = 0; i < ne; i++){
    // number of surfge on i-th element - gtopology
    nsurf = top->give_nsurf(i);
    nnsurf = top->give_nnsurf(i);
    for(j = 0; j < nsurf; j++ ){
      top->give_surf_nodes(i,j,aux);
      m = 0;
      for(l = 0; l < nnsurf; l++){
	  if(nodmultip[aux[l]] > 1){
	    m++;
	  }
      }
      switch(nnsurf){
      case 3:
      case 6:
	if(m > 2){
	  auxnsurf++;
	}
	break;
      case 4:
      case 8:
	if(m > 3){
	  auxnsurf++;
	}
	break;
      }
    }
    
  }

  fprintf(out,"pocet ploch %ld\n",auxnsurf);
  
  auxnnsurf = new long[auxnsurf];
  auxsurfnodes = new long*[auxnsurf];
  duplsurf = new long[auxnsurf];
  k = 0;
  for(i = 0; i < ne; i++){
    nsurf = top->give_nsurf(i);
    nnsurf = top->give_nnsurf(i);
    for(j = 0; j < nsurf; j++ ){
      top->give_surf_nodes(i,j,aux);
      m = 0;
      for(l = 0; l < nnsurf; l++){
	if(nodmultip[aux[l]] > 1){
	  m++;
	}
      }
      switch (nnsurf){
      case 3:
      case 6:
	if(m > 2){
	  auxnnsurf[k] = nnsurf;
	  auxsurfnodes[k] = new long[nnsurf];
	  for(l = 0; l < nnsurf; l++){
	    auxsurfnodes[k][l] = aux[l];
	  }
	  for(l = nnsurf - 1; l > 0; l--){
	    for(n = 0; n < l; n++){
	      if(auxsurfnodes[k][n] > auxsurfnodes[k][n+1]){
		a = auxsurfnodes[k][n];   
		auxsurfnodes[k][n] = auxsurfnodes[k][n+1];
		auxsurfnodes[k][n+1] = a;
	      }
	    } 
	  }
	  duplsurf[k] = 0;
	  k++;
	}
	break;
      case 4:
      case 8:
	if(m > 3){
	  auxnnsurf[k] = nnsurf;
	  auxsurfnodes[k] = new long[nnsurf];
	  for(l = 0; l < nnsurf; l++){
	    auxsurfnodes[k][l] = aux[l];
	  }
	  for(l = nnsurf - 1; l > 0; l--){
	    for(n = 0; n < l; n++){
	      if(auxsurfnodes[k][n] > auxsurfnodes[k][n+1]){
		a = auxsurfnodes[k][n];   
		auxsurfnodes[k][n] = auxsurfnodes[k][n+1];
		auxsurfnodes[k][n+1] = a;
	      }
	    } 
	  }
	  duplsurf[k] = 0;
	  k++;
	}
	break;
      }
    }
  }
  delete []aux;
  //fprintf(out,"pocet hran %ld\n",auxnsurf);
    
    
  // kontrolni tisk
  //for(i = 0; i < auxnsurf; i++){
  //fprintf(out,"hrana %ld ma uzly:",i);
  //for(j = 0; j < auxnnsurf[i]; j++ ){
  //  fprintf(out,"   %ld",auxsurfnodes[i][j]);
  //}
  //fprintf(out,"\n");
  //}
  
    // hledani duplikovanych hran na hranici
    for(i = 0; i < auxnsurf; i++){
      for(j = i+1; j < auxnsurf; j++){
	if(auxsurfnodes[i][0] == auxsurfnodes[j][0]){
	  if(auxnnsurf[i] == 3 && auxnnsurf[j] == 3){
	    if(auxsurfnodes[i][1] == auxsurfnodes[j][1]){
	      if(auxsurfnodes[i][2] == auxsurfnodes[j][2]){
		duplsurf[j]++;
		//duplsurf[i]++;
	      }
	    }
	  }
	  if(auxnnsurf[i] == 4 && auxnnsurf[i] == 4){
	    if(auxsurfnodes[i][1] == auxsurfnodes[j][1]){
	      if(auxsurfnodes[i][2] == auxsurfnodes[j][2]){
		if(auxsurfnodes[i][3] == auxsurfnodes[j][3]){
		  duplsurf[j]++;
		  //duplsurf[i]++;
		}
	      }
	    }
	  }
	  if(auxnnsurf[i] == 6 && auxnnsurf[i] == 6){
	    if(auxsurfnodes[i][1] == auxsurfnodes[j][1]){
	      if(auxsurfnodes[i][2] == auxsurfnodes[j][2]){
		if(auxsurfnodes[i][3] == auxsurfnodes[j][3]){
		  if(auxsurfnodes[i][4] == auxsurfnodes[j][4]){
		    if(auxsurfnodes[i][5] == auxsurfnodes[j][5]){
		      duplsurf[j]++;
		      //duplsurf[i]++;
		    }
		  }
		}
	      }
	    }
	  }
	  if(auxnnsurf[i] == 8 && auxnnsurf[i] == 8){
	    if(auxsurfnodes[i][1] == auxsurfnodes[j][1]){
	      if(auxsurfnodes[i][2] == auxsurfnodes[j][2]){
		if(auxsurfnodes[i][3] == auxsurfnodes[j][3]){
		  if(auxsurfnodes[i][4] == auxsurfnodes[j][4]){
		    if(auxsurfnodes[i][5] == auxsurfnodes[j][5]){
		      if(auxsurfnodes[i][6] == auxsurfnodes[j][6]){
			if(auxsurfnodes[i][7] == auxsurfnodes[j][7]){
			  duplsurf[j]++;
			  //duplsurf[i]++;
			}
		      }
		    }
		  }
		}
	      }
	    }
	  }
	}
      }
    }
    
    // kontrolni tisk
    for(i = 0; i < auxnsurf; i++){
      fprintf(out,"%ld duplsurf %ld\n",i,duplsurf[i]);
    }
    k = 0;
    for(i = 0; i < auxnsurf; i++){
      if(duplsurf[i] == 0){
	k++;
      }
    }
    
    
    long auxnadjacboundnod;
    long max;
    max = 0;
    for(i = 0; i < nbn; i++){
     auxnadjacboundnod = 0;
      for(j = 0; j < auxnsurf; j++){
	if(duplsurf[j] == 0) {
	  for(k = 0; k < auxnnsurf[j]; k++){
	    if(auxsurfnodes[j][k] == lnbndom[i]){
	      auxsurfnodes[j][k] = i;
	      auxnadjacboundnod+=(auxnnsurf[j]-1);
	    }
	  }
	  if(auxnadjacboundnod > max){
	    max = auxnadjacboundnod;
	  }
	}
      } 
    }
    
    fprintf(out,"max je %ld\n",max);
      
    
    long *auxadjacboundnod;
    auxadjacboundnod = new long[max];
    nadjacboundnod = new long[nbn];
    adjacboundnod = new long *[nbn];
    for(i = 0; i < nbn; i++){
      m = 0;
      auxnadjacboundnod = 0;
      for(j = 0; j < auxnsurf; j++){
	if(duplsurf[j] == 0) {
	  for(k = 0; k < auxnnsurf[j]; k++){
	    if(auxsurfnodes[j][k] == i){
	      for(a = 0; a < auxnnsurf[j]; a++){
		if(auxsurfnodes[j][a] != i){
		  auxadjacboundnod[m] = auxsurfnodes[j][a];
		  m++;
		  auxnadjacboundnod++;
		}
	      }
	      break;
	    }
	  }
	}
      }
      for(l = auxnadjacboundnod - 1; l > 0; l--){
	for(n = 0; n < l; n++){
	  if( auxadjacboundnod[n] >  auxadjacboundnod[n+1]){
	    a =  auxadjacboundnod[n];   
	    auxadjacboundnod[n] =  auxadjacboundnod[n+1];
	    auxadjacboundnod[n+1] = a;
	  }
	} 
      }
      nadjacboundnod[i] = 1;
      for(l = 1; l < auxnadjacboundnod; l++){
	if(auxadjacboundnod[l] != auxadjacboundnod[l-1]){
	  nadjacboundnod[i]++;
	}
      }
      adjacboundnod[i] = new long[nadjacboundnod[i]];
      adjacboundnod[i][0] = auxadjacboundnod[0];
      m = 1;
      for(l = 1; l < auxnadjacboundnod; l++){
	if(auxadjacboundnod[l] != auxadjacboundnod[l-1]){
	  adjacboundnod[i][m] = auxadjacboundnod[l];
	  m++;
	}
      }
    }
    delete []auxadjacboundnod;
    
    // kontrolni tisk
    for(i = 0; i < nbn; i++){
      fprintf(out,"uzel %ld ma  %ld sousedu: ",i,nadjacboundnod[i]);
      for(j = 0; j < nadjacboundnod[i]; j++){
	fprintf(out,"   %ld",adjacboundnod[i][j]);
      }
      fprintf(out,"\n");
    }
    for(i = 0; i < nbn; i++){
      fprintf(out,"uzel %ld ma icnbn %ld\n",i,icnbn[i]);
    }


}

void partop::send_and_check_corner_nodes(long *domproc,long *ltg,FILE *out)
{
  long i,j,k,m,a,b;
  long buffsize,maxnv,max,adr1,adr2;
  long *buff,*nvdom,**aux,**vazbanod,**vazbadom,*naux,**auxnadjac,*help;
  MPI_Status stat;
  
  buffsize = maxnbn;
  buff = new long[buffsize];
  
  for(i = 0; i < nbn; i++){
    buff[i] = nodeidentif[i];
  }
  
//   for(j = 0; j < nbn; j++){
//     fprintf(out,"buff %ld %ld\n",j,nodeidentif[j]);
//   }
  
  if(myrank == 0){
    
    // vazby
    vazbanod = new long*[tnbn];
    vazbadom = new long*[tnbn];
    
    for(i = 0; i < tnbn; i++){
      vazbanod[i] = new long[icmultip[i]];
      vazbadom[i] = new long[icmultip[i]];
    }
    
    for(i = 0; i < tnbn; i++){
      m = 0;
      for(j = 0; j < nproc; j++){
	for(k = 0; k < nbnd[j]; k++){
	  if(icnbnmas[j][k] == i){
	    vazbanod[i][m] = k;
	    vazbadom[i][m] = j;
	    m++;
	    break;
	  }
	}
      }
    }
    
//     for(i = 0; i < tnbn; i++){
//       fprintf(out,"%ld vazbadom na ",i);
//       for(j = 0; j < icmultip[i]; j++){
// 	fprintf(out,"  %ld",vazbadom[i][j]);
//       }
//       fprintf(out,"\n");
//     }
//     for(i = 0; i < tnbn; i++){
//       fprintf(out,"%ld vazbanod na ",i);
//       for(j = 0; j < icmultip[i]; j++){
// 	fprintf(out,"  %ld",vazbanod[i][j]);
//       }
//       fprintf(out,"\n");
//     }
    
    aux = new long* [nproc];
    //  master contribution
    k=domproc[0];
    aux[k] = new long[nbnd[k]];
    for(j = 0; j < nbnd[k]; j++){
      aux[k][j] = buff[j];
    }
    //slaves contributions
    for (i=1;i<nproc;i++){
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      k=domproc[stat.MPI_TAG];
      aux[k] = new long[nbnd[k]];
      for(j = 0; j < nbnd[k]; j++){
	aux[k][j] = buff[j];
      } 
    }
    
//     for(i = 0; i < nproc; i++){
//       fprintf(out,"domain %ld\n",i);
//       for(j = 0; j < nbnd[i]; j++){
// 	fprintf(out,"%ld\n",aux[i][j]);
//       }
//     }
    
    coarseidentif = new long[tnbn];
    
    for(i = 0; i < tnbn; i++){
      // initial import
      coarseidentif[i] = aux[vazbadom[i][0]][vazbanod[i][0]];
      // controlling
      m = 0;
      if (coarseidentif[i] == 3){
	a = 1;
	b = 0;
      }
      else{
	a = 0;
	b = 1;
      }	
      for(j = 1; j < icmultip[i]; j++){
	if(aux[vazbadom[i][j]][vazbanod[i][j]] == 3){
	  a++;
	}
	if(aux[vazbadom[i][j]][vazbanod[i][j]] == 2){
	  b++;
	}
      }
      if((a != 0 && b > 0)||(a > 0 && b != 0)){
	if(a>=b){
	  coarseidentif[i] = 3;
	  fprintf(out,"m = %ld prepis %ld z 2 na %ld\n",m,i,coarseidentif[i]);
	  
	}
	else{
	  coarseidentif[i] = 2;
	  fprintf(out," m = %ld prepis %ld z 3 na %ld\n",m,i,coarseidentif[i]);
	}
      }
    }
    
//     for(i = 0; i < tnbn; i++){
//       fprintf(out,"%ld coarseidentif %ld\n",i,coarseidentif[i]);
//     }
    
    nvdom = new long[nproc];
    maxnv = 0;
    for(i = 0; i < nproc; i++){
      nvdom[i] = 0;
      for(j = 0; j < nbnd[i]; j++){
	if(aux[i][j] == 3){
	  nvdom[i]++;
	}
      }
      if(nvdom[i] >= maxnv){
	maxnv = nvdom[i];
      }
      delete []aux[i];
    }
    delete []aux;
    
    for(i= 0; i < nproc; i++){
      fprintf(out,"Domain %ld has %ld corner nodes\n",i,nvdom[i]);
    }
    fprintf(out,"Maximum number of corner nodes on one domain  %ld\n",maxnv);
    
  }
  else{
   MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  delete []buff;
  
  
}


//
// corner detection 

//08.05.07  JB
//
void partop::assemble_ltg_with_corners(long *domproc, long *ltg,FILE *out)
{
  long i,j,k;
  long buffsize,c,b,a;
  long *buff,*master_corner,*coarsenumb;
  MPI_Status stat;
  
  
   
  buffsize = maxnbn;
  buff = new long[buffsize];
  
  //  numbering of corner and edge nodes on master
  if(myrank==0){
    coarsenumb = new long[tnbn];
    // numbering of corner and edge nodes on master
    c = 1;
    b = 1;
    //fprintf(out,"tncen je %ld\n\n",tncen);
    for(i = 0; i < tnbn; i++){
      //fprintf(out,"%ld %ld\n",i+1,coarseidentif[i]);
      if(coarseidentif[i] == 3){
        coarsenumb[i] = c*(-1);
        c++;
	//fprintf(out,"now %ld %ld\n",i+1,coarsenumb[i]);
      }
      if(coarseidentif[i] == 2){
	coarsenumb[i] = b;
	b++;
	//fprintf(out,"now %ld %ld\n",i+1,coarsenumb[i]);
      }
    }
    fprintf(out,"\n\n\n Number of corner nodes in whole problem is %ld\n",c-1);
    fprintf(out,"\n\n\n Number of boundary nodes in whole problem is %ld\n",b-1);
    
    if((c-1)+(b-1) != tnbn){
      fprintf (stderr,"Problem in function corner_detection - the number of corner and edge nodes is not equal to total number of boundary nodes (file %s, line %d).\n",__FILE__,__LINE__);
      fprintf (out,"Problem in function corner_detection - the number of corner and edge nodes is not equal to total number of boundary nodes (file %s, line %d).\n",__FILE__,__LINE__);
      
    }
    
    // For Slaves
    for (i = 1;i < nproc; i++){
      for(k = 0; k < nbnd[i]; k++){
        buff[k] =  coarsenumb[icnbnmas[i][k]];
      }
      MPI_Send (buff,buffsize,MPI_LONG,i,myrank,MPI_COMM_WORLD);
    }
    
    // For master
    for(k = 0; k < nbnd[0]; k++){
      buff[k] =  coarsenumb[icnbnmas[0][k]];
    }
    delete []coarseidentif;
    delete []coarsenumb;
  }
  
  // slave
  else{
    MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  
  
  //   fprintf(out,"\n\ncorner nodes - buff\n");
  //    for(i = 0; i < nbn; i++){
  //      fprintf(out,"%ld %ld\n",i+1,buff[i]);
  //    }
  
  for(i = 0; i < nbn; i++){
    ltg[lnbndom[i]]=buff[i]-1;
  }
  for(i = 0; i < nin; i++){
    ltg[lnindom[i]]=-1;
  }

  
  fflush(out);
  // deleting 
  delete []buff;
  
}
































/*
  ZACATEK ROZSAHLYCH UPRAV
  27.7.2009
*/

/**
   function collects numbers of all nodes on subdomains
   
   the following function has to be called before this function:
   void initiation (gtopology *top,long *ltg);
   
   
   function establishes:

   maxnn - maximum number of nodes on one subdomain
   nnsd - (M) array containing numbers of nodes on subdomains
   
   @param domproc - array containing domain-processor correspondence
   @param out - output file (for auxiliary output)

   JK, 30.6.2005
*/
void partop::numbers_of_all_nodes_on_subdomains (long *domproc,FILE *out)
{
  long i,j,k;
  MPI_Status stat;
  
  //  the maximum number of nodes defined on subdomain (will be sent to the master)
  maxnn=nn;
  
  // *********
  //  master
  // *********
  if (myrank==0){
    //  array containing number of nodes on subdomains
    if (nnsd!=NULL)
      delete [] nnsd;
    nnsd = new long [nproc];
    
    //  master contribution
    j=domproc[0];
    nnsd[j]=maxnn;
    
    for (i=1;i<nproc;i++){
      MPI_Recv (&k,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      
      //  slave contributions
      j=domproc[stat.MPI_TAG];
      if (maxnn<k)  maxnn=k;
      nnsd[j]=k;
    }
    
    for (i=1;i<nproc;i++){
      MPI_Send (&maxnn,1,MPI_LONG,i,myrank,MPI_COMM_WORLD);
    }
  }
  // ********
  // slaves
  // ********
  else{
    MPI_Send (&nn,1,MPI_LONG,0,myrank,MPI_COMM_WORLD);
    MPI_Recv (&maxnn,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  

  if (myrank==0){
    fprintf (out,"\n\n\n the maximum number of nodes on subdomains (maxnn) %ld\n",maxnn);
    fprintf (out,"\n\n\n array nnsd \n\n");
    for (i=0;i<nproc;i++){
      fprintf (out,"\n %ld",nnsd[i]);
    }
  }
  
}

/**
   function assembles the array bmultip or amultip
   function computes the variables tnnp or tnbn
   
   this function can assemble also arrays nbnd and nind
   but the coupled DOFs generate additional boundary/interface nodes
   which are not known in this function
   therefore, there is function assemble_nbnd which does it
   
   function determines:
   
   maxnbn - maximum number of boundary/interface nodes on subdomain
   nbn - number of boundary/interface nodes
   tnbn - total number of boundary/interface nodes in the problem
   bnodes - (M)
   bmultip - (M)
   maxnbn - maximum number of boundary/interface nodes on subdomain
   tnnp - (M) total number of nodes in the problem
   allnodes - (M)
   amultip - (M) 

   @param ltg - local to global correspondence
   @param domproc - array containing domain-processor correspondence
   @param out - output file for auxiliary print
   @param proc_name - processor name
   
   27.7.2009, JK
*/
void partop::assemble_multip (long *ltg,long *domproc,FILE *out,char *proc_name)
{
  long i,j,k,buffsize;
  long *buff;
  MPI_Status stat;
  
  if (maxnn==0){
    par_print_err(myrank,proc_name,"maximum number of nodes on one subdomain is equal to zero",__FILE__, __LINE__, __func__);
    par_print_err(myrank,proc_name,"function numbers_of_all_nodes_on_subdomains has to be called first",__FILE__, __LINE__, __func__);
  }
  if (myrank==0){
    if (nnsd==NULL){
      par_print_err(myrank,proc_name,"array nnsd (number of nodes on subdomains) is not assembled",__FILE__, __LINE__, __func__);
      par_print_err(myrank,proc_name,"function numbers_of_all_nodes_on_subdomains has to be called first",__FILE__, __LINE__, __func__);
    }
  }
  
  switch (md){
  case bound_nodes:{
    //  ltg=-1 for the internal nodes
    //  ltg>-1 for the interface/boundary nodes
    
    /*
    buffsize=2;
    buff = new long [buffsize];
    
    //  the number of boundary/interface nodes on subdomain
    nbn = 0;
    //  the maximum number of boundary/interface nodes on subdomain
    maxnbn = 0;
    //  the total number of boundary/interface nodes in the problem
    tnbn = 0;

    //  loop over the number of all nodes on subdomain
    for (i=0;i<nn;i++){
      if (tnbn<ltg[i])
        tnbn=ltg[i];
      if (ltg[i]>-1)
        nbn++;
    }
    
    buff[0]=tnbn;
    buff[1]=nbn;
    
    if (myrank==0){
      //  master contribution
      j=domproc[0];
      if (tnbn<buff[0])  tnbn=buff[0];
      if (maxnbn<buff[1])  maxnbn=buff[1];
      nbnd[j]=buff[1];
      
      for (i=1;i<nproc;i++){
        MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
        j=domproc[stat.MPI_TAG];
        //  slave contributions
        if (tnbn<buff[0])
          tnbn=buff[0];
        if (maxnbn<buff[1])
          maxnbn=buff[1];
      }
      tnbn++;
      
      buff[0]=tnbn;
      buff[1]=maxnbn;
      
      for (i=1;i<nproc;i++){
        MPI_Send (buff,buffsize,MPI_LONG,i,myrank,MPI_COMM_WORLD);
      }
    }
    else{
      MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    }
    MPI_Barrier (MPI_COMM_WORLD);
    
    tnbn=buff[0];
    maxnbn=buff[1];
    
    delete [] buff;
    */



    //  ltg=-1 for the internal nodes
    //  ltg>-1 for the interface/boundary nodes
    
    buffsize=1;
    buff = new long [buffsize];
    
    //  the total number of boundary/interface nodes in the problem
    tnbn = 0;

    //  loop over the number of all nodes on subdomain
    for (i=0;i<nn;i++){
      if (tnbn<ltg[i])
        tnbn=ltg[i];
    }
    
    buff[0]=tnbn;
    
    if (myrank==0){
      //  master contribution
      j=domproc[0];
      if (tnbn<buff[0])  tnbn=buff[0];
      
      for (i=1;i<nproc;i++){
        MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
        j=domproc[stat.MPI_TAG];
        //  slave contributions
        if (tnbn<buff[0])
          tnbn=buff[0];
      }
      tnbn++;
      
      buff[0]=tnbn;
      
      for (i=1;i<nproc;i++){
        MPI_Send (buff,buffsize,MPI_LONG,i,myrank,MPI_COMM_WORLD);
      }
    }
    else{
      MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    }
    MPI_Barrier (MPI_COMM_WORLD);
    
    tnbn=buff[0];
    
    delete [] buff;
    
    
    
    buffsize = maxnn;
    buff = new long [buffsize];
    for (i=0;i<maxnn;i++){
      buff[i]=-1;
    }
    
    for(i=0;i<nn;i++){
      buff[i] = ltg[i];
    }
    
    
    if(myrank == 0){

      if (bnodes != NULL){
        for (i=0;i<nproc;i++){
          delete [] bnodes[i];
        }
        delete [] bnodes;
      }
      bnodes = new long* [nproc];
      for (i=0;i<nproc;i++){
        bnodes[i] = new long [nnsd[i]];
      }
      
      if (bmultip!=NULL){
        delete [] bmultip;
      }
      bmultip = new long [tnbn];
      for(i=0;i<tnbn;i++){
        bmultip[i]=0;
      }
      
      // master contribution
      k=domproc[0];
      for(j=0;j<nnsd[k];j++){
	bnodes[k][j]=buff[j];
        if (buff[j]>-1){
          bmultip[buff[j]]++;
        }
      }
      
      //  slave contributions
      for (i=1;i<nproc;i++){
        MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);

	k=domproc[stat.MPI_TAG];
        for(j=0;j<nnsd[k];j++){
	  bnodes[k][j]=buff[j];
          if (buff[j]>-1){
            bmultip[buff[j]]++;
          }
        }
      }
      
    }
    else{
      MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
    }
    MPI_Barrier (MPI_COMM_WORLD);
    delete [] buff;
    
    buffsize=tnbn;
    buff = new long [buffsize];
    
    if (myrank==0){
      for (i=0;i<tnbn;i++){
        buff[i]=bmultip[i];
      }
      
      for (i=1;i<nproc;i++){
        MPI_Send (buff,buffsize,MPI_LONG,i,myrank,MPI_COMM_WORLD);
      }
    }
    else{
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    }
    MPI_Barrier (MPI_COMM_WORLD);
    
    delete [] buff;
       
    break;
  }



  case all_nodes:
  case neg_bound_nodes:{
    
    buffsize=maxnn+1;
    buff = new long [buffsize];

    //  the total number of nodes in the problem
    tnnp=0;
    //  loop over the number of nodes on subdomain
    for (i=0;i<nn;i++){
      buff[i]=ltg[i];
      if (tnnp<ltg[i])
        tnnp=ltg[i];
    }
    buff[maxnn]=tnnp;
    
    // *********
    //  master
    // *********
    if (myrank==0){
      if (allnodes != NULL){
        for (i=0;i<nproc;i++){
          delete [] allnodes[i];
        }
        delete [] allnodes;
      }
      allnodes = new long* [nproc];
      for (i=0;i<nproc;i++){
        allnodes[i] = new long [nnsd[i]];
      }
      
      
      k=domproc[0];
      for (j=0;j<nnsd[k];j++){
        allnodes[k][j]=buff[j];
      }
      if (tnnp<buff[maxnn])
        tnnp=buff[maxnn];
      

      for (i=1;i<nproc;i++){
        MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
        
        k=domproc[stat.MPI_TAG];
        for (j=0;j<nnsd[k];j++){
          allnodes[k][j]=buff[j];
        }
        if (tnnp<buff[maxnn])
          tnnp=buff[maxnn];
      }
    }
    
    // ********
    // slaves
    // ********
    else{
      MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
    }
    
    if (myrank==0){
      //  array containing number of subdomains which share node
      //  this is called node multiplicity
      
      //  must be increased due to indices from 0 instead of 1
      tnnp++;
      
      if (amultip != NULL)
        delete [] amultip;
      amultip = new long [tnnp];
      for (i=0;i<tnnp;i++){
        amultip[i]=0;
      }
      
      //  computation of node incidences
      for (i=0;i<nproc;i++){
        for (j=0;j<nnsd[i];j++){
          amultip[allnodes[i][j]]++;
        }
      }
    }
    
    break;
  }
  default:{
    par_print_err(myrank,proc_name,"unknown type of mesh description",__FILE__, __LINE__, __func__);
  }
  }
  MPI_Barrier (MPI_COMM_WORLD);

  
  // *******************
  //  auxiliary output
  // *******************
  if (myrank==0){
    if (allnodes!=NULL){
      fprintf (out,"\n\n array allnodes \n");
      for (i=0;i<nproc;i++){
	fprintf (out,"\n");
	for (j=0;j<nnsd[i];j++){
	  fprintf (out,"\n nproc %3ld  node %6ld   allnode %6ld",i,j,allnodes[i][j]);
	}
      }
    }
    if (amultip!=NULL){
      fprintf(out,"\n\n\n tnnp - total number of nodes in whole problem is %ld\n\n",tnnp);
      for (i=0;i<tnnp;i++){
	fprintf (out,"\n node %6ld  amultip %3ld",i,amultip[i]);
      }
    }
    if (bmultip!=NULL){
      for (i=0;i<tnbn;i++){
	fprintf (out,"\n boundary/interface node %6ld  bmultip %3ld",i,bmultip[i]);
      }
    }
    
    if (bnodes!=NULL){
      fprintf(out,"\n\n\n array bnodes \n\n");
      for (i=0;i<nproc;i++){
	fprintf (out,"\n domain %ld\n",i);
	for (j=0;j<nnsd[i];j++){
	  fprintf (out,"\n bnodes  %6ld %6ld",j,bnodes[i][j]);
 	}
      }
    }

    fprintf(out,"\n\n\n total number of boundary/interface nodes in whole problem (tnbn) is %ld\n\n",tnbn);
    fprintf (out,"\n\n\n maximum number of boundary/interface nodes on subdomain (maxnbn) is %ld\n\n",maxnbn);
  }
  
}



/**
   function searches for coupled DOFs which are shared
   by interfaces

   ncdof - the number of coupled DOFs
   coupdof - (M) array containing number of coupled DOFs on subdomains
   coupdofmas - (M) array containing suspicious indicators of coupled DOFs
   nbcdof - the number of boundary/interface coupled DOFs
   
   @param top - pointer to the general topology
   @param domproc - array containing domain-processor correspondence
   @param out - output file
   
   28.7.2009, JK
*/
void partop::coupled_dofs (gtopology *top,long *domproc,FILE *out)
{
  long i,j,k,l,ndofn;
  long *buff,buffsize;
  MPI_Status stat;
  
  //  the number of coupled DOFs
  //  
  ncdof=0;

  //  loop over the number of nodes on subdomain
  for (i=0;i<nn;i++){
    //  the number of DOFs on node
    ndofn=top->give_ndofn (i);
    //  loop over the number of DOFs on node
    for (j=0;j<ndofn;j++){
      k=top->give_dof (i,j);
      if (k>1){
	if (ncdof<k)
	  ncdof=k;
      }
    }
  }
  
  
  // *********
  //  master
  // *********
  if (myrank==0){
    //  loop over the number of slaves
    for (i=1;i<nproc;i++){
      MPI_Recv (&k,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      //  slave contributions
      if (ncdof<k)  ncdof=k;
      
    }
    
    //  loop over the number of slaves
    for (i=1;i<nproc;i++){
      MPI_Send (&ncdof,1,MPI_LONG,i,myrank,MPI_COMM_WORLD);
    }
    
  }
  // ********
  // slaves
  // ********
  else{
    MPI_Send (&ncdof,1,MPI_LONG,0,myrank,MPI_COMM_WORLD);
    MPI_Recv (&ncdof,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  
  buffsize=ncdof;
  buff=new long [buffsize];
  for (i=0;i<buffsize;i++){
    buff[i]=0;
  }
  
  //  loop over the number of nodes on subdomain
  for (i=0;i<nn;i++){
    //  the number of DOFs on node
    ndofn=top->give_ndofn (i);
    //  loop over the number of DOFs on node
    for (j=0;j<ndofn;j++){
      k=top->give_dof (i,j);
      if (k>1){
	buff[k-1]=k;
      }
    }
  }
  
  fprintf (out,"\n\n");
  for (i=0;i<buffsize;i++){
    fprintf (out,"\n buff %3ld  %ld",i,buff[i]);
  }


  // *********
  //  master
  // *********
  if (myrank==0){
    
    //  array containing number of coupled DOFs on subdomains
    //  the indicator is greater than one, otherwise the
    //  code number is not coupled
    if (coupdof!=NULL){
      for (i=0;i<nproc;i++){
	delete [] coupdof[i];
      }
      delete [] coupdof;
    }
    coupdof = new long* [nproc];
    for (i=0;i<nproc;i++){
      coupdof[i] = new long [ncdof];
      for (j=0;j<ncdof;j++){
	coupdof[i][j]=0;
      }
    }

    //  master contribution
    k=domproc[0];
    for (j=0;j<ncdof;j++){
      l=buff[j];
      if (l>1)
	coupdof[k][l-1]++;
    }
    
    //  loop over the number of slaves
    for (i=1;i<nproc;i++){
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      //  slave contributions
      k=domproc[stat.MPI_TAG];
      for (j=0;j<ncdof;j++){
	l=buff[j];
	if (l>1)
	  coupdof[k][l-1]++;
      }
    }
    
    //  array containing suspicious indicators of coupled DOFs
    if (coupdofmas!=NULL){
      delete [] coupdofmas;
    }
    coupdofmas = new long [ncdof];
    for (i=0;i<ncdof;i++){
      coupdofmas[i]=0;
    }
    
    //  loop over the number of coupled DOFs
    for (i=0;i<ncdof;i++){
      k=0;
      //  loop over the number of subdomains
      for (j=0;j<nproc;j++){
	if (coupdof[j][i]>0)
	  k++;
      }
      if (k>1){
	//  coupled DOFs described by the i-th indicator occur at least on two subdomains
	//  such coupled DOFs have to be treated as the boundary/interface DOFs
	coupdofmas[i]=1;
      }
    }
    
    //  the number of boundary/interface coupled DOFs
    nbcdof=0;
    for (i=0;i<ncdof;i++){
      if (coupdofmas[i]==1)
	nbcdof++;
    }
    
  }
  // ********
  // slaves
  // ********
  else{
    MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  
  
  // *******************
  //  auxiliary output
  // *******************
  if (myrank==0){
    fprintf (out,"\n\n\n ncdof - the number of coupled DOFs %ld",ncdof);
    fprintf (out,"\n\n\n nbcdof - the number of boundary/interface coupled DOFs %ld",nbcdof);
    fprintf (out,"\n\n");
    for (i=0;i<nproc;i++){
      for (j=0;j<ncdof;j++){
	fprintf (out,"\n coupdof %3ld %3ld   %ld",i,j,coupdof[i][j]);
      }
    }
    fprintf (out,"\n\n");
    for (i=0;i<ncdof;i++){
      fprintf (out,"\n coupdofmas %6ld   %ld",i,coupdofmas[i]);
    }
  }
  
}



/**
   function collects the numbers of all degrees of freedom on subdomains
   constrained and prescribed DOFs are taken into account, it means
   they are added to the number of DOFs
   
   function establishes:
   
   maxndof - maximum number of degrees of freedom on one subdomain
             it summarizes number of DOFs at nodes belonging to one subdomain
             it differs from the actual degrees of freedom of subdomains in constraints
             
   nalldof - (M) array of numbers of degrees of freedom on subdomains


   @param top - pointer to the sequential general topology
   @param domproc - array containing domain-processor correspondence
   @param out - output file (for auxiliary output)
   
   JK, 20.3.2007
*/
void partop::numbers_of_all_dofs_on_subdomains (gtopology *top,long *domproc,FILE *out)
{
  long i,j,k;
  MPI_Status stat;
  
  //  maximum number of nodes defined on subdomain (will be sent to the master)
  maxndof=0;
  for (i=0;i<nn;i++){
    maxndof+=top->give_ndofn (i);
  }
  
  // *********
  //  master
  // *********
  if (myrank==0){
    //  array containing number of nodes on subdomains
    if (nalldof!=NULL)
      delete [] nalldof;
    nalldof = new long [nproc];
    
    //  master contribution
    j=domproc[0];
    nalldof[j]=maxndof;
    
    for (i=1;i<nproc;i++){
      MPI_Recv (&k,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      
      //  slave contributions
      j=domproc[stat.MPI_TAG];
      if (maxndof<k)  maxndof=k;
      nalldof[j]=k;
    }
    
    for (i=1;i<nproc;i++){
      MPI_Send (&maxndof,1,MPI_LONG,i,myrank,MPI_COMM_WORLD);
    }
  }
  // ********
  // slaves
  // ********
  else{
    MPI_Send (&maxndof,1,MPI_LONG,0,myrank,MPI_COMM_WORLD);
    MPI_Recv (&maxndof,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  
  
  if (myrank==0){
    fprintf (out,"\n\n\n the maximum number of DOFs on subdomain (maxndof) %ld\n",maxndof);
    fprintf (out,"\n\n\n the numbers of all DOFs on each subdomain (alldof)\n\n");
    for (i=0;i<nproc;i++){
      fprintf (out,"\n %ld",nalldof[i]);
    }
  }

}

/**
   function assembles array ndofnmas
   it contains the number of DOFs on nodes
   
   @param top - pointer to the general topology
   @param domproc - array containing domain-processor correspondence
   @param out - output file
   @param proc_name - processor name
   
   JK, 28.7.2009
*/
void partop::ndofn_master (gtopology *top,long *domproc,FILE *out,char *proc_name)
{
  long i,j,k,l,ndofn;
  long *buff,buffsize;
  MPI_Status stat;
  
  /*
  if (md!=all_nodes){
    par_print_err(myrank,proc_name,"wrong mesh description is used",__FILE__, __LINE__, __func__);
    abort ();
  }
  */

  if (maxnn==0){
    par_print_err(myrank,proc_name,"maximum number of nodes on one subdomain is equal to zero",__FILE__, __LINE__, __func__);
    par_print_err(myrank,proc_name,"function numbers_of_all_nodes_on_subdomains has to be called first",__FILE__, __LINE__, __func__);
    abort ();
  }
  
  buffsize=maxnn;
  buff = new long [buffsize];
  
  //  loop over the number of nodes on subdomain
  for (i=0;i<nn;i++){
    //  number of DOFs on nodes
    buff[i] = top->give_ndofn (i);
  }
  
  // *********
  //  master
  // *********
  if (myrank==0){
    
    if (tnnp==0){
      par_print_err(myrank,proc_name,"total number of nodes in the problem (tnnp) is zero",__FILE__, __LINE__, __func__);
      par_print_err(myrank,proc_name,"function partop::assemble_multip has to be called first",__FILE__, __LINE__, __func__);
    }
    if (nnsd==NULL){
      par_print_err(myrank,proc_name,"array nnsd (number of nodes on subdomains) is not assembled",__FILE__, __LINE__, __func__);
      par_print_err(myrank,proc_name,"function partop::numbers_of_all_nodes_on_subdomains has to be called first",__FILE__, __LINE__, __func__);
    }
    if (allnodes==NULL){
      par_print_err(myrank,proc_name,"array allnodes is not assembled",__FILE__, __LINE__, __func__);
      par_print_err(myrank,proc_name,"function partop::assemble_multip has to be called first",__FILE__, __LINE__, __func__);
    }
    
    //  array containing the numbers of DOF on nodes
    if (ndofnmas!=NULL){
      delete [] ndofnmas;
    }
    ndofnmas = new long [tnnp];
    
    
    //  master contribution
    
    //  subdomain id
    k=domproc[0];
    //  loop over the number of nodes
    for (j=0;j<nnsd[k];j++){
      //  global node number
      l=allnodes[k][j];
      //  the number of DOFs on the node
      ndofnmas[l]=buff[j];
    }
    
    for (i=1;i<nproc;i++){
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      
      //  slave contributions
      
      //  subdomain id
      k=domproc[stat.MPI_TAG];
      //  loop over the number of nodes
      for (j=0;j<nnsd[k];j++){
	//  global node number
	l=allnodes[k][j];
	//  the number of DOFs on the node
	ndofnmas[l]=buff[j];
      }
      
    }
    
  }
  // *********
  //  slaves
  // *********
  else{
    MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  
  // *******************
  //  auxiliary output
  // *******************
  if (myrank==0){
    fprintf(out,"\n\n\n the numbers of DOFs on nodes (ndofnmas)\n\n");
    for (i=0;i<tnnp;i++){
      fprintf (out,"\n node %6ld  ndofnmas %3ld",i,ndofnmas[i]);
    }
  }  
  
    
}

/**
   function collects id of DOFs on the master
   
   function establishes:
   
   difind_mas - 

   @param top - pointer to the general topology
   @param domproc - array containing domain-processor correspondence
   @param out - output file
   @param proc_name - processor name
   
   JK, 28.7.2009
*/
void partop::dofind_master (gtopology *top,long *domproc,FILE *out,char *proc_name)
{
  long i,j,k,l,m,ii,ndofn;
  long *buff,buffsize;
  MPI_Status stat;
  
  /*
  if (md!=all_nodes){
    par_print_err(myrank,proc_name,"wrong mesh description is used",__FILE__, __LINE__, __func__);
    abort ();
  }
  */

  if (maxndof==0){
    par_print_err(myrank,proc_name,"maximum number of DOFs is zero",__FILE__, __LINE__, __func__);
    par_print_err(myrank,proc_name,"function partop::numbers_of_all_dofs_on_subdomains has to be called first",__FILE__, __LINE__, __func__);
  }
  
  buffsize=maxndof;
  buff = new long [buffsize];
  
  //  index of array buff
  k=0;
  //  loop over the number of nodes on subdomain
  for (i=0;i<nn;i++){
    //  number of DOFs on nodes
    ndofn = top->give_ndofn (i);
    //  loop over the number of DOFs on node
    for (j=0;j<ndofn;j++){
      buff[k]=top->give_dof (i,j);
      k++;
    }
  }
  
  // *********
  //  master
  // *********
  if (myrank==0){
    
    //  array containing DOF indicators
    if (dofindmas!=NULL){
      for (i=0;i<tnnp;i++){
	delete [] dofindmas[i];
      }
      delete [] dofindmas;
    }
    dofindmas = new long* [tnnp];
    for (i=0;i<tnnp;i++){
      dofindmas[i] = new long [ndofnmas[i]];
    }

    
    //  master contribution
    
    ii=0;
    //  subdomain id
    k=domproc[0];
    //  loop over the number of nodes
    for (j=0;j<nnsd[k];j++){
      //  global node number
      l=allnodes[k][j];
      //  loop over the number of DOFs in node
      for (m=0;m<ndofnmas[l];m++){
	dofindmas[l][m]=buff[ii];
	ii++;
      }
    }
    
    
    //  loop over the number of slaves
    for (i=1;i<nproc;i++){
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      
      //  slave contributions
      
      ii=0;
      //  subdomain id
      k=domproc[stat.MPI_TAG];
      //  loop over the number of nodes
      for (j=0;j<nnsd[k];j++){
	//  global node number
        l=allnodes[k][j];
	//  loop over the number of DOFs in node
	for (m=0;m<ndofnmas[l];m++){
	  dofindmas[l][m]=buff[ii];
	  ii++;
	}
      }
      
    }
    
  }
  // *********
  //  slaves
  // *********
  else{
    MPI_Send (buff,buffsize,MPI_LONG,0,myrank,MPI_COMM_WORLD);
  }
  MPI_Barrier (MPI_COMM_WORLD);

  // *******************
  //  auxiliary output
  // *******************
  if (myrank==0){
    fprintf(out,"\n\n\n DOF id (dofindmas)\n\n");
    for (i=0;i<tnnp;i++){
      fprintf (out,"\n node %6ld  ",i);
      for (j=0;j<ndofnmas[i];j++){
	fprintf (out,"  dofindmas %3ld",dofindmas[i][j]);
      }
    }
    fprintf (out,"\n\n");
  }  

}




/**
   function updates node multiplicity

   
   function establishes:
   
   tnbn - (M) total number of boundary nodes
   amultip - (M) array of all node multiplicity, it is assembled if md=all_nodes
   dofind - (M) array containing DOF indicators
   
   @param top - pointer to the general topology
   @param out - output file for auxiliary print
   @param proc_name - processor name
   
   JK, 28.7.2009
*/
void partop::update_multip (gtopology *top,FILE *out,char *proc_name)
{
  long i,j,k,l,ndofn,dofid;
  
  /*
  if (md!=all_nodes){
    par_print_err(myrank,proc_name,"wrong mesh description is used",__FILE__, __LINE__, __func__);
    abort ();
  }
  */

  if (myrank==0){
    
    //  array containing DOF indicators
    if (dofind!=NULL){
      for (i=0;i<tnnp;i++){
	delete [] dofind[i];
      }
      delete [] dofind;
    }
    dofind = new long* [tnnp];
    for (i=0;i<tnnp;i++){
      dofind[i] = new long [ndofnmas[i]];
      for (j=0;j<ndofnmas[i];j++){
	dofind[i][j]=0;
      }
    }

    // *********************************************************************************
    //  in the case of coupled DOFs, treatment with nodes is not enough
    //  DOFs have to be split to group of internal DOFs and boundary/interface DOFs
    //
    //  there are three types of boundary/interface DOFs:
    //
    //  all DOFs belonging to the boundary/interface nodes are boundary/interface DOFs
    //
    //  coupled DOFs connected to any boundary/interface node are boundary/interface DOFs
    //
    //  coupled DOFs not connected to any boundary/interface node but shared by at least
    //  two subdomains are also boundary/interface DOFs
    //
    //  all remaining DOFs are internal DOFs and can be eliminated on subdomains
    //
    // *********************************************************************************
    
    //  loop over the number of subdomains
    for (i=0;i<nproc;i++){
      //  loop over the number of nodes on subdomains
      for (j=0;j<nnsd[i];j++){
	//  global node number
        l=allnodes[i][j];
	//  the number of DOFs in node
	ndofn = ndofnmas[l];
	//  loop over the number of DOFs in node
	for (k=0;k<ndofn;k++){
	  //  DOF indicator
	  dofid=dofindmas[l][k];
	  if (amultip[l]>1){
	    //  the i-th node is boundary/interface node, therefore this is boundary/interface DOF
	    dofind[l][k]=1;
	    if (dofid>1){
	      //  there is coupled DOF defined in this boundary/interface node
	      //  this information has to be stored in the array coupdofmas
	      //  it will be used in the following loop
	      if (coupdofmas[dofid-1]==0)
		coupdofmas[dofid-1]=1;
	    }
	  }
	  if (dofid>1){
	    
	    //  there is coupled DOF connected to at least two subdomains
	    if (coupdofmas[dofid-1]==1){
	      //  this coupled DOF has to be taken into account
	      //  this is boundary/interface DOF
	      dofind[l][k]=1;
	    }

	  }
	}
      }
    }
    
    fprintf (out,"\n\n\n");
    for (i=0;i<tnnp;i++){
      fprintf (out,"\n dofind %6ld ",i);
      for (j=0;j<ndofnmas[i];j++){
	fprintf (out,"  %3ld",dofind[i][j]);
      }
    }
    
    
    // *********************************************************************************
    //  modification of the array amultip
    //  the array has to be modified with respect to the array dofind
    //  if any DOF in node is denoted as the boundary/interface DOF, the appropriate node
    //  has to be denoted as the boundary/interface node, this fact is
    //  assured with the help of the array amultip, where value 1 is rewritten to the value 2
    // *********************************************************************************
    
    //  total number of boundary/interface nodes
    tnbn=0;
    //  loop over the number of subdomain
    for (i=0;i<nproc;i++){
      //  loop over the number of nodes on subdomain
      for (j=0;j<nnsd[i];j++){
	//  global node number
        l=allnodes[i][j];
	//  the number of DOFs in node
	ndofn = ndofnmas[l];
	//  loop over the number of DOFs in node
	for (k=0;k<ndofn;k++){
	  if (dofind[l][k]==1){
	    //  the DOF is boundary/interface DOF
	    if (amultip[l]==1){
	      //  the node is internal node
	      //  it has to be changed from internal to boundary/interface
	      amultip[l]=2;
	    }
	  }
	}
      }
    }
    // *********************************************************************************
    //  end of modification of the array amultip
    // *********************************************************************************
    
  }
  
  // *******************
  //  auxiliary output
  // *******************
  if (myrank==0){
    if (amultip!=NULL){
      fprintf(out,"\n\n\n modified array amultip with respect to the coupled DOFs\n");
      for (i=0;i<tnnp;i++){
	fprintf (out,"\n node %6ld  amultip %3ld",i,amultip[i]);
      }
    }
    fprintf(out,"\n\n\n DOF id (dofind)\n\n");
    for (i=0;i<tnnp;i++){
      fprintf (out,"\n node %6ld  ",i);
      for (j=0;j<ndofnmas[i];j++){
	fprintf (out,"  dofind %3ld",dofind[i][j]);
      }
    }
    fprintf (out,"\n\n");
  }
  
}

/**
   function assembles the arrays nbnd a nind
   
   these arrays are assembled separately in this function
   they can be assembled in the function assemble_multip
   but the coupled DOFs generate additional boundary/interface
   nodes, these nodes are not known in the function assemble_multip
   
   function assembles:
   nbnd - (M) array of the numbers of boundary/interface nodes
   nind - (M) array of the numbers of internal nodes
   tnbn - the total number of all boundary/interface nodes
          it is computed once again because additional boundary/interface nodes
	  may be added due to the coupled DOFs
   maxnbn - maximum number of boundary/interface nodes
   nbn - the number of boundary/interface nodes
   nin - the number of internal nodes
   

   @param ltg - local to global correspondence
   @param domproc - array containing domain-processor correspondence
   @param out - output file (for auxiliary output)
   @param proc_name - processor name

   JK, 29.7.2009
*/
void partop::assemble_nbnd_nind (long *ltg,long *domproc,FILE *out,char *proc_name)
{
  long i,j,k;
  long *buff,buffsize;
  MPI_Status stat;
  
  switch (md){
  case bound_nodes:{
    
    if (maxnn==0){
      par_print_err(myrank,proc_name,"maximum number of nodes on one subdomain is equal to zero",__FILE__, __LINE__, __func__);
      par_print_err(myrank,proc_name,"function numbers_of_all_nodes_on_subdomains has to be called first",__FILE__, __LINE__, __func__);
    }
    
    //  determination of total number of boundary nodes
    nbn=0;
    for (i=0;i<nn;i++){
      if (ltg[i]>-1)
        nbn++;
    }
    
    if (myrank==0){
      //  list of number of boundary nodes on subdomains
      if (nbnd != NULL)
        delete [] nbnd;
      nbnd = new long [nproc];
      
      maxnbn=0;
      
      //  master contribution
      j=domproc[0];
      nbnd[j]=nbn;
      if (maxnbn<nbn)
	maxnbn=nbn;
      
      for (i=1;i<nproc;i++){
        MPI_Recv (&k,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
	
        //  slave contributions
        j=domproc[stat.MPI_TAG];
        nbnd[j]=k;
	if (maxnbn<k)
	  maxnbn=nbnd[j];
	
      }
      for (i=1;i<nproc;i++){
	MPI_Send (&maxnbn,1,MPI_LONG,i,myrank,MPI_COMM_WORLD);
      }
    }
    else{
      MPI_Send (&nbn,1,MPI_LONG,0,myrank,MPI_COMM_WORLD);
      MPI_Recv (&maxnbn,1,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    }
    MPI_Barrier (MPI_COMM_WORLD);
    
    break;
  }
    
    
    
    
  case all_nodes:
  case neg_bound_nodes:{
    
    buffsize=2;
    buff = new long [buffsize];
    
    // *********
    //  master
    // *********
    if (myrank==0){
      
      if (nnsd==NULL){
	par_print_err(myrank,proc_name,"array nnsd is not assembled",__FILE__, __LINE__, __func__);
	par_print_err(myrank,proc_name,"function partop::numbers_of_all_nodes_on_subdomains has to be called first",__FILE__, __LINE__, __func__);
      }
      if (amultip==NULL){
	par_print_err(myrank,proc_name,"array amultip is not assembled",__FILE__, __LINE__, __func__);
	par_print_err(myrank,proc_name,"function partop::assemble_multip has to be called first",__FILE__, __LINE__, __func__);
      }
      if (allnodes==NULL){
	par_print_err(myrank,proc_name,"array allnodes is not assembled",__FILE__, __LINE__, __func__);
	par_print_err(myrank,proc_name,"function partop::assemble_multip has to be called first",__FILE__, __LINE__, __func__);
      }

      //  list of the numbers of boundary nodes on subdomains
      if (nbnd != NULL)
        delete [] nbnd;
      nbnd = new long [nproc];
      
      //  maximum number of boundary/interface nodes on subdomain
      maxnbn=0;
      for (i=0;i<nproc;i++){
        k=0;
        for (j=0;j<nnsd[i];j++){
          if (amultip[allnodes[i][j]]>1)  k++;
        }
        nbnd[i]=k;
	if (maxnbn<nbnd[i])
	  maxnbn=nbnd[i];
      }
      
      //  the total number of all boundary nodes
      //  it has to be computed once again because additional
      //  boundary/interface nodes have been defined
      tnbn=0;
      for (i=0;i<tnnp;i++){
        if (amultip[i]!=1)  tnbn++;
      }

      //  loop over the number of slaves
      for (i=1;i<nproc;i++){
	k=domproc[i];
	buff[0]=nbnd[k];
	buff[1]=maxnbn;
	MPI_Send (buff,buffsize,MPI_LONG,i,myrank,MPI_COMM_WORLD);
      }
      
      k=domproc[0];
      nbn=nbnd[k];
    }
    else{
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
      
      nbn = buff[0];
      maxnbn = buff[1];
    }
    MPI_Barrier (MPI_COMM_WORLD);
    
    delete [] buff;
    
    break;
  }
  default:{
    par_print_err(myrank,proc_name,"unknown type of mesh description is required",__FILE__, __LINE__, __func__);
  }
  }
  MPI_Barrier (MPI_COMM_WORLD);
  
  //  number of internal nodes
  nin = nn - nbn;
  
  if (myrank==0){
    //  list of the numbers of internal nodes on subdomains
    if (nind != NULL)
      delete [] nind;
    nind = new long [nproc];
    
    for (i=0;i<nproc;i++){
      nind[i]=nnsd[i]-nbnd[i];
    }
    
  }
  
  // *******************
  //  auxiliary output
  // *******************
  fprintf (out,"\n\n\n the number of boundary/interface nodes on subdomain (nbn)             %ld\n",nbn);
  fprintf (out,"\n the number of internal nodes on subdomain (nin)                       %ld\n",nin);
  fprintf (out,"\n the number of all nodes on subdomain (nn)                             %ld\n",nn);
  fprintf (out,"\n the maximum number of all nodes on subdomain (maxnn)                  %ld\n",maxnn);
  fprintf (out,"\n the maximum number of boundary/interface nodes on subdomain (maxnbn)  %ld\n",maxnbn);
  if (myrank==0){
    fprintf (out,"\n\n\n the numbers of boundary/interface nodes on each subdomain (nbnd)\n\n");
    for (i=0;i<nproc;i++){
      fprintf (out,"\n %ld",nbnd[i]);
    }
    fprintf (out,"\n\n\n the numbers of internal nodes on each subdomain (nind)\n\n");
    for (i=0;i<nproc;i++){
      fprintf (out,"\n %ld",nind[i]);
    }
  }  

}




/**
   function assembles the array nodmultip
   
   the array nodmultip can be assembled in the function assemble_multip
   but the coupled DOFs can generate additional boundary/interface nodes
   
   @param ltg - local to global correspondence
   @param domproc - array containing domain-processor correspondence
   @param out - output file for auxiliary print
   @param proc_name - processor name
   
   27.7.2009, JK
*/
void partop::assemble_nodmultip (long *ltg,long *domproc,FILE *out,char *proc_name)
{
  long i,j,k,buffsize;
  long *buff;
  MPI_Status stat;
  
  switch (md){
  case bound_nodes:{
    //  ltg=-1 for the internal nodes
    //  ltg>-1 for the interface/boundary nodes
    
    buffsize=tnbn;
    buff = new long [buffsize];
    
    if (myrank==0){
      for (i=0;i<tnbn;i++){
        buff[i]=bmultip[i];
      }
      
      for (i=1;i<nproc;i++){
        MPI_Send (buff,buffsize,MPI_LONG,i,myrank,MPI_COMM_WORLD);
      }
    }
    else{
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    }
    MPI_Barrier (MPI_COMM_WORLD);
    
    if (nodmultip!=NULL)
      delete [] nodmultip;
    nodmultip = new long [nn];
    
    for (i=0;i<nn;i++){
      if (ltg[i]>-1){
        nodmultip[i]=buff[ltg[i]];
      }
      else{
        nodmultip[i]=1;
      }
    }
    
    delete [] buff;
    
    break;
  }
    
    
    
  case all_nodes:
  case neg_bound_nodes:{
    
    if (maxnn==0){
      par_print_err(myrank,proc_name,"maximum number of nodes on one subdomain is equal to zero",__FILE__, __LINE__, __func__);
      par_print_err(myrank,proc_name,"function numbers_of_all_nodes_on_subdomains has to be called first",__FILE__, __LINE__, __func__);
      abort ();
    }
    if (myrank==0){
      if (nnsd==NULL){
	par_print_err(myrank,proc_name,"array nnsd is not assembled",__FILE__, __LINE__, __func__);
	par_print_err(myrank,proc_name,"function partop::numbers_of_all_nodes_on_subdomains has to be called first",__FILE__, __LINE__, __func__);
      }
      if (amultip==NULL){
	par_print_err(myrank,proc_name,"array amultip is not assembled",__FILE__, __LINE__, __func__);
	par_print_err(myrank,proc_name,"function partop::assemble_multip has to be called first",__FILE__, __LINE__, __func__);
      }
      if (allnodes==NULL){
	par_print_err(myrank,proc_name,"array allnodes is not assembled",__FILE__, __LINE__, __func__);
	par_print_err(myrank,proc_name,"function partop::assemble_multip has to be called first",__FILE__, __LINE__, __func__);
      }
    }
    
    buffsize=maxnn;
    buff = new long [buffsize];
    
    if (myrank==0){
      for (i=1;i<nproc;i++){
        k=domproc[i];
        for (j=0;j<nnsd[k];j++){
          buff[j]=amultip[allnodes[k][j]];
        }
        MPI_Send (buff,buffsize,MPI_LONG,i,myrank,MPI_COMM_WORLD);
      }
      
      k=domproc[0];
      for (j=0;j<nnsd[k];j++){
        buff[j]=amultip[allnodes[k][j]];
      }
    }
    else{
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    }
    
    //  nodal multiplicity
    if (nodmultip != NULL)
      delete [] nodmultip;
    nodmultip = new long [nn];
    for (i=0;i<nn;i++){
      nodmultip[i] = buff[i];
    }
    
    delete [] buff;
    
    break;
  }
  default:{
    par_print_err(myrank,proc_name,"unknown type of mesh description is required",__FILE__, __LINE__, __func__);
  }
  }
  MPI_Barrier (MPI_COMM_WORLD);

  
  // *******************
  //  auxiliary output
  // *******************
  fprintf(out,"\n\n\n Multiplicity of nodes on subdomain (nodmultip)\n\n");  
  for (i=0;i<nn;i++){
    fprintf(out,"%ld   %ld\n",i,nodmultip[i]);  
  }
  
  
}


/**
   function assembles global numbers of nodes
   
   the following arrays are assembled:
   
   gnbn - (M) global numbers of interface/boundary nodes
   gnin - (M) global numbers of internal nodes
   gnbndom -  array containing global numbers of interface/boundary nodes on subdomain

   @param domproc - array containing domain-processor correspondence
   @param out - output file for auxiliary print
   @param proc_name - processor name
   
   JK, 30.7.2009
*/
void partop::node_global_numbers (long *domproc,FILE *out,char *proc_name)
{
  long i,j,k,ii,jj,kk;
  long *buff,buffsize;
  MPI_Status stat;
  
  if (myrank==0){
    //  array containing global numbers of internal nodes
    if (gnin != NULL){
      for (i=0;i<nproc;i++){
	delete [] gnin[i];
      }
      delete [] gnin;
    }
    gnin = new long* [nproc];
    for (i=0;i<nproc;i++){
      gnin[i] = new long [nind[i]];
    }
    //  array containing global numbers of interface/boundary nodes
    if (gnbn != NULL){
      for (i=0;i<nproc;i++){
	delete [] gnbn[i];
      }
      delete [] gnbn;
    }
    gnbn = new long* [nproc];
    for (i=0;i<nproc;i++){
      gnbn[i] = new long [nbnd[i]];
    }
  }
  
  //  array containing global numbers of interface/boundary nodes on subdomain
  if (gnbndom != NULL){
    delete [] gnbndom;
  }
  gnbndom = new long [nbn];
  

  switch (md){
  case bound_nodes:{
    par_print_err(myrank,proc_name,"mesh description bound_nodes cannot be used",__FILE__, __LINE__, __func__);
    break;
  }
  case metis:
  case all_nodes:{
    
    if (myrank==0){
      
      if (nnsd==NULL){
	par_print_err(myrank,proc_name,"array nnsd is not assembled",__FILE__, __LINE__, __func__);
	par_print_err(myrank,proc_name,"function partop::numbers_of_all_nodes_on_subdomains has to be called first",__FILE__, __LINE__, __func__);
      }
      if (amultip==NULL){
	par_print_err(myrank,proc_name,"array amultip is not assembled",__FILE__, __LINE__, __func__);
	par_print_err(myrank,proc_name,"function partop::assemble_multip has to be called first",__FILE__, __LINE__, __func__);
      }
      if (allnodes==NULL){
	par_print_err(myrank,proc_name,"array allnodes is not assembled",__FILE__, __LINE__, __func__);
	par_print_err(myrank,proc_name,"function partop::assemble_multip has to be called first",__FILE__, __LINE__, __func__);
      }

      //  loop over the number of subdomains
      for (i=0;i<nproc;i++){
	//  index in the array gnbn
	ii=0;
	//  index in the array gnin
	jj=0;
	//  loop over the number of all nodes on subdomain
	for (j=0;j<nnsd[i];j++){
	  if (amultip[allnodes[i][j]]>1){
	    //  boundary/interface node
	    
	    //  global number of boundary/interface node
	    gnbn[i][ii]=allnodes[i][j];
	    ii++;
	  }
	  if (amultip[allnodes[i][j]]==1){
	    //  internal node
	    
	    //  global number of internal node
	    gnin[i][jj]=allnodes[i][j];
	    jj++;
	  }
	  
	}
	//printf ("\n ii %ld     jj %ld",ii,jj);
      }
      
    }
    break;
  }
  case neg_bound_nodes:{
    
    if (myrank==0){
      
      if (nnsd==NULL){
	par_print_err(myrank,proc_name,"array nnsd is not assembled",__FILE__, __LINE__, __func__);
	par_print_err(myrank,proc_name,"function partop::numbers_of_all_nodes_on_subdomains has to be called first",__FILE__, __LINE__, __func__);
      }
      if (allnodes==NULL){
	par_print_err(myrank,proc_name,"array allnodes is not assembled",__FILE__, __LINE__, __func__);
	par_print_err(myrank,proc_name,"function partop::assemble_multip has to be called first",__FILE__, __LINE__, __func__);
      }

      //  loop over the number of subdomains
      for (i=0;i<nproc;i++){
	//  index in the array gnbn
	ii=0;
	//  index in the array gnin
	jj=0;
	//  loop over the number of all nodes on subdomain
	for (j=0;j<nnsd[i];j++){
	  if (allnodes[i][j]<0){
	    //  boundary/interface node
	    gnbn[i][ii]=0-allnodes[i][j]-2;
	    ii++;
	  }
	  if (allnodes[i][j]>-1){
	    //  internal node
	    gnin[i][jj]=allnodes[i][j];
	    jj++;
	  }
	}
      }
    }
    
    break;
  }
  default:{
    par_print_err(myrank,proc_name,"unknown type of mesh description is required",__FILE__, __LINE__, __func__);
  }
  }
  
  buffsize=maxnbn;
  buff = new long [buffsize];
  
  if (myrank==0){
    
    //  loop over the number of slaves
    for (i=1;i<nproc;i++){
      //  subdomain id
      k=domproc[i];
      //  loop over the number of boundary/interface nodes
      for (j=0;j<nbnd[k];j++){
	buff[j]=gnbn[k][j];
      }
      MPI_Send (buff,buffsize,MPI_LONG,i,myrank,MPI_COMM_WORLD);
    }

    //  subdomain id
    k=domproc[0];
    //  loop over the number of boundary/interface nodes
    for (j=0;j<nbnd[k];j++){
      buff[j]=gnbn[k][j];
    }
    
  }
  else{
    MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  
  for (i=0;i<nbn;i++){
    gnbndom[i]=buff[i];
  }
  
  delete [] buff;
  
  
  // *******************
  //  auxiliary output
  // *******************
  fprintf (out,"\n\n\n global numbers of interface/boundary nodes on subdomain (gnbndom)");
  for (i=0;i<nbn;i++){
    fprintf (out,"\n %6ld   %6ld",i,gnbndom[i]);
  }
  if (myrank==0){
    fprintf (out,"\n\n\n global numbers of interface/boundary nodes on subdomain (gnbn)");
    for (i=0;i<nproc;i++){
      fprintf (out,"\n\n domain %ld",i);
      for (j=0;j<nbnd[i];j++){
	fprintf (out,"\n %6ld   %6ld",j,gnbn[i][j]);
      }
    }
    
    fprintf (out,"\n\n\n global numbers of internal nodes on subdomain (gnin)");
    for (i=0;i<nproc;i++){
      fprintf (out,"\n\n domain %ld",i);
      for (j=0;j<nind[i];j++){
	fprintf (out,"\n %6ld   %6ld",j,gnin[i][j]);
      }
    }
  }

}

/**
   function assembles local numbers of nodes
   
   the following arrays are assembled:
   
   lnbndom - local numbers of interface/boundary nodes
   lnindom - local numbers of internal nodes

   @param out - output file for auxiliary print
   
   JK, 16.9.2009
*/
void partop::node_local_numbers (FILE *out)
{
  long i,j,k,ii,jj,kk;
  long *buff,buffsize;
  MPI_Status stat;
  
  //  array containing local numbers of internal nodes on subdomain
  if (lnindom != NULL){
    delete [] lnindom;
  }
  lnindom = new long [nin];
  //  array containing local numbers of interface/boundary nodes on subdomain
  if (lnbndom != NULL){
    delete [] lnbndom;
  }
  lnbndom = new long [nbn];
  
  j=0;
  for (i=0;i<nn;i++){
    if (nodmultip[i]>1){
      lnbndom[j]=i;
      j++;
    }
  }
  j=0;
  for (i=0;i<nn;i++){
    if (nodmultip[i]==1){
      lnindom[j]=i;
      j++;
    }
  }
  
  
  // *******************
  //  auxiliary output
  // *******************
  fprintf (out,"\n\n\n local numbers of interface/boundary nodes on subdomain (lnbndom)");
  for (i=0;i<nbn;i++){
    fprintf (out,"\n %6ld   %6ld",i,lnbndom[i]);
  }
  fprintf (out,"\n\n\n local numbers of internal nodes on subdomain (lnindom)");
  for (i=0;i<nin;i++){
    fprintf (out,"\n %6ld   %6ld",i,lnindom[i]);
  }

}


/**
   function assembles coarse numbers of nodes
   
   the following arrays are assembled:
   
   icnbnmas - (M) coarse numbers of interface/boundary nodes
   icmultip - (M) number of multiplicity of boundary/interface nodes
   icnbn - array containing coarse numbers of interface/boundary nodes on one subdomain

   @param domproc - array containing domain-processor correspondence
   @param out - output file for auxiliary print
   @param proc_name - processor name
   
   JK, 29.7.2009
*/
void partop::node_coarse_numbers (long *domproc,FILE *out,char *proc_name)
{
  long i,j,k,l,ii,jj,ln,cn,buffsize;
  long *av,*buff;
  av=NULL;
  MPI_Status stat;
  
  if (myrank==0){
    
    
    //  array containing coarse numbers of interface/boundary nodes
    if (icnbnmas != NULL){
      for (i=0;i<nproc;i++){
	delete [] icnbnmas[i];
      }
      delete [] icnbnmas;
    }
    icnbnmas = new long* [nproc];
    for (i=0;i<nproc;i++){
      icnbnmas[i] = new long [nbnd[i]];
    }
    

    switch (md){
    case bound_nodes:{
      
      if (nnsd==NULL){
	par_print_err(myrank,proc_name,"array nnsd is not assembled",__FILE__, __LINE__, __func__);
	par_print_err(myrank,proc_name,"function partop::numbers_of_all_nodes_on_subdomains has to be called first",__FILE__, __LINE__, __func__);
      }
      if (bnodes==NULL){
	par_print_err(myrank,proc_name,"array bnodes is not assembled",__FILE__, __LINE__, __func__);
	par_print_err(myrank,proc_name,"function partop::assemble_multip has to be called first",__FILE__, __LINE__, __func__);
      }
      
      //  loop over the number of subdomains
      for (i=0;i<nproc;i++){
	//  index
	l=0;
	//  loop over the number of boundary/interface nodes
	for (j=0;j<nnsd[i];j++){
	  k=bnodes[i][j];
	  if (k>-1){
	    icnbnmas[i][l]=k;
	    l++;
	  }
	}
      }
      
      break;
    }
    case metis:
    case all_nodes:
    case neg_bound_nodes:{
      
      if (tnnp==0){
	par_print_err(myrank,proc_name,"total number of nodes in the problem (tnnp) is zero",__FILE__, __LINE__, __func__);
	par_print_err(myrank,proc_name,"function partop::assemble_multip has to be called first",__FILE__, __LINE__, __func__);
      }
      if (amultip==NULL){
	par_print_err(myrank,proc_name,"array amultip is not assembled",__FILE__, __LINE__, __func__);
	par_print_err(myrank,proc_name,"function partop::assemble_multip has to be called first",__FILE__, __LINE__, __func__);
      }
      if (allnodes==NULL){
	par_print_err(myrank,proc_name,"array allnodes is not assembled",__FILE__, __LINE__, __func__);
	par_print_err(myrank,proc_name,"function partop::assemble_multip has to be called first",__FILE__, __LINE__, __func__);
      }
      
      j=0;
      av = new long [tnnp];
      for (i=0;i<tnnp;i++){
	av[i]=amultip[i];
	if (av[i]==1)
	  av[i]=-1;
	if (av[i]>1){
	  av[i]=j;
	  j++;
	}
      }
      break;
    }
    default:{
      print_err("unknown type of mesh description is required", __FILE__, __LINE__, __func__);
    }
    }
    
    switch (md){
    case bound_nodes:{
      break;
    }
    case metis:
    case all_nodes:{
      //  loop over the number of subdomains
      for (i=0;i<nproc;i++){
	//  index
	l=0;
	//  loop over the number of boundary/interface nodes
	for (j=0;j<nnsd[i];j++){
	  k=allnodes[i][j];
	  if (av[k]>-1){
	    icnbnmas[i][l]=av[k];
	    l++;
	  }
	}
      }
      break;
    }
    case neg_bound_nodes:{
      //  loop over the number of subdomains
      for (i=0;i<nproc;i++){
	//  index
	l=0;
	//  loop over the number of boundary/interface nodes
	for (j=0;j<nnsd[i];j++){
	  k=allnodes[i][j];
	  if (k<0){
	    k=0-k-2;
	    icnbnmas[i][j]=av[k];
	    l++;
	  }
	}
      }
      break;
    }  
    default:{
      print_err("unknown type of mesh description is required", __FILE__, __LINE__, __func__);
    }
    }
    
    
    //  number of multiplicity of boundary/interface nodes
    if (icmultip!=NULL){
      delete [] icmultip;
    }
    icmultip = new long [tnbn];
    for (i=0;i<tnbn;i++){
      icmultip[i]=0;
    }
    
    //  loop over the number of subdomains
    for (i=0;i<nproc;i++){
      //  loop over the number of boundary/interface nodes
      for (j=0;j<nbnd[i];j++){
	icmultip[icnbnmas[i][j]]++;
      }
    }

    if (av!=NULL)
      delete [] av;
  }

  buffsize=maxnbn;
  buff = new long [buffsize];
  
  //  array containing coarse numbers of interface/boundary nodes on one subdomain
  if (icnbn!=NULL){
    delete [] icnbn;
  }
  icnbn = new long [nbn];
  
  if (myrank==0){
    
    //  loop over the number of slaves
    for (i=1;i<nproc;i++){
      //  subdomain id
      k=domproc[i];
      //  loop over the number of boundary/interface nodes
      for (j=0;j<nbnd[k];j++){
	buff[j]=icnbnmas[k][j];
      }
      MPI_Send (buff,buffsize,MPI_LONG,i,myrank,MPI_COMM_WORLD);
    }

    //  subdomain id
    k=domproc[0];
    //  loop over the number of boundary/interface nodes
    for (j=0;j<nbnd[k];j++){
      buff[j]=icnbnmas[k][j];
    }
    
  }
  else{
    MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
  }
  MPI_Barrier (MPI_COMM_WORLD);
  
  for (i=0;i<nbn;i++){
    icnbn[i]=buff[i];
  }
  
  delete [] buff;
  

  // *******************
  //  auxiliary output
  // *******************
  if (myrank==0){
    fprintf (out,"\n\n\n coarse numbers of interface/boundary nodes on subdomain");
    for (i=0;i<nproc;i++){
      fprintf (out,"\n\n domain %ld",i);
      for (j=0;j<nbnd[i];j++){
	fprintf (out,"\n %6ld   %6ld",j,icnbnmas[i][j]);
      }
    }
    
    fprintf (out,"\n\n\n coarse-local correspondence \n");
    for (i=0;i<tnbn;i++){
      fprintf (out,"\n coarse node %6ld, number of nodes %2ld",i,icmultip[i]);
    }
  }
  fprintf (out,"\n\n\n coarse numbers of interface/boundary nodes on subdomain");
  for (i=0;i<nbn;i++){
    fprintf (out,"\n %6ld   %6ld",i,icnbn[i]);
  }

}



/**
   function assembles coarse - global numbers map
   
   the following arrays are assembled:
   
   gnbncn - (M) global numbers of boundary/interface nodes of the coarse node
   sid - (M) subdomain id of interface/boundary nodes of the coarse node
   
   @param out - output file for auxiliary print
   @param proc_name - processor name
   
   JK, 30.7.2009
*/
void partop::node_coarse_global_map (FILE *out,char *proc_name)
{
  long i,j,k,ii,jj,ln,cn;
  
  if (myrank==0){
    
    if (tnbn==0){
      par_print_err(myrank,proc_name,"total number of boundary/interface nodes (tnbn) is zero",__FILE__, __LINE__, __func__);
      par_print_err(myrank,proc_name,"function partop::assemble_nbnd_nind or partop::update_multip or partop::assemble_multip has to be called first",__FILE__, __LINE__, __func__);
    }
    if (icmultip==NULL){
      par_print_err(myrank,proc_name,"array icmultip is not assembled",__FILE__, __LINE__, __func__);
      par_print_err(myrank,proc_name,"function partop::node_coarse_numbers has to be called first",__FILE__, __LINE__, __func__);
    }
    if (icnbnmas==NULL){
      par_print_err(myrank,proc_name,"array icnbnmas is not assembled",__FILE__, __LINE__, __func__);
      par_print_err(myrank,proc_name,"function partop::node_coarse_numbers has to be called first",__FILE__, __LINE__, __func__);
    }
    if (gnbn==NULL){
      par_print_err(myrank,proc_name,"array gnbn is not assembled",__FILE__, __LINE__, __func__);
      par_print_err(myrank,proc_name,"function partop::node_global_numbers has to be called first",__FILE__, __LINE__, __func__);
    }

    //  global numbers of boundary/interface nodes of the coarse node
    if (gnbncn!=NULL){
      for (i=0;i<tnbn;i++){
	delete [] gnbncn[i];
      }
      delete [] gnbncn;
    }
    gnbncn = new long* [tnbn];
    for (i=0;i<tnbn;i++){
      gnbncn[i]=new long [icmultip[i]];
    }
    //  subdomain id of interface/boundary nodes of the coarse node
    if (sid!=NULL){
      for (i=0;i<tnbn;i++){
	delete [] sid[i];
      }
      delete [] sid;
    }
    sid = new long* [tnbn];
    for (i=0;i<tnbn;i++){
      sid[i] = new long [icmultip[i]];
    }
    
    //  loop over the number of subdomains
    for (i=0;i<tnbn;i++){
      icmultip[i]=0;
    }
    
    //  loop over the number of subdomains
    for (i=0;i<nproc;i++){
      //  loop over the number of boundary/interface nodes
      for (j=0;j<nbnd[i];j++){
	//  coarse number
	cn=icnbnmas[i][j];
	//  global number
	ln=gnbn[i][j];
	
	//  global number of boundary/interface node
	gnbncn[cn][icmultip[cn]]=ln;
	//  subdomain id if boundary/interface node
	sid[cn][icmultip[cn]]=i;
	icmultip[cn]++;
      }
    }
  }
  
  // *******************
  //  auxiliary output
  // *******************
  if (myrank==0){
    fprintf (out,"\n\n\n coarse-global correspondence \n");
    for (i=0;i<tnbn;i++){
      fprintf (out,"\n coarse node %6ld, number of nodes %2ld",i,icmultip[i]);
      for (j=0;j<icmultip[i];j++){
	fprintf (out,"   node %ld, global number %6ld, subdomain id %4ld",j,gnbncn[i][j],sid[i][j]);
      }
    }
    fprintf (out,"\n\n\n coarse-global correspondence \n");
    for (i=0;i<tnbn;i++){
      fprintf (out,"\n coarse node %6ld",i);
      for (j=0;j<icmultip[i];j++){
	fprintf (out,"   ggn %6ld, sub.%4ld",gnbncn[i][j],sid[i][j]);
      }
    }
  }
  
}




/**
   function generates the Schur complement ordering
   it takes into account the coupled DOFs
   
   ordering of subdomains is performed
   ordering of the coarse problem is in selnodes::schur_ordering
   
   function determines:
   ndof - the number of all DOFs (unknowns)
   nidof - the number of internal DOFs (unknowns)
   nbdof - the number of boundary/interface DOFs (unknowns)
   
   @param top - pointer to general topology
   @param domproc - array containing domain-processor correspondence
   @param out - output file
   @param proc_name - processor name
   
   JK, 29.7.2009
*/
long partop::schur_ordering (gtopology *top,long *domproc,FILE *out,char *proc_name)
{
  long i,j,k,l,m,ndofn,nid,ncdof,buffsize;
  long *aux,**auxdofind,*buff;
  MPI_Status stat;
  
  switch (md){
  case bound_nodes:{
    
    // ********************************
    //  ordering of internal unknowns
    // ********************************
    ndof=1;
    for (i=0;i<nin;i++){
      //  node id (in local ordering)
      nid=lnindom[i];
      //  number of DOFs at required node
      ndofn = top->give_ndofn (nid);
      
      for (j=0;j<ndofn;j++){
	k=top->give_dof (nid,j);
	if (k>0){
	  top->save_dof (nid,j,ndof);
	  ndof++;
	}
      }
    }
    //  number of internal DOFs (unknowns)
    nidof = ndof-1;
    
    
    // ********************************
    //  ordering of boundary unknowns
    // ********************************
    for (i=0;i<nbn;i++){
      //  node id (in local ordering)
      nid=lnbndom[i];
      //  number of DOFs at required node
      ndofn = top->give_ndofn (nid);
      
      for (j=0;j<ndofn;j++){
	k=top->give_dof (nid,j);
	if (k>0){
	  top->save_dof (nid,j,ndof);
	  ndof++;
	}
      }
    }
    ndof--;
    
    
    //  contributions from elements
    for (i=0;i<top->ne;i++){
      if (top->give_cne (i)==1){
	for (j=0;j<top->give_ndofe(i);j++){
	  if (top->give_dof (i,j)>ndof)  ndof=top->give_dof (i,j);
	}
      }
    }
    
    //  number of boundary DOFs (unknowns)
    nbdof = ndof-nidof;
    
    break;
  }
  case metis:
  case all_nodes:{
    
    // *********************************************
    //  distribution of the array dofind to slaves
    // *********************************************
    if (maxndof==0){
      par_print_err(myrank,proc_name,"maximum number of DOFs on one subdomain is equal to zero",__FILE__, __LINE__, __func__);
      par_print_err(myrank,proc_name,"function numbers_of_all_dofs_on_subdomains has to be called first",__FILE__, __LINE__, __func__);
    }
    
    buffsize=maxndof;
    buff = new long [buffsize];
    for (i=0;i<buffsize;i++){
      buff[i]=0;
    }
    
    if (myrank==0){
      
      if (nnsd==NULL){
	par_print_err(myrank,proc_name,"array nnsd is not assembled",__FILE__, __LINE__, __func__);
	par_print_err(myrank,proc_name,"function partop::numbers_of_all_nodes_on_subdomains has to be called first",__FILE__, __LINE__, __func__);
      }
      if (allnodes==NULL){
	par_print_err(myrank,proc_name,"array allnodes is not assembled",__FILE__, __LINE__, __func__);
	par_print_err(myrank,proc_name,"function partop::assemble_multip has to be called first",__FILE__, __LINE__, __func__);
      }
      if (ndofnmas==NULL){
	par_print_err(myrank,proc_name,"array ndofnmas is not assembled",__FILE__, __LINE__, __func__);
	par_print_err(myrank,proc_name,"function partop::ndofn_master has to be called first",__FILE__, __LINE__, __func__);
      }
      if (dofind==NULL){
	par_print_err(myrank,proc_name,"array dofind is not assembled",__FILE__, __LINE__, __func__);
	par_print_err(myrank,proc_name,"function partop::update_multip has to be called first",__FILE__, __LINE__, __func__);
      }
      
      
      //  loop over the number of slaves
      for (i=1;i<nproc;i++){
	//  subdomain id
	k=domproc[i];
	
	//  index in the array buff
	m=0;
	//  loop over the number of all nodes on the k-th subdomain
	for (j=0;j<nnsd[k];j++){
	  //  global node number
	  nid=allnodes[k][j];
	  //printf ("  nid %ld ",nid);
	  //  the number of DOFs in node
	  ndofn=ndofnmas[nid];
	  //  loop over the number of DOFs in node
	  for (l=0;l<ndofn;l++){
	    buff[m]=dofind[nid][l];
	    m++;
	  }
	}
	//printf ("\n m %ld",m);
	
	MPI_Send (buff,buffsize,MPI_LONG,i,myrank,MPI_COMM_WORLD);
      }
      //  master part
      
      //  subdomain id
      k=domproc[0];
      
      //  index in the array buff
      m=0;
      //  loop over the number of all nodes on the k-th subdomain
      for (j=0;j<nnsd[k];j++){
	//  global node number
	nid=allnodes[k][j];
	//  the number of DOFs in node
	ndofn=ndofnmas[nid];
	//  loop over the number of DOFs in node
	for (l=0;l<ndofn;l++){
	  buff[m]=dofind[nid][l];
	  m++;
	}
      }
      
    }
    else{
      MPI_Recv (buff,buffsize,MPI_LONG,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&stat);
    }
    MPI_Barrier (MPI_COMM_WORLD);
    
    auxdofind = new long* [nn];
    for (i=0;i<nn;i++){
      auxdofind[i] = new long [top->give_ndofn(i)];
    }
    
    //  index in the array buff
    k=0;
    //  loop over the number of all nodes on subdomain
    for (i=0;i<nn;i++){
      //  the number of DOFs in node
      ndofn=top->give_ndofn(i);
      //  loop over the number of DOFs in node
      for (j=0;j<ndofn;j++){
	auxdofind[i][j]=buff[k];
	k++;
      }
    }
    
    delete [] buff;
    
    // ****************************************************
    //  end of distribution of the array dofind to slaves
    // ****************************************************
    
    //  searching of maximum code number indicator
    ndof=0;
    //  loop over the number of all nodes in the problem
    for (i=0;i<nn;i++){
      //  the number of DOFs in node
      ndofn=top->give_ndofn (i);
      //  loop over the number of DOFs in node
      for (j=0;j<ndofn;j++){
	k=top->give_dof (i,j);
	if (k>ndof)  ndof=k;
      }
    }
    ndof--;
    if (ndof<0)  ndof=0;
    aux = new long [ndof];
    for (i=0;i<ndof;i++){
      aux[i]=-1;
    }
    
    
    
    //  number of actual DOF
    ndof=1;
    
    // ******************************
    //  generation of internal DOFs
    // ******************************
    
    //  loop over the number of nodes on subdomains
    for (j=0;j<nn;j++){
      //  the number of DOFs
      ndofn=top->give_ndofn (j);
      //  loop over the number of DOFs in node
      for (l=0;l<ndofn;l++){
	if (auxdofind[j][l]==0){
	  //  this is internal DOF
	  
	  k=top->give_dof (j,l);
	  if (k<0)  continue;
	  if (k==0)  continue;
	  if (k==1){
	    top->gnodes[j].cn[l]=ndof;  ndof++;
	  }
	  if (k>1){
	    if (aux[k-2]==-1){
	      top->gnodes[j].cn[l]=ndof;
	      aux[k-2]=ndof;
	      ndof++;
	    }
	    else{
	      top->gnodes[j].cn[l]=aux[k-2];
	    }
	  }
	}
      }
    }
    
    //  the number of internal DOFs (unknowns)
    nidof=ndof-1;
    
    // ****************************************
    //  generation of boundary/interface DOFs
    // ****************************************
    
    //  loop over the number of all nodes on subdomains
    for (j=0;j<nn;j++){
      //  the number of DOFs
      ndofn=top->give_ndofn (j);
      //  loop over the number of DOFs in node
      for (l=0;l<ndofn;l++){
	if (auxdofind[j][l]==1){
	  //  this is boundary/interface DOF
	  
	  k=top->give_dof (j,l);
	  if (k<0)  continue;
	  if (k==0)  continue;
	  if (k==1){
	    top->gnodes[j].cn[l]=ndof;  ndof++;
	  }
	  if (k>1){
	    if (aux[k-2]==-1){
	      top->gnodes[j].cn[l]=ndof;
	      aux[k-2]=ndof;
	      ndof++;
	    }
	    else{
	      top->gnodes[j].cn[l]=aux[k-2];
	    }
	  }
	}
      }
    }
    ndof--;
    
    //  number of boundary DOFs (unknowns)
    nbdof = ndof-nidof;
    
    if (aux != NULL)
      delete [] aux;
    
    for (i=0;i<nn;i++){
      delete [] auxdofind[i];
    }
    delete [] auxdofind;
    
    break;
  }
  default:{
    print_err("unknown type of mesh description is required", __FILE__, __LINE__, __func__);
  }
  }
  
  //  state of code numbers is changed
  //  code numbers are generated
  top->cnstate=1;
  
  
  // *******************
  //  auxiliary output
  // *******************
  fprintf (out,"\n\n nidof   %6ld",nidof);
  fprintf (out,"\n\n nbdof   %6ld",nbdof);
  fprintf (out,"\n\n ndof    %6ld",ndof);
  fprintf (out,"\n\n\n kontrola kodovych cisel \n\n");
  //  loop over the number of all nodes on subdomains
  for (j=0;j<nn;j++){
    fprintf (out,"\n node %6ld  ",j);
    //  the number of DOFs
    ndofn=top->give_ndofn (j);
    //  loop over the number of DOFs in node
    for (l=0;l<ndofn;l++){
      fprintf (out,"  %6ld",top->gnodes[j].cn[l]);
    }
  }
  
  
  return ndof;
}

/**
   function prepares all data needed

   @param domproc - array containing domain-processor correspondence
   @param ltg - local to global correspondence
   @param top - pointer to the sequential general topology
   @param out - output file (used for auxiliary output)
   @param proc_name - processor name
   
   JK, 16.9.2009
*/
void partop::prepare_data (long *domproc,long *ltg,gtopology *top,FILE *out,char *proc_name)
{
  initiation (top,ltg);
  numbers_of_all_nodes_on_subdomains (domproc,out);
  assemble_multip (ltg,domproc,out,proc_name);
  coupled_dofs (top,domproc,out);
  numbers_of_all_dofs_on_subdomains (top,domproc,out);
  
  if (md==all_nodes || md==neg_bound_nodes){
    ndofn_master (top,domproc,out,proc_name);
    dofind_master (top,domproc,out,proc_name);
    update_multip (top,out,proc_name);
  }

  assemble_nbnd_nind (ltg,domproc,out,proc_name);
  assemble_nodmultip (ltg,domproc,out,proc_name);
  if (md==all_nodes || md==neg_bound_nodes){
    node_global_numbers (domproc,out,proc_name);
  }
  node_local_numbers (out);

  node_coarse_numbers (domproc,out,proc_name);
  
  if (md==all_nodes || md==neg_bound_nodes){
    node_coarse_global_map (out,proc_name);
  }
  
}
